package molmaster;

import java.util.*;
import javax.vecmath.*;

import molmaster.model.*;

/** Contains the information used to display and
 *  keep track of a particular atom in the system.
 *
 * @author  rjpower
 */
public class Atom extends MolObject implements WatchedObject, java.util.Comparator {
    
    protected Vector3d  location;
    protected List      listeners = new ArrayList(1);
    protected List      bondList = new ArrayList(1);
    protected List      connected = new ArrayList(1);
    protected AtomAttributes attributes;
    
    protected int atomID;
    
    public MolAttributes getAttributes() { return attributes; }
    
    /** Creates a new Atom with default attributes.
     */    
    public Atom() {
        
        location = new Vector3d(0,0,0);
        atomID = 0;
        
        attributes = new AtomAttributes();
        attributes.bondingRadius = 0;
        attributes.radius = 0.1;
        attributes.atomName = "Unnamed";
        attributes.color = new Color3f( 0.3f, 0.3f, 0.3f );
    }
    
    /**
     */    
    public Atom(AtomAttributes a){
        this();
        attributes = a;
    }
    
    public Atom(AtomAttributes a, Vector3d location, int id ) {
        this(a);
        this.atomID = id;
        this.location = location;
    }
    
    public Atom( Atom copy ){
        this ( new AtomAttributes(copy.attributes), new Vector3d(copy.getLocation()), copy.atomID );
        listeners = copy.listeners;
        owner = copy.owner;
        bondList = copy.bondList;//new ArrayList(copy.bondList);
        connected = copy.connected;//new ArrayList(copy.connected);
    }
    
    /** Return all Bond's this Atom is a member of.
     */    
    public List getBonds(){ return bondList; }
    /** Add the given bond to the list of connected atoms.
     */    
    public void addBond(Bond b) { 
        bondList.add(b); 
        if ( b.getStart() != this ) {
            connected.add( b.getStart() );
        }else {
            connected.add( b.getEnd() );
        }
    }
    
    /** Returns all Atom's currently connected to this Atom.
     */    
    public List getConnected() { return connected; }
    /** Removes the given bond from this Atom's connected list.
     */    
    public void removeBond(Bond b) { 
        bondList.remove(b); 
        connected.remove( b.getStart() );
        connected.remove( b.getEnd() );
    }
    /** Returns the location of this atom in the scene.
     */    
    public Vector3d getLocation() {
        return location;
    }
    
    /** Sets the location of this atom in the scene.  <b> If this atom is
     * active ( visible in the scene ) this call will also change the
     * display of the AtomView representing the Atom to the new location. </b>
     */    
    public void setLocation(Vector3d l) {
        location.set(l);
        fire( MOVED, l );
    }
    
    public void fire( int type, Object value ) {
        Iterator i = listeners.iterator();
        while ( i.hasNext() ) {
            ((Listener)i.next()).wakeUp( this, type, value );
        }
    }
    /** Returns the ID number assigned to this atom by the owner file.
     */    
    public int getID() { return atomID; }
    /** Sets the id of this atom to the given number.
     */    
    public void setID(int id ) { atomID = id; }
    
    public String toString() {
        return attributes.atomName + " " + atomID;
    }
    
    /**
     */    
    public void addListener(Listener l) {
        listeners.add( l );
    }
    
    public void removeListener(Listener l) {
        listeners.add( l );
    }
    
    public int compare(Object o1, Object o2) {
        try {
            Atom a = (Atom)o1;
            Atom b = (Atom)o2;
            if ( a.getID() > b.getID() ) { return -1; }
            if ( a.equals(b) ) { return 0; }
            return 1;
        }catch(Exception e) {
            Log.writeLog("testing...");
            return 0;
        }
        
    }
   
    public int hashCode() {
        return getID();
    }
    
}
