package molmaster;

import molmaster.model.*;

/** Represents an atomic or hydrogen bond.
 *
 *
 * @author  rjpower
 */
public class Bond extends MolObject implements java.util.Comparator{
    
    /** Represents the default bond types.
     */    
    static public final int BOND_TYPE_NORMAL=   0x0001;
    /** Identifier for hydrogen bonds.
     */    
    static public final int BOND_TYPE_HYDROGEN= 0x0002;
    
    private Atom start;
    private Atom end;
    
    /** The type of this bond - regular or hydrogen.    */
    int bondType;
    
    /** The 'fudge factor' to use when determining the length of valid
     * bonds.  The covalent radius of the input atoms is multiplied by
     * this value to obtain the distance used.
     */    
    public static float fudge = 1.25f;
    /** The amount to fudge covalent radaii when determining the validity of a
     * hydrogen bond.
     */    
    public static float hydFudge = 1.5f;
    
    private double maxLength;
    
    /** The empty constructor is used to allow bond comparisons...  */
    public Bond() {
        super();
    }
    
    /** Creates a new bond between the two given atoms.
     * The maximum bond distance is determined by the attributes
     * of the two target atoms, and the bond type.
     * @param end The ending point for the bond.
     * @param start The starting point for the bond.
     * @param type The type of bond being created.
     */
    public Bond(Atom start, Atom end, int type) {
        this( start, end, type, 
        ((AtomAttributes)start.getAttributes()).bondingRadius + 
        ((AtomAttributes)end.getAttributes()).bondingRadius );
    }
    
    public Bond(Atom start, Atom end, int type, double maxLength ) {
        if ( maxLength < 0 ) {
            maxLength = ((AtomAttributes)start.getAttributes()).bondingRadius + 
            ((AtomAttributes)end.getAttributes()).bondingRadius;
            maxLength *= fudge;
        }
        
        if ( start.getUniqueID() < end.getUniqueID() ) {
            this.start = start;
            this.end = end;
        }
        else {
            this.start = end;
            this.end = start;
        }
        this.bondType = type;
        this.maxLength = maxLength;
        this.attributes.radius = molmaster.gui.BondView.BOND_RADIUS;
        this.attributes.color = new javax.vecmath.Color3f( 0.7f, 0.7f, 0.7f );
        this.uniqueID = MolModel.getModel().getUniqueID();
        
        if ( type == BOND_TYPE_HYDROGEN )
            this.maxLength *= hydFudge;
        
        owner = start.getOwner();
    }
    
    /** Returns the atom that originated this bond.  Note that the
     *  terms 'start' and end are applied arbitrarily and do not
     *  infer any special relationship between the bond and the 
     *  atoms.
     *
     * @return The atom originating this bond.
     */    
    public Atom getStart() { return start; }
    
    
    /** Returns the atom ending this bond.  Note that the
     *  terms 'start' and end are applied arbitrarily and do not
     *  infer any special relationship between the bond and the 
     *  atoms.
     *
     * @return The atom ending this bond.
     */    
    public Atom getEnd() { return end; }
    
    
    /** Returns the type of this bond.
     * @return An integer indicating the type of Bond - either hydrogen or normal.
     */    
    public int getType() { return bondType; }
    
    /** Sets the type of this bond.  The input should be either
     * BOND_TYPE_HYDROGEN or BOND_TYPE_NORMAL.
     * @param type The type to set this bond as.
     */    
    public void setType( int type ) { bondType = type; }
    
    /** Returns the current length of this bond.
     */    
    public double  getLength() {
        javax.vecmath.Vector3d a = start.getLocation();
        javax.vecmath.Vector3d b = end.getLocation();
        javax.vecmath.Vector3d c = new javax.vecmath.Vector3d();
        c.sub( a, b );
        return c.length();
    }
    
    /** Returns a boolean value indicating whether the length
     * of this bond is less then the maximum length allowed for
     * it's type.
     */    
    public boolean isValid() {
        return ( getLength() < maxLength );
    }
    
    /** Returns true if the input Object is a Bond and has the
     * same originating and ending Atoms.
     */    
    public boolean equals( Object o ) {
        if ( ! ( o instanceof Bond ) ) { return false; }
        Bond b = (Bond)o;
        if ( b.start.equals(start) || b.end.equals(start) ){
            if ( b.end.equals(end) || b.start.equals(end) ) {
                if ( b.bondType == bondType ) {
                    return true;
                }
            }
        }
        return false;
        
    }
    
    /** Returns a hash of the input atom id's.
     */    
    public int hashCode() {
        return start.getID() + end.getID();
    }
    
    /** Returns a String uniquely identifying this bond within a
     * particular file.  This is done by concatenating the id's
     * of the individual component Atoms.
     */    
    public String getBondTypeID(){
        String s = ((AtomAttributes)start.getAttributes()).atomicSymbol;
        String e = ((AtomAttributes)end.getAttributes()).atomicSymbol;
        if ( s == null ) { s = ((AtomAttributes)start.getAttributes()).atomName; }
        if ( e == null ) { e = ((AtomAttributes)end.getAttributes()).atomName; }
        String outString = "";
        if ( s.compareTo(e) <= 0){
            outString = s + "-" + e;
        }
        else {
            outString = e + "-" + s;
        }
        if ( bondType == BOND_TYPE_HYDROGEN ) {
            outString += "( h-bond )";
        }
        return outString;
    }
    
    /** Returns a string representing this bond.  Of the form:
     *
     * " [atom1type](id) - [atom2type](id) "
     */    
    public String toString() {
        return start.toString() + "<->" + end.toString();
    }
    
    /** Compares the two input bonds for equality.  The test is
     * a string equality based on the return of getBondTypeID().
     *
     */    
    public int compare(Object obj, Object obj1) {
        Bond a = (Bond)obj;
        Bond b = (Bond)obj1;
        if ( !a.getBondTypeID().equals( b.getBondTypeID() ) ){
            return (a.getBondTypeID().compareTo( b.getBondTypeID()));
        }
        if ( a.start.getID() < b.start.getID() ) { return -1; }
        return 1;
        
    }    
    
    /** Sets the 'fudge factor' used to extend bonding radaii a certain
     * constant factor beyond the lengths given by the atom table.
     */    
    public static final void setBondingFudgeFactor( float f ) { fudge = f; }
    /** Returns the current bonding 'fudge factor'.
     */    
    public static final float getBondingFudgeFactor( ) { return fudge; }
    
    /** Sets the 'fudge factor' used when checking hydrogen bonds.
     */    
    public static final void setHydrogenAdjustmentFactor( float f ) { hydFudge = f; }
    /** Returns the fudge factor used for hydrogen bonds.
     */    
    public static final float getHydrogenAdjustmentFactor( ) { return hydFudge; }
    
    /** Returns the distance between the two input atoms.
     */    
    public static final double bondLength( Atom typeA, Atom typeB ) {
        return new Bond( typeA, typeB, BOND_TYPE_NORMAL ).maxLength * fudge;
    }
}
