package molmaster.gui;

import molmaster.*;
import molmaster.util.*;
import molmaster.model.*;

import javax.vecmath.*;
import javax.media.j3d.*;

import com.sun.j3d.utils.geometry.*;

/** Renders a single atom object using its
 *  AtomAttributes.
 *
 * @author  Russell Power
 */
public class AtomView extends Shape3D
implements WatchedObject, MolView, ModelListener, java.io.Serializable, Listener {
    
    protected Atom myAtom;
    
    /** Used to send messages to the model when events occur. */
    protected DefaultWatchedObject watched = new DefaultWatchedObject();
    static protected java.util.Map atomToView = new java.util.WeakHashMap();
    static protected DataStore info = new DataStore("appSettings");
    
    private static int highResQuality = 18;
    private static int lowResQuality = 4;
    
    private static float selectScale = 1.25f;
    
    private double currentSize = 1.0;
    private double scaleFactor = 1.0;
    
    static {
        Integer t = (Integer)info.get( "AtomView.highRes" );
        if ( t != null ) { highResQuality = t.intValue(); }
        t = (Integer)info.get( "AtomView.lowRes" );
        if ( t != null ) { lowResQuality = t.intValue(); }
    }
    
    static protected Geometry highRes = new Sphere( 1, Sphere.GENERATE_NORMALS | Sphere.GEOMETRY_NOT_SHARED, highResQuality).getShape().getGeometry();
    static protected Geometry lowRes = new Sphere( 1, Sphere.GENERATE_NORMALS | Sphere.GEOMETRY_NOT_SHARED, lowResQuality).getShape().getGeometry();
    static protected Geometry currentRes = lowRes;
    static protected Transform3D trans = new Transform3D();
    
    static {
        highRes.setCapability( Geometry.ALLOW_INTERSECT );
        lowRes.setCapability( Geometry.ALLOW_INTERSECT );
    }
    
    /** Returns the AtomView that represents the given input Atom.  If no view is
     * available ( the Atom has not been added to the scene ) the return is
     * <code>null</code>
     */
    static public AtomView atomToView( Atom a ) {
        return (AtomView)atomToView.get( a );
    }
    
    protected TransformGroup ownerGroup;
    protected Appearance defaultAppearance;
    
    /** For whatever reason, the geometry created by the Sphere object
     *  is a bit smaller than the size picked up by the object picking
     *  layer.  In order to allow for bounds based picking, we used this
     *  constant.
     */
    public static final float boundsScale = 1.0f;
    
    public AtomView() {}
    
    /** Creates a new atomview representing the given atom, rooted below
     *  the given transform group.  */
    public AtomView(Atom atom, TransformGroup root) {
        
        myAtom = atom;
        myAtom.addListener( this );
        ownerGroup = root;
        
        setCapability( ALLOW_APPEARANCE_WRITE );
        setCapability( ALLOW_APPEARANCE_READ );
        setCapability( ALLOW_GEOMETRY_WRITE );
        setCapability( ALLOW_GEOMETRY_READ );
        setGeometry( currentRes );
        
        root.addChild( this );
        root.getTransform( trans );
        trans.setScale( myAtom.getAttributes().radius );
        currentSize = myAtom.getAttributes().radius;
        root.setTransform( trans );
        
        setDefaultAppearance( AtomTable.getActive().getAppearance( atom ) );
        
        setAppearance( defaultAppearance );
        Point3d center = new Point3d( atom.getLocation() );
        setBounds( new BoundingSphere( center, boundsScale) );
        atomToView.put( atom, this );
    }
    
    /** Adds the given listener to the notification list for this AtomView.  When
     * changes occur to this AtomView, the specified listener(s) will receive
     * confirmation.
     */
    public void addListener(Listener l) {
        watched.addListener(l);
    }
    
    /** Removes the given listener from this AtomView's list.
     */
    public void removeListener(Listener l) {
        watched.removeListener(l);
    }
    
    protected boolean selected = false;
    
    /** Change this object so it appears to
     *  be selected.     */
    public void select() {
        if ( !vis  ) { return; }
        if ( selected ) { return; }
        selected = true;
        java.util.Iterator i = myAtom.getBonds().iterator();
        
        Log.info( "Changing appearance to selected." );
        setAppearance( AtomTable.selectedAppearance );
        
        Transform3D trans = new Transform3D();
        ownerGroup.getTransform( trans );
        trans.setScale( currentSize * scaleFactor * selectScale );
        ownerGroup.setTransform( trans );
        
        watched.signal(this, WatchedObject.SELECTED, null );
    }
    
    /** Change this object so it appears to be
     *  unselected.  */
    public void unselect() {
        if ( !selected ) { return; }
        selected = false;
        java.util.Iterator i = myAtom.getBonds().iterator();
        
        Log.info( "Changing appearance to default." );
        setAppearance( defaultAppearance );
        Transform3D trans = new Transform3D();
        ownerGroup.getTransform( trans );

        trans.setScale( currentSize * scaleFactor );
        ownerGroup.setTransform( trans );
        watched.signal( this, WatchedObject.UNSELECTED, null );
    }
    
    /** Change the rendering of this object to
     *  the type specified, if possible.     */
    public void setViewType(int viewType) {
    }
    
    /** Checks to see if the atom removed from the model belonged to
     *  this view.  If so, we remove ourself.   */
    public void modelUpdated(MolModel scene, int opType, Object opValue) {
        if ( (opType & ModelListener.ATOM_REMOVED) != 0 ) {
            if ( scene.contains(myAtom) ) {
                //  TODO - is this a remove from scene?
                Group g = (Group)getParent();
                g.removeChild(0);
            }
        }
    }
    
    /** Returns the java3d node that this AtomView is positioned under in the scenegraph.
     */
    public javax.media.j3d.Group getOwner() { return ownerGroup; }
    /** Returns the Atom that this AtomView represents.
     */
    public Atom getAtom() { return myAtom; }
    /** Sets the Atom that this AtomView represents.  The position of this AtomView is
     * updated to reflect the target Atom.
     */
    public void setAtom( Atom a ) { myAtom = a; }
    /** Returns a String representation of this view.
     */
    public String toString() { return myAtom.toString(); }
    /** Sets the appearance that this AtomView object reverts to when unselected or
     * reset.
     */
    public void setDefaultAppearance( Appearance a ) {
        defaultAppearance = a;
        setAppearance(a);
        setVisible( true );
    }
    
    /**
     */
    public void setVisible( boolean b ) {
        vis = b;
        if ( !b ) {
            removeAllGeometries();
        }
        else {
            setGeometry( currentRes );
        }
    }
    
    public boolean isVisible() { return vis; }
    
    /** Returns the object that this view is representing.   */
    public molmaster.MolObject getModelObject() { return myAtom; }
    
    /** Allows the user to change the basic geometry resolution used
     *  by the program.
     */
    public static void setLowResolutionQuality( int val ) { lowResQuality = val; updateGeom(); }
    /** Returns the quality factor used to represent low resolution AtomView rendering.
     */
    public static int getLowResolutionQuality() { return lowResQuality; }
    /** Sets the quality factor used to represent high resolution AtomView.
     */
    public static void setHighResolutionQuality(int val) { highResQuality = val; updateGeom(); }
    /** Returns the quality factor used to represent high resolution AtomView.
     */
    public static int getHighResolutionQuality() { return highResQuality; }
    
    protected static void updateGeom() {
        info.put( "AtomView.lowRes", new Integer(lowResQuality) );
        info.put( "AtomView.highRes", new Integer(highResQuality) );
        info.flush();
        
        highRes = new Sphere( 1, Sphere.GENERATE_NORMALS | Sphere.GEOMETRY_NOT_SHARED, highResQuality).getShape().getGeometry();
        lowRes = new Sphere( 1, Sphere.GENERATE_NORMALS | Sphere.GEOMETRY_NOT_SHARED, lowResQuality).getShape().getGeometry();
        
        highRes.setCapability( Geometry.ALLOW_INTERSECT );
        lowRes.setCapability( Geometry.ALLOW_INTERSECT );
        
        currentRes = lowRes;
        java.util.Collection c = atomToView.values();
        java.util.Iterator i = c.iterator();
        while ( i.hasNext() ) {
            AtomView a = (AtomView)i.next();
            a.setGeometry( currentRes );
        }
    }
    /** Specify the resolution to display.   */
    public static void changeResolution(int resValue) {
        if ( resValue == RESOLUTION_HIGH ){
            if ( currentRes == highRes ) { return; }
            currentRes = highRes;
        }
        else {
            if ( currentRes == lowRes ) { return; }
            currentRes = lowRes;
        }
        
        java.util.Collection c = atomToView.values();
        java.util.Iterator i = c.iterator();
        //if ( c.size() == 0 ) { return; }
        while ( i.hasNext() ) {
            AtomView a = (AtomView)i.next();
            a.setGeometry( currentRes );
        }
    }
    
    static protected Matrix3d rot = new Matrix3d();
    static protected Vector3d translate = new Vector3d();
    
    /** Only signaled by child atom.    */
    public void wakeUp(Object signaler, int signalType, Object value) {
        if ( (signalType & Atom.MOVED) != 0 ) {
            Log.info( "Atom event fired." );
            Vector3d newLoc = (Vector3d)value;
            ownerGroup.getTransform( trans );
            trans.get( translate );
            Log.info( translate + " to " + newLoc );
            trans.getRotationScale(rot);
            trans.setTranslation( newLoc );
            trans.setRotationScale( rot );
            ownerGroup.setTransform( trans );
            
        }
    }
    
    public boolean equals( Object o ) {
        if ( !(o instanceof AtomView) ) { return false; }
        return myAtom.equals( ((AtomView)o).myAtom );
    }
    
    boolean removed = false;
    public void removeFromScene() {
        removeAllGeometries();
        removed = true;
    }
    
    /** Sets the Geometry of this AtomView to the input.
     */
    public void setGeometry( Geometry g ) {
        if ( removed || !isVisible() ) { return; }
        removeAllGeometries();
        //g.setCapability( g.ALLOW_INTERSECT );
        super.addGeometry( g );
    }
    
    /** Sets the amount to scale this AtomView when rendering.  This should only
     *  be used on a file wide basis from the MoleculeFile call setScaleFactor().
     */
    public void setScaleFactor( float f ){
        Transform3D trans = new Transform3D();
        ownerGroup.getTransform( trans );
        scaleFactor = f;
        trans.setScale( f * currentSize );
        ownerGroup.setTransform( trans );
    }
    
    /** Returns the amount this AtomView is scaled by in the scene.
     */
    public float getScaleFactor() {
        return (float)scaleFactor;
    }
    
    public void setSize( double size ) {
        currentSize = size;
        Transform3D trans = new Transform3D();
        ownerGroup.getTransform( trans );
        trans.setScale( currentSize * scaleFactor );
        ownerGroup.setTransform( trans );
    }
    
    public double getSize() {
        return currentSize;
    }
    
    /** Returns the Appearance that this AtomView reverts to when unselected or reset.
     */
    public Appearance getDefaultAppearance() { return defaultAppearance; }
    
    /** Returns the appearance of this atom to the default
     *  found in the AtomTable, and unselects this atom.
     */
    public void reset() { 
        unselect();
        setDefaultAppearance( AtomTable.getActive().getAppearance( myAtom ) );
        setSize( myAtom.getAttributes().radius );
    }
        
    
    protected boolean vis = true;
}