package molmaster.gui;

import molmaster.*;
import molmaster.model.*;

import java.util.*;

import javax.vecmath.*;
import javax.media.j3d.*;
import com.sun.j3d.utils.geometry.*;

/** Provides a visual representation of a
 *  single bond.
 *
 * @author  rjpower
 */
public class BondView extends Shape3D
implements WatchedObject, MolView, Comparator {
    
    protected static GeomHolder highRes;
    protected static GeomHolder lowRes;
    protected static GeomHolder currentRes;
    
    public static final float BOND_RADIUS = 0.2f;
    
    protected static Map bondToView = new WeakHashMap();
    /** Given an input Bond, returns the BondView object representing the input.  If no
     * BondView is found for the given Bond, or the input is not a Bond object, the
     * return value is null.
     */    
    public static BondView bondToView(Object b) { return (BondView)bondToView.get(b); }
    
    protected static Appearance defaultBondAppearance = new Appearance();
    protected static Appearance selectedAppearance = new Appearance();
    protected static Appearance hydrogenAppearance = new Appearance();
    protected static Transform3D trans = new Transform3D();
    
    protected DefaultWatchedObject watched = new DefaultWatchedObject();
    protected Bond myBond;
    
    private double currentSize;
    private double scaleFactor;
    
    static {
        Material m = new Material();
        m.setEmissiveColor( 0.4f, 0.4f, 0.4f );
        m.setDiffuseColor( 0.4f, 0.4f, 0.4f );
        m.setShininess( 10 );
        
        defaultBondAppearance.setMaterial( m );
        m = new Material();
        m.setEmissiveColor( 0.4f, 0.4f, 0.9f );
        selectedAppearance.setMaterial( m );
        
        TextureAttributes ta = new TextureAttributes();
        ta.setTextureMode( TextureAttributes.MODULATE );
        ta.setTextureBlendColor( 1.0f, 1.0f, 1.0f, 0.0f );
        
        TransparencyAttributes tAtt = new
        TransparencyAttributes(TransparencyAttributes.BLENDED, 0.0f);
        hydrogenAppearance.setTransparencyAttributes(tAtt);
        
        hydrogenAppearance.setTexture( TextureCache.getTexture("hyd.png") );
        hydrogenAppearance.setTextureAttributes( ta );
        
        highRes = new GeomHolder();
        Cylinder c = new Cylinder( 1.0f, 1.0f, Cylinder.GENERATE_NORMALS | Cylinder.GENERATE_TEXTURE_COORDS, 10, 2, defaultBondAppearance );
        highRes.top = c.getShape( Cylinder.TOP ).getGeometry();
        highRes.middle = c.getShape( Cylinder.BODY ).getGeometry();
        highRes.bottom = c.getShape( Cylinder.BOTTOM).getGeometry();
        
        lowRes = new GeomHolder();
        c = new Cylinder( 1.0f, 1.0f, Cylinder.GENERATE_NORMALS | Cylinder.GENERATE_TEXTURE_COORDS, 3, 2, defaultBondAppearance );
        lowRes.top = c.getShape( Cylinder.TOP ).getGeometry();
        lowRes.middle = c.getShape( Cylinder.BODY ).getGeometry();
        lowRes.bottom = c.getShape( Cylinder.BOTTOM).getGeometry();
        
        setCapabilities( lowRes );
        setCapabilities( highRes );
        updateTexCoords( lowRes );
        updateTexCoords( highRes );
        
        currentRes = lowRes;
    }
    
    private static void setCapabilities( GeomHolder h ) {
        h.bottom.setCapability( Geometry.ALLOW_INTERSECT );
        h.middle.setCapability( Geometry.ALLOW_INTERSECT );
        h.top.setCapability( Geometry.ALLOW_INTERSECT );
    }
    
    static protected void updateTexCoords( GeomHolder h ) {
        GeometryArray g = (GeometryArray)h.middle;
        float[] tex = new float[g.getVertexCount() * 2];
        g.getTextureCoordinates(0, 0, tex);
        for ( int i = 0; i < tex.length; i++ ){
            tex[i] *= 10;
        }
        g.setTextureCoordinates( 0, 0, tex );
    }
    
    protected TransformGroup ownerGroup;
    protected Appearance myAppearance;
    protected Appearance defaultAppearance;
    
    /** Creates a new instance of BondView */
    public BondView() {
        super();
    }
    
    public BondView( Bond b, TransformGroup owner ){
        this( b.getLength(), defaultBondAppearance, b, owner );
    }
    
    public BondView( double length, Appearance ap, Bond b, TransformGroup owner ) {
        myBond = b;
        ownerGroup = owner;
        defaultAppearance = ap;
        currentSize = 1;
        scaleFactor = 1;
        init();
    }
    
    public void init() {
        bondToView.put( myBond, this );
        setPickable(true);
        
        /** We can't call getParent() when the scene graph
         *  is live... so we store it here...
         * Shape3D s = super.getShape( BODY );
         * setupShape( s, BODY );
         *
         * s = super.getShape( TOP );
         * setupShape( s, TOP );
         *
         * s = super.getShape( BOTTOM );
         * setupShape( s, BOTTOM );
         */
        
        //setUserData( this );
        
        addGeometry( currentRes.top );
        addGeometry( currentRes.middle );
        addGeometry( currentRes.bottom );
        
        setCapability( ALLOW_GEOMETRY_READ );
        setCapability( ALLOW_GEOMETRY_WRITE );
        setCapability( ALLOW_APPEARANCE_READ );
        setCapability( ALLOW_APPEARANCE_WRITE );
        
        if ( myBond.getType() == Bond.BOND_TYPE_HYDROGEN ) {
            myAppearance = hydrogenAppearance;
        }
        else {
            myAppearance = defaultAppearance;
        }
        setAppearance( myAppearance );
        
        ownerGroup.getTransform( trans );
        trans.setScale( new Vector3d( BOND_RADIUS, myBond.getLength(), BOND_RADIUS ) );
        ownerGroup.setTransform( trans );
        
        ownerGroup.addChild( this );
    }

    /** Returns the Java3D group that this Bond is positioned beneath.
     */    
    public Group getOwner() { return ownerGroup; }
    
    /** Adds a listener for BondView related events.
     */    
    public void addListener(Listener l) {
        watched.addListener(l);
    }
    
    public void removeListener(Listener l) {
        watched.removeListener(l);
    }
    
    /** Change this object so it appears to
     *  be selected.     */
    public void select() {
        setAppearance( selectedAppearance );
    }
    
    /** Change the rendering of this object to
     *  the type specified, if possible.     */
    public void setViewType(int viewType) {
    }
    
    /** Change this object so it appears to be
     *  unselected.  */
    public void unselect() {
        setAppearance( myAppearance );
    }
    
    /** Removes the given listener from the event notification list.
     */    
    public Bond getBond() {return myBond;}
    
    /** Sets the Bond that this BondView represents.
     */    
    public void setBond( Bond b ) {myBond = b;}
    
    /** Forces an update of this BondView object.
     */    
    public void refresh() {
        Appearance old = myAppearance;
        if ( myBond.getType() == Bond.BOND_TYPE_HYDROGEN ) {
            myAppearance = hydrogenAppearance;
        }
        else {
            myAppearance = defaultAppearance;
        }
        if( !( old == myAppearance ) ) {
            setAppearance( myAppearance );
        }
        
        //  update transform...
        Atom st = myBond.getStart();
        Atom end = myBond.getEnd();
        
        Vector3d a = st.getLocation(); Vector3d b = end.getLocation();
        Transform3D t3d = molmaster.gui.input.AttachPoints.getTransform( a, b);
        t3d.setScale( new Vector3d( BOND_RADIUS, myBond.getLength(), BOND_RADIUS ) );
        ownerGroup.setTransform( t3d );
    }
    
    public void setVisible( boolean b ) {
        if ( !b ) {
            removeAllGeometries();
        }
        else {
            addGeometry( currentRes.top );
            addGeometry( currentRes.middle );
            addGeometry( currentRes.bottom );
        }
        vis = b;
    }
    
    public boolean isVisible() { return vis; }
    public String toString() { return myBond.toString(); }
    
    /** Returns the object that this view is representing.   */
    public molmaster.MolObject getModelObject() { return myBond; }
    
    /** Sets the default appearance this BondView reverts to if unselected or reset.
     */    
    public void setDefaultAppearance(javax.media.j3d.Appearance a) {
        defaultAppearance = a;
        myAppearance = a;
        setAppearance( a );
    }
    
    /** Sets the color used by this BondView.
     */    
    public void setColor( Color3f c ) {
        Material m = new Material();
        m.setShininess( 10 );
        m.setAmbientColor( new Color3f( 0.5f, 0.5f, 0.5f ) );
        m.setEmissiveColor ( c );
        m.setSpecularColor ( new Color3f( 0.8f, 0.8f, 0.8f ) );
        m.setDiffuseColor ( c );
        myAppearance = new Appearance();
        myAppearance.setMaterial( m );
        setAppearance( myAppearance );
    }
    
    public int compare(Object a, Object b) {
        return compare.compare( ((BondView)a).getBond(), ((BondView)b).getBond());
    }
    
    /** Specify the resolution to display.   */
    public static void setResolution(int resValue) {
        if ( resValue == RESOLUTION_HIGH ){
            if ( currentRes == highRes ) { return; }
            currentRes = highRes;
        }
        if ( resValue == RESOLUTION_LOW ){
            if ( currentRes == lowRes ) { return; }
            currentRes = lowRes;
        }
        Collection c = bondToView.values();
        if( c.size() == 0 ) { return; }
        Iterator i = c.iterator();
        while ( i.hasNext() ){
            BondView b = (BondView)i.next();
            b.removeAllGeometries();
            b.addGeometry( currentRes.top );
            b.addGeometry( currentRes.middle );
            b.addGeometry( currentRes.bottom );
            /*
            b.getShape( TOP ).setGeometry( currentRes.top );
            b.getShape( BODY ).setGeometry( currentRes.middle );
            b.getShape( BOTTOM ).setGeometry( currentRes.bottom );
             */
        }
    }
    
    public boolean equals( Object o ) {
        if ( !(o instanceof BondView) ) { return false; }
        return myBond.equals( ((BondView)o).myBond );
    }
    
    /** Removes this BondView from the scene.
     */    
    public void removeFromScene() {
        removeAllGeometries();
        removed = true;
    }
      
    public void addGeometry( Geometry g ) {
        if ( !isVisible() || removed ) { return; }
        super.addGeometry( g );
    }
    
    /** Sets the amount to scale the BondView by.  This scale factor is only applied to
     * the width of the view - the length remains the same.
     */    
    public void setScaleFactor( float f ) {
        scaleFactor = f;
        Transform3D trans = new Transform3D();
        ownerGroup.getTransform( trans );
        trans.setScale( new Vector3d(BOND_RADIUS * f * currentSize, myBond.getLength(), BOND_RADIUS * f * currentSize) );
        ownerGroup.setTransform( trans );
    }
    
    public int hashCode() { return myBond.hashCode(); }
    
    /** Sets the size of this object to the given value.  For atoms
     *  this results in an increase of the radius to the given value,
     *  for bonds, the cylinder radius is increased to the given value.
     *
     */
    public void setSize(double size) {
        currentSize = size;
        setScaleFactor( (float)scaleFactor );
    }
    
    public double getSize() {return currentSize;}
    protected boolean vis = true;
    protected boolean removed = false;
    static protected Bond compare = new Bond();
}


class GeomHolder {
    Geometry top;
    Geometry middle;
    Geometry bottom;
}