package molmaster.gui;

import java.util.*;

import javax.vecmath.*;
import javax.media.j3d.*;

import molmaster.util.*;

/** Provides an interface to examine and change the
 *  properties of lights within molmaster.
 *
 * @author  Russell Power
 */
public class MolLight implements java.io.Serializable {
    
    protected String lightType = AMBIENT;
    protected float lightIntensity = 1.0f;
    protected float extent = Float.POSITIVE_INFINITY;
    
    protected Vector3d lightPosition = new Vector3d();
    protected Vector3d lightDirection = new Vector3d();
    protected Color3f lightColor = new Color3f(0.6f, 0.6f, 0.6f);
    
    protected String name = "";
    
    protected transient BranchGroup attachGroup = new BranchGroup();
    protected transient Light delegate;
    
    public static ArrayList lightList;
    
    public static void initLights() {
        DataStore ds = new DataStore( "lightSettings" );
        lightList = (ArrayList)ds.get( "MolLight.LightList" );
        
        if ( lightList == null ) {
            lightList = new ArrayList();
            MolLight ambient = new MolLight( MolLight.AMBIENT, new Color3f( 0.6f, 0.6f, 0.6f ) );
            ambient.setName( "Default Ambient light" );
            ambient.addToScene();
            
            MolLight dir = new MolLight( MolLight.DIRECTIONAL, new Color3f( 0.6f, 0.6f, 0.6f ) );
            dir.setDirection( new Vector3d( 0, 0, -1 ) );
            dir.setName( "Default Directional light" );
            dir.addToScene();
            
            MolLight dir2 = new MolLight( MolLight.DIRECTIONAL, new Color3f( 0.6f, 0.6f, 0.6f ) );
            dir2.setDirection( new Vector3d( 0, -1, 1 ) );
            dir2.setName( "Default Directional Light" );
            dir2.addToScene();
        }else {
            Iterator i = lightList.iterator();
            while ( i.hasNext() ) {
                ((MolLight)i.next()).addToScene();
            }
        }
    }
    
    /** Represents the java3d/opengl ambient light.
     */
    public static final String AMBIENT =  "AMBIENT";
    /** Represents the java3d/opengl directional light.
     */
    public static final String DIRECTIONAL =   "DIRECTIONAL";
    /** Represents the java3d/opengl point light.
     */
    public static final String POINT =         "POINT";
    /** Represents the java3d/opengl spotlight.
     */
    public static final String SPOTLIGHT =     "SPOTLIGHT";
    
    static private List lightTypes = Arrays.asList(
    new Object[] { AMBIENT, DIRECTIONAL, POINT, SPOTLIGHT } );
    
    /** Creates a new instance of MolLight */
    public MolLight() {
        lightType = AMBIENT;
        lightIntensity = 1.0f;
        delegate = null;
        name = "Light " + lightList.size();
        addToScene();
    }
    
    /** Creates a new MolLight of the given type and color.
     */
    public MolLight( String type, Tuple3f color ) {
        this();
        try {
            setLightType( type );
        } catch(Exception e){
            type = AMBIENT;
        }
        lightColor.set( color );
        addToScene();
    }
    
    /** Sets this light to be the given type.
     */
    public void setLightType( String type ) throws IllegalArgumentException {
        type = type.toUpperCase();
        if ( !lightTypes.contains( type ) ) {
            throw new IllegalArgumentException("Invalid light type specified.");
        }
        lightType = (String)lightTypes.get( lightTypes.indexOf( type ) );
    }
    
    /** Returns the current type of this light.
     */
    public String getLightType() { return lightType; }
    
    /** Sets the position of this light in space.  Ignored for AMBIENT and DIRECTIONAL.
     */
    public void setPosition(Vector3d pos){ lightPosition = pos; }
    /** Returns the position of this light in space.
     */
    public Vector3d getPosition() { return lightPosition; }
    
    /** Sets the direction of this light.  Ignored for POINTLIGHT and AMBIENT.
     */
    public void setDirection(Vector3d dir){ lightDirection = dir; }
    public Vector3d getDirection() { return lightDirection; }
    
    public Color3f getColor() { return lightColor; }
    public void setColor( Tuple3f c ) { lightColor = new Color3f(c); }
    
    /** Returns the name of this light.
     */
    public String getName() { return name; }
    public void setName( String s ) { name = s; }
    
    /** Returns the intensity of this light.
     */
    public float getIntensity() { return lightIntensity; }
    /** Set the intensity value for this light.  This value is multiplied by the color
     * to obtain the final color values to use for the light.
     */
    public void setIntensity( float f ) { lightIntensity = f; }
    /** Sets the intensity value for this light.  This value is multiplied by the color
     * to obtain the final color values to use for the light.
     */
    public void setIntensity( Float f ) { lightIntensity = f.floatValue(); }
    
    /** Returns the radius of the sphere in which this light is applicable.
     */
    public float getExtent() { return extent; }
    /** Sets the application radius of this light.
     */
    public void setExtent( float f ) { extent = f; }
    public void setExtent( Float f ) { extent = f.floatValue(); }
    
    public String toString() { return name; }
    
    public void refresh() {
        removeFromScene();
        addToScene();
    }
    
    /** Removes this light from the scene.
     */
    public void removeFromScene() {
        attachGroup.detach();
        lightList.remove( this );
    }
    
    protected Light getLight() {
        if ( lightType.equals(AMBIENT) ) {return new AmbientLight();}
        if ( lightType.equals(POINT) ) {
            PointLight p = new PointLight();
            p.setPosition( new Point3f( getPosition() ) );
            p.setAttenuation( getIntensity(), getIntensity(), getIntensity() );
            return p;
        }
        if ( lightType.equals(DIRECTIONAL) ) {
            DirectionalLight d = new DirectionalLight();
            d.setDirection( new Vector3f( getDirection() ) );
            return d;
        }
        if ( lightType.equals(SPOTLIGHT) ) {
            SpotLight s = new SpotLight();
            s.setPosition( new Point3f(getPosition()) );
            s.setAttenuation( getIntensity(), getIntensity(), getIntensity() );
            s.setDirection( new Vector3f(getDirection()) );
            return s;
        }
        return null;
    }
    
    public void addToScene() {
        if ( delegate != null && delegate.isLive() ){
            attachGroup.detach();
            attachGroup.removeAllChildren();
        }
        
        if ( !lightList.contains( this ) ) {
            lightList.add( this );
        }
        
        delegate = getLight();
        delegate.setInfluencingBounds( new BoundingSphere( origin, extent ) );
        Color3f c = new Color3f( getColor() );
        c.scale( lightIntensity );
        c.clampMax( 1.0f );
        
        delegate.setColor(c);
        
        attachGroup = new BranchGroup();
        
        attachGroup.setCapability( BranchGroup.ALLOW_CHILDREN_WRITE );
        attachGroup.setCapability( BranchGroup.ALLOW_DETACH );
        attachGroup.clearCapabilityIsFrequent( BranchGroup.ALLOW_DETACH );
        attachGroup.clearCapabilityIsFrequent( BranchGroup.ALLOW_CHILDREN_WRITE );
        
        attachGroup.addChild( delegate );
        MainFrame.uni.addBranchGraph( attachGroup );
        
    }
    
    static protected Point3d origin = new Point3d( 0,0,0 );
}
