package molmaster.gui;

import java.util.*;
import javax.media.j3d.*;
import javax.vecmath.*;

import molmaster.*;
import molmaster.model.*;
import molmaster.gui.input.*;

/** An instance of this class handles basic
 *  functions related to the current scene -
 *  it provides for adding and removing objects,
 *  as well as some basic utility functions.
 *
 * @author  Russell
 */
public class SceneControl implements ModelListener  {
    
    MolUniverse uni;
    MolModel model;
    public Group rootGroup;
    int currentAtoms = 0;
    
    public Map fileMap = new HashMap();
    public Map atomMap = new WeakHashMap();
    
    /** The current scene.
     */    
    static public SceneControl activeScene;
    /** Creates a new instance of SceneControl */
    public SceneControl(MolModel model, Group root, MolUniverse universe ) {
        this.model = model;
        this.uni = universe;
        this.rootGroup = root;
        activeScene = this;
        model.addListener( this );
    }
    
    /** Returns an AtomView attached to a TransformGroup.  The
     *  returned TransformGroup has been translated (relative to the
     *  origin) to the position of the given Atom in space. */
    protected void addAtomView(Atom a, Group root) {
        
        Transform3D trans = new Transform3D();
        trans.setTranslation( new Vector3d(a.getLocation()) );
        
        currentAtoms++;
        
        //BranchGroup subGroup = new BranchGroup();
        //subGroup.setCapability( subGroup.ALLOW_DETACH );
        
        TransformGroup tg = new TransformGroup();
        tg.setTransform( trans );
        
        tg.setCapability(TransformGroup.ALLOW_TRANSFORM_WRITE);
        tg.setCapability(TransformGroup.ALLOW_TRANSFORM_READ);
        //tg.setCapability(TransformGroup.ALLOW_CHILDREN_EXTEND);
        //tg.setCapability(TransformGroup.ALLOW_CHILDREN_WRITE);
        tg.clearCapabilityIsFrequent( TransformGroup.ALLOW_TRANSFORM_WRITE);
        tg.clearCapabilityIsFrequent( TransformGroup.ALLOW_TRANSFORM_READ );
        //tg.clearCapabilityIsFrequent(TransformGroup.ALLOW_CHILDREN_EXTEND );
        //tg.clearCapabilityIsFrequent(TransformGroup.ALLOW_CHILDREN_WRITE);
        AtomView av = new AtomView( a, tg );
        
        //subGroup.addChild( tg );
        //root.addChild( subGroup );
        root.addChild( tg );
        
        //a.addListener( this );
        //atomMap.put( av, subGroup );
    }
    
    /** Processes model events.  SceneControl watches for various
     *  events that from the model that are relevant to updating the
     *  scene.  It then acts to make the scene conform with the model,
     *  if possible.
     */
    public void modelUpdated(MolModel scene, int opType, Object opValue) {
        if ( (opType & ATOM_ADDED) != 0 ) {
            Atom a = (Atom)opValue;
            BranchGroup childGroup = new BranchGroup();
            //childGroup.addChild(getTransform(a));
            addAtomView( a, childGroup );
            childGroup.compile();
            rootGroup.addChild(childGroup);
        }
        
        else if( (opType & FILE_ADDED) != 0 ) {
            
            MoleculeFile mf = (MoleculeFile)opValue;
            Collection c = mf.getAtoms();
            BranchGroup g = (BranchGroup)fileMap.get(opValue);
            
            if ( g != null ) {
                if ( !g.isLive() ){
                    rootGroup.addChild( g );
                }else {
                    Log.writeLog( "Attempted to add group twice!" );
                }
                return;
            }
            
            Iterator i = c.iterator();
            BranchGroup fileGroup = new BranchGroup();
            
            fileGroup.setCapability( BranchGroup.ALLOW_DETACH );
            fileGroup.setCapability( BranchGroup.ALLOW_CHILDREN_WRITE );
            fileGroup.setCapability( BranchGroup.ALLOW_CHILDREN_EXTEND );
            
            fileGroup.clearCapabilityIsFrequent( BranchGroup.ALLOW_DETACH );
            fileGroup.clearCapabilityIsFrequent( BranchGroup.ALLOW_CHILDREN_WRITE );
            fileGroup.clearCapabilityIsFrequent( BranchGroup.ALLOW_CHILDREN_EXTEND );
            
            //  we store the file's branch group so that we can remove
            //  it all at once later...
            fileMap.put( opValue, fileGroup );
            
            while ( i.hasNext() ) {
                Atom a = (Atom)i.next();
                addAtomView( a, fileGroup );
            }
            
            fileGroup.addChild( mf.getAdditionalData() );
            
            fileGroup.compile();
            rootGroup.addChild( fileGroup );
        }
        
        else if ((opType & FILE_REMOVED) != 0 ) {
            BranchGroup g = (BranchGroup)fileMap.get(opValue);
            if ( g != null ) {
                g.detach();
            }
        }
        else if ((opType & BOND_GROUP_ADDED) != 0 ) {
            BranchGroup bondGroup = new BranchGroup();
            bondGroup.setCapability( BranchGroup.ALLOW_DETACH );
            bondGroup.setCapability( BranchGroup.ALLOW_CHILDREN_WRITE );
            bondGroup.clearCapabilityIsFrequent( BranchGroup.ALLOW_DETACH );
            bondGroup.clearCapabilityIsFrequent( BranchGroup.ALLOW_CHILDREN_WRITE );
            
            MoleculeFile owner = null;
            Collection l = (Collection)opValue;
            if ( l.size() < 1 ) { return; }
            Iterator i = l.iterator();
            while (i.hasNext()){
                Bond b = (Bond)i.next();
                if ( owner == null ) {
                    owner = b.getOwner();
                }
                connectBond(b, bondGroup );
            }
            
            bondGroup.compile();
            BranchGroup g = (BranchGroup)fileMap.get( owner );
            g.addChild( bondGroup );
        }
        
        else if ((opType & BOND_ADDED) != 0) {
            Bond b = (Bond)opValue;
            Log.writeLog( "Adding " + b + " to scene." );
            connectBond(b, (BranchGroup)fileMap.get(b.getOwner() ) );
        }
        
        else if((opType & BOND_REMOVED) != 0) {
            /*
            BranchGroup g = (BranchGroup)AttachPoints.bondMap.get( opValue );
            AttachPoints.bondMap.remove( opValue );
            if ( g != null )
                g.detach();
            else
             
             */
            BondView bv = BondView.bondToView( opValue );
            if ( bv != null ) {
                bv.removeFromScene();
                BranchGroup g = (BranchGroup)AttachPoints.bondMap.get( bv );
                AttachPoints.bondMap.remove( opValue );
                if ( g != null )
                    g.detach();
            }
            else {
                Log.writeLog( "huh?" );
            }
        }
        else if((opType & ATOM_REMOVED) != 0) {
            AtomView av = AtomView.atomToView((Atom)opValue);
            av.removeFromScene();
            //BranchGroup g = (BranchGroup)atomMap.get( av );
            //g.detach();
        }
    }
    
    protected void connectBond( Bond b, Group root ) {
        Vector3d start = b.getStart().getLocation();
        Vector3d end = b.getEnd().getLocation();
        
        AttachPoints.attach( root, b.getStart(), b.getEnd(), b );
    }
    
    
}