package molmaster.gui.editbar;

import molmaster.*;
import molmaster.gui.*;
import molmaster.gui.toolbar.*;
import molmaster.util.*;

import java.io.*;
import java.awt.event.*;
import java.util.*;

import javax.swing.*;

/** Interface to all plugins loaded by molmaster.  This class
 *  provides for dynamic loading and unloading of 'plugins' - java
 *  classes implementing the EditAction interface.  These classes
 *  once loaded are allowed to manipulate the molmaster environment.
 *
 * @author  Russell Power
 */
public class EditBar extends ToolBar implements ComponentListener {
    
    /** Creates a new instance of EditBar */
    public EditBar() {
        super( "Plugins" );
        theBar = this;
        loadHandlers();
        restoreState();
    }

    protected void loadClasses( String[] classes ) {
        for ( int i = 0; i < classes.length; i++ ){
            if ( !classes[i].endsWith( ".class" ) ){ continue; }
            
            String cName = classes[i].substring( 0, classes[i].length() - 6 );
            
            //  don't try and load ourself...
            if ( cName.equals("EditBar") ){ continue; }
            
            cName = "molmaster.gui.editbar." + cName;
            Log.info( "Trying to load " + cName );
            loadClass( cName );
        }
    }
    
    /** A list of plugins that are always in existence - they are located in the
     * molmaster.jar file.
     */
    protected String[] basePlugins = new String[] {
        "AngleCalculator",
        "AntialiasAction",
        "AxisDisplay",
        "BondDistance",
        "ClearScene",
        "analysis.CDAnalysis",
        "analysis.CNAnalysis",
        "analysis.RadialDensity",
        "analysis.RingAnalysis",
        "DihedralCalculator",
        "IsoSurface",
        "MultiSelect",
        "RecenterRotation",
        "ViewSelect"
    };
    
    
    protected void loadHandlers() {
        String[] classes = null;
        String[] tmpClass = new String[basePlugins.length];
        
        for ( int i = 0; i < basePlugins.length; i++ ) {
            String className = "molmaster.gui.editbar." + basePlugins[i];
            loadClass(className);
        }
        
        File f = new File("plugins");
        classes = f.list();
        if ( classes != null ) {
            loadClasses( classes );
        }
        
        addComponents();
    }
    
    private void addComponents() {
        currentButtons.clear();
        buttons = new ButtonGroup();
        HashMap groups = new HashMap();
        
        Collection c = activeHandlers.values();
        Iterator i = c.iterator();
        
        while ( i.hasNext( ) ){
            Iterator j = ((List)i.next()).iterator();
            while ( j.hasNext() ){
                EditAction e = (EditAction)j.next();
                AbstractButton b = null;
                if ( e instanceof CustomView ){
                    b = (AbstractButton)((CustomView)e).getButton();
                }
                else {
                    b = new SelectButton(e);
                    buttons.add( b );
                }
                currentButtons.add( b );
                
                JToolBar panel = (JToolBar)groups.get( e.getHandlerGroup() );
                if( panel == null ) {
                    panel = new JToolBar();
                    panel.setName( e.getHandlerGroup() );
                    add( panel );
                    groups.put( e.getHandlerGroup(), panel );
                }
                
                panel.add(b);
                
                if  ( e.getHandlerName().equals( "Object Select" ) ) {
                    defaultButton = b;
                    defaultHandler = e;
                }
            }
        }
        
        java.awt.Component[] child = this.getComponents();
        
        for ( int j = 0; j < child.length; j++ ){
            java.awt.Component[] d = ((JToolBar)child[j]).getComponents();
            for ( int k = 0; k < d.length; k++ ) {
                add( d[k] );
            }
            EditButton sep= new molmaster.gui.toolbar.Separator();
            sep.setName( "separator" );
            add ( sep );
            remove( child[j] );
        }
    }
    
    /** Attempts to find and load the given class from
     *  the hard disk.  At first we attempt to realize a class
     *  already in our namespace (loaded via the -classpath directive).
     *
     *  If this fails, we search the plugins directory for a 
     *  class matching the given name.
     */
    protected void loadClass( String cName ) {
        try {
            Class c = Class.forName( cName );
            EditAction test = (EditAction)c.newInstance();
            List l = (List)activeHandlers.get( test.getHandlerGroup() );
            
            actionToFile.put( test, cName );
            currentPlugins.add( test );
            
            //  check to see if we want to load this value
            Boolean b = (Boolean)pluginInfo.get( cName );
            if ( b != null && !b.booleanValue() ){
                return;
            }
            
            if ( l == null ) {
                l = new ArrayList();
                activeHandlers.put( test.getHandlerGroup(), l );                
                if ( test.getHandlerName().equals( "Object Select" ) ){
                    defaultHandler = test;
                }
            }
            removeItem( l, test );
            l.add( test );
            
            Log.info( "Loaded " + cName );
        }catch(Exception e){
            Log.info( "Couldn't load " + cName );
            e.printStackTrace( Log.getLogStream() );
        }catch(Error er ){
            Log.info( "Plugin caused an error condition." );
            er.printStackTrace();
        }
    }
    
    /** Removes the given EditAction from the list. */
    protected void removeItem( List l, EditAction e ){
        Iterator i = l.iterator();
        while ( i.hasNext() ) {
            EditAction n = (EditAction)i.next();
            if ( n.getHandlerName().equals( e.getHandlerName() ) ) {
                i.remove();
            }
        }
    }
    
    /** Returns a list of all plugins found. */
    static public List getPlugins() {
        return currentPlugins;
    }
    
    /** Returns whether the current action has been added to the
     * editing toolbar.
     */
    static public boolean isActive(EditAction e) {
        Boolean b = (Boolean)pluginInfo.get( getPath( e ) );
        if ( b == null ) { return true; }
        return b.booleanValue();
    }
    
    /** Returns the path to the disk file for the given EditAction. */
    static public String getPath(EditAction e ) {
        return (String)actionToFile.get( e );
    }
    
    /** Resets the EditBar.  Forces the reloading of all classes and
     * the reinitialization of the toolbar display.
     */
    static public void reset() {
        theBar.removeAll();
        activeHandlers.clear();
        currentPlugins.clear();
        actionToFile.clear();
        theBar.loadHandlers();
        JFrame jf = MainFrame.getActiveWindow();
        jf.setSize( jf.getWidth() - 1, jf.getHeight() );
        jf.setSize( jf.getWidth() + 1, jf.getHeight() );
        theBar.repaint(0);
    }
    
    static public EditBar getBar() { return theBar; }
    
    /** The plugins currently active in the scene. */
    static private List currentPlugins = new LinkedList();
    
    /** Keeps track of which plugins have been activated by the user. */
    static protected DataStore pluginInfo = new DataStore( "plugins" );
    
    static private Map activeHandlers = new HashMap();
    static private Map actionToFile = new HashMap();
    
    private List currentButtons = new ArrayList();
    private ButtonGroup buttons = new ButtonGroup();
    
    /** The button associated with the default handler. */
    protected AbstractButton defaultButton;
    
    /** The default handler.  This reference is used when the EditBar is first loaded,
     * reset, or setDefaultHandler() is called in ActiveAction. */
    protected EditAction defaultHandler;
    
    /** The singular instance of the EditBar. */
    static private EditBar theBar;
}

/** A simple wrapper class to relay activation events
 *  to an EditAction object.
 */
class SelectButton extends EditButton implements ActionListener  {
    private EditAction buttonHandler;
    
    SelectButton( EditAction action ) {
        super ( action.getHandlerName() );
        buttonHandler = action;
        addActionListener( this );
        setText( action.getHandlerName() );
        setToolTipText( action.getHandlerName() );
        java.awt.Image img = action.getHandlerImage();
        if ( img != null ) {
            myIcon = new ImageIcon( img );
        }
    }
    
    public void actionPerformed(java.awt.event.ActionEvent evt) {
        ActiveAction.setHandler( buttonHandler );
    }
    
}
