package molmaster.gui.editbar.analysis;

import java.io.*;
import java.util.*;
import javax.swing.*;

import molmaster.gui.*;
import molmaster.model.*;
import molmaster.model.loaders.*;
import molmaster.gui.editbar.*;
import molmaster.gui.tools.ErrorDialog;

/** Provides an interface to the CD analysis library
 *  developed by Andri Arnaldsson and Graeme Henkelman.
 *
 *  Allows a user to load a given CHG file, which is then
 *  analyzed.  The results may then be displayed via
 *  molmaster.
 *
 * @author  Russell Power
 */
public class CDAnalysis
	extends DefaultAction
	implements molmaster.model.Listener {

	private static final String CD_SURFACE_NAME = "_Charge Density Surface_";
	
	private double[][][] baseField;
	protected MaxRhoLoader mr = new MaxRhoLoader();
	protected VaspLoader vr = new VaspLoader();

	/** Creates a new instance of CNAnalysis */
	public CDAnalysis() {
	}

	static private boolean loadSuccess = false;
	static private boolean loadAttempted = false;

	static {
		String sep = System.getProperty("file.separator");
		try {
			String nativePath =
				System.getProperty("libdir")
					+ sep
					+ "charge"
					+ sep
					+ "charge.dll";
			molmaster.Log.writeLog("Loading native library: " + nativePath);
			Runtime.getRuntime().load(nativePath);
			loadSuccess = true;
		} catch (Throwable t) {
			molmaster.Log.writeLog("Couldn't load native CD analysis library.");
			molmaster.Log.writeLog(t);
		}
	}

	/** Native call to analyze the given file.  The report
	 * file is then inspected and displayed.
	 */
	private native void calcBader(String fileName);

	/** Called when a significant update has occured.  Significant is
	 *  defined as 'whatever Russell felt like implementing', but in general
	 *  all expected events ( mouse clicks, drags, etc. ) will be reported.
	 */
	public void actionPerformed(Object source, molmaster.Event evt) {
		if (evt.getID() == EditAction.handlerActivate) {
			if (loadSuccess) {
				try {
					JFileChooser f =
						new JFileChooser(System.getProperty("user.dir"));
					f.setDialogTitle("Select CHG file");
					f.showDialog(
						molmaster.gui.MainFrame.getActiveWindow(),
						"Analyze");

					final File selected = f.getSelectedFile();
					if (selected == null || !selected.exists()) {
						molmaster.Log.writeLog("Couldn't load file.");
						return;
					}
					Thread t = new CalcThread(selected.getCanonicalPath());
					new CDDialog(t).show();

					/** If the analysis proceeded correctly, we will now have
					 *  a few (currently one) file to read in from java and
					 *  display.    */
					File maxRho = new File("max_rho.dat");
					if (!maxRho.exists()) {
						molmaster.Log.writeLog("Couldn't find output file.");
						molmaster.Log.writeLog("Aborting analysis process...");
						return;
					}

					final String path = maxRho.getCanonicalPath();

					Thread lt = new Thread() {
						public void run() {
							try {
								MolModel theModel = MolModel.getModel();
								vr.loadFromFile(selected.getCanonicalPath());
								mr.loadFromFile(path);
								
								//	this is a hack to handle the fact that the current maxrho
								//	format doesn't export a basis matrix 
								mr.setBasis(vr.getBasis());
								baseField = mr.getField();
								
								vr.addToScene(theModel);
								mr.addToScene(theModel);
								
								Collection c = theModel.getAtoms();
								Iterator i = c.iterator();
								while (i.hasNext()) {
									AtomView.atomToView(
										(molmaster.Atom) i.next()).addListener(
										CDAnalysis.this);
								}
								JOptionPane.showMessageDialog(
									molmaster.gui.MainFrame.getActiveWindow(),
									"Field data has been loaded for the specified file.  You may\n"
										+ "now choose the Isosurface Visualization option to view the data.");
							} catch (Exception e) {
								new ErrorDialog( e, "Couldn't load max_rho file!", true).show();
							}
						}
					};
					lt.start();
				} catch (Exception ex) {
					molmaster.Log.writeLog(ex);
				} catch (Error e) {
					molmaster.Log.writeLog(e);
				}
			} else {
				JOptionPane.showMessageDialog(
					null,
					"The charge density analysis library failed to load.  Aborting analysis attempt.\n"
						+ "Please ensure that you have installed the native library pack for your OS.",
					"Error loading library",
					JOptionPane.ERROR_MESSAGE);
			}
		}

	}

	/** Returns the handler group this belongs to - selection handlers might compose
	 *  one group, while angle calculations would belong to another.  The groups
	 *  will be clustered on the screen to indicate their relation.
	 */
	public String getHandlerGroup() {
		return "Analysis";
	}

	/** Optional.  Return null if not implemented.  Returns a graphical icon to
	 *  display along with the text for this handler.    */
	public java.awt.Image getHandlerImage() {
		return null;
	}

	/** This is displayed to the user on the selection button for this handler.  */
	public String getHandlerName() {
		return "Charge Density Analysis";
	}

	//protected BranchGroup root;
	//private int quality = 5;
	
	private double[][][] tempField;
	
	/** We have been signaled by an AtomView that is has been
	 *  selected or unselected - we deal with that here.
	 */
	public void wakeUp(Object signaler, int signalType, Object value) {
		int x = baseField.length; int y = baseField[0].length; int z = baseField[0][0].length;
		if ( tempField == null) {
			tempField = new double[x][y][z];
		}
		
		AtomView av = (AtomView) signaler;
		double cur = av.getAtom().getID() + 1;
		molmaster.Log.info("Redetermining field for atom: " + cur);
		
		for ( int i = 0; i < x; ++i )
			for ( int j = 0; j < y; ++j )
				for ( int k = 0; k < z; ++k )
					tempField[i][j][k] = (baseField[i][j][k] == cur) ? 1 : 0;
		
		mr.setField(tempField);
			
		IsoDialog.Surface s = IsoSurface.getDialog().getSurface(CD_SURFACE_NAME);
		if(s == null) {
			s = new IsoDialog.Surface();
			IsoSurface.getDialog().addSurface(CD_SURFACE_NAME);
		}
		
		s.setIsoValue(0.5);
		

	}
	class CalcThread extends Thread {
		String name;

		CalcThread(String fileName) {
			name = fileName;
		}

		public void run() {
			try {
				Thread.currentThread().setPriority(Thread.MIN_PRIORITY);
				calcBader(name);
			} catch (Throwable t) {
				//new molmaster.gui.tools.ErrorDialog( t ).show();
				molmaster.Log.writeLog(t);
				t.printStackTrace(molmaster.Log.getLogStream());
			}
		}
	}
}
