package molmaster.gui.editbar.analysis;

import molmaster.*;
import molmaster.gui.*;
import molmaster.util.*;
import molmaster.model.*;
import molmaster.model.loaders.*;
import molmaster.gui.tools.reflect.*;

import java.io.*;
import java.util.*;
import java.awt.event.*;

import javax.swing.*;
import javax.vecmath.*;
import javax.swing.table.*;

/** Presents a table allowing a user to select various
 *  ring characteristics to apply to a file which can be
 *  highlighted.
 *
 * @author Russell Power
 */
public class RingDialog
	extends javax.swing.JDialog
	implements ActionListener, ComponentListener {

	static private boolean loadSuccess = false;
	protected FileList theList = FileList.getDefaultList();
	protected Map bins;
	protected Map numToAtom = new HashMap();
	protected Map finished = new HashMap();
	protected Collection atoms;

	private Object selected;
	private RingSettings settings = new RingSettings();
	private double maxBond = settings.getMaximumBondSize();
	private int maxRank = settings.getMaximumRingSize();
	private String allFiles = "All Files";
	private boolean working = false;
	private JButton btnCutoff = new JButton("Bond Cutoff");
	private Map specialBonds = new HashMap();

	private Runnable updateThread = new Runnable() {
		public void run() {
			updateFile();
		}
	};

	static {
		try {
			String sep = System.getProperty("file.separator");
			String nativePath =
				System.getProperty("libdir") + sep + "ring" + sep + "ring.dll";
			molmaster.Log.writeLog("Loading native library: " + nativePath);
			System.load(nativePath);
			loadSuccess = true;
		} catch (Throwable t) {
			molmaster.Log.writeLog("Couldn't load ring analysis plugin...");
			molmaster.Log.writeLog(t);
		}
	}

	public RingDialog() {
		super(molmaster.gui.MainFrame.getActiveWindow(), false);

		if (!loadSuccess) {
			JOptionPane.showMessageDialog(
				this,
				"The ring analysis library failed to load.  Aborting analysis attempt.\n"
					+ "Please ensure that you have installed the native library pack for your OS.",
				"Error loading library",
				JOptionPane.ERROR_MESSAGE);
			hide();
			dispose();
			return;
		}
		initComponents();
		theList.addListener(this);
		addComponentListener(this);

		setTitle("Ring Analysis");

		btnPanel.add(btnCutoff);

		btnCutoff.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				updateBins();
				List typeList = new ArrayList();
				Iterator i = bins.values().iterator();
				while (i.hasNext()) {
					Atom a = (Atom) ((List) i.next()).get(0);
					typeList.add(a);
				}

				BondList b = new BondList(typeList, specialBonds);
				b.show();
				specialBonds = b.getSpecial();
				double max = -1;
				i = specialBonds.values().iterator();
				while (i.hasNext()) {
					BondSize bs = (BondSize) i.next();
					if (bs.getMaxBondSize() > max) {
						max = bs.getMaxBondSize();
					}
				}
				if (max < maxBond) {
					maxBond = max;
				}

			}
		});
	}

	public void show() {
		if (!loadSuccess) {
			return;
		}
		theList.showListDialog(this);
		loadCombo();
		super.show();
	}

	private void initComponents() { //GEN-BEGIN:initComponents
		splitter = new javax.swing.JSplitPane();
		btnPanel = new javax.swing.JPanel();
		btnHelp = new javax.swing.JButton();
		btnRender = new javax.swing.JButton();
		btnAddFilter = new javax.swing.JButton();
		btnRemoveFilter = new javax.swing.JButton();
		btnSettings = new javax.swing.JButton();
		btnHistogram = new javax.swing.JButton();
		btnExport = new javax.swing.JButton();
		lstPanel = new javax.swing.JPanel();
		valSplitter = new javax.swing.JSplitPane();
		topPanel = new javax.swing.JPanel();
		lblFiles = new javax.swing.JLabel();
		currentFile = new javax.swing.JComboBox();
		jLabel1 = new javax.swing.JLabel();
		cmbTypes = new javax.swing.JComboBox();
		listScroll = new javax.swing.JScrollPane();
		filterTable = new javax.swing.JTable();

		addWindowListener(new java.awt.event.WindowAdapter() {
			public void windowClosing(java.awt.event.WindowEvent evt) {
				closeDialog(evt);
			}
		});

		splitter.setDividerLocation(130);
		btnPanel.setLayout(
			new javax.swing.BoxLayout(btnPanel, javax.swing.BoxLayout.Y_AXIS));

		btnHelp.setText("Help");
		btnHelp.setToolTipText("Display help for this dialog");
		btnHelp.addActionListener(new java.awt.event.ActionListener() {
			public void actionPerformed(java.awt.event.ActionEvent evt) {
				btnHelpActionPerformed(evt);
			}
		});

		btnPanel.add(btnHelp);

		btnRender.setText("Render");
		btnRender.setToolTipText(
			"Recalculate new rings given the current filters and render the data onscreen.");
		btnRender.addActionListener(new java.awt.event.ActionListener() {
			public void actionPerformed(java.awt.event.ActionEvent evt) {
				btnRenderActionPerformed(evt);
			}
		});

		btnPanel.add(btnRender);

		btnAddFilter.setText("Add Filter");
		btnAddFilter.setToolTipText("Add a new filter to the current list.");
		btnAddFilter.addActionListener(new java.awt.event.ActionListener() {
			public void actionPerformed(java.awt.event.ActionEvent evt) {
				btnAddFilterActionPerformed(evt);
			}
		});

		btnPanel.add(btnAddFilter);

		btnRemoveFilter.setText("Remove Filter");
		btnRemoveFilter.setToolTipText(
			"Removes the currently selected filter from the active list");
		btnRemoveFilter.addActionListener(new java.awt.event.ActionListener() {
			public void actionPerformed(java.awt.event.ActionEvent evt) {
				btnRemoveFilterActionPerformed(evt);
			}
		});

		btnPanel.add(btnRemoveFilter);

		btnSettings.setMnemonic('s');
		btnSettings.setText("Settings");
		btnSettings.setToolTipText("Change ring calculation options");
		btnSettings.addActionListener(new java.awt.event.ActionListener() {
			public void actionPerformed(java.awt.event.ActionEvent evt) {
				btnSettingsActionPerformed(evt);
			}
		});

		btnPanel.add(btnSettings);

		btnHistogram.setText("View Histogram");
		btnHistogram.addActionListener(new java.awt.event.ActionListener() {
			public void actionPerformed(java.awt.event.ActionEvent evt) {
				btnHistogramActionPerformed(evt);
			}
		});

		btnPanel.add(btnHistogram);

		btnExport.setText("Export Data");
		btnExport.addActionListener(new java.awt.event.ActionListener() {
			public void actionPerformed(java.awt.event.ActionEvent evt) {
				btnExportActionPerformed(evt);
			}
		});

		btnPanel.add(btnExport);

		splitter.setLeftComponent(btnPanel);

		lstPanel.setLayout(new java.awt.BorderLayout());

		valSplitter.setDividerLocation(50);
		valSplitter.setOrientation(javax.swing.JSplitPane.VERTICAL_SPLIT);
		topPanel.setLayout(new java.awt.GridLayout(2, 2));

		lblFiles.setText("Current File");
		topPanel.add(lblFiles);

		currentFile.addActionListener(new java.awt.event.ActionListener() {
			public void actionPerformed(java.awt.event.ActionEvent evt) {
				currentFileActionPerformed(evt);
			}
		});

		topPanel.add(currentFile);

		jLabel1.setText("Show only");
		topPanel.add(jLabel1);

		topPanel.add(cmbTypes);

		valSplitter.setLeftComponent(topPanel);

		filterTable
			.setModel(
				new javax
				.swing
				.table
				.DefaultTableModel(new Object[][] { {
					new Integer(8),
					new Integer(4)}
		}, new String[] { "ring count", "rank" }) {
			Class[] types =
				new Class[] {
					java.lang.Integer.class,
					java.lang.Integer.class };

			public Class getColumnClass(int columnIndex) {
				return types[columnIndex];
			}
		});
		listScroll.setViewportView(filterTable);

		valSplitter.setRightComponent(listScroll);

		lstPanel.add(valSplitter, java.awt.BorderLayout.CENTER);

		splitter.setRightComponent(lstPanel);

		getContentPane().add(splitter, java.awt.BorderLayout.CENTER);

		setSize(new java.awt.Dimension(400, 300));
	} //GEN-END:initComponents

	private void btnExportActionPerformed(java.awt.event.ActionEvent evt) {
		//GEN-FIRST:event_btnExportActionPerformed
		Runnable callback = new Runnable() {
			public void run() {
				btnExportActionPerformed(new ActionEvent(this, 0, "refresh"));
			}
		};
		MoleculeFile mf = (MoleculeFile) theList.getLoaderMap().get(selected);
		if (!isLoaded(selected)) {
			new CallbackThread(updateThread, callback).start();
			return;
		}
		Collection c = mf.getAtoms();
		Iterator i = c.iterator();
		JFileChooser fc = new JFileChooser();
		fc.setSelectedFile(new File("ringdata.sav"));
		fc.showSaveDialog(this);
		if (fc.getSelectedFile() != null) {
			File out = fc.getSelectedFile();
			try {
				FileWriter fOut = new FileWriter(out);
				PrintWriter pOut = new PrintWriter(fOut);
				pOut.println(
					"#Molmaster ring output: <atom number> - <ring size> <ring size> ... ");
				while (i.hasNext()) {
					Atom a = (Atom) i.next();
					List l = (List) a.getUserObject();
					Iterator j = l.iterator();
					StringBuffer lineBuf = new StringBuffer();
					lineBuf.append(a.getID());
					lineBuf.append(' ');
					while (j.hasNext()) {
						lineBuf.append(j.next());
						lineBuf.append(' ');
					}
					pOut.println(lineBuf.toString());
				}
				pOut.close();
				fOut.close();
			} catch (Exception e) {
				Log.writeLog(
					"Error writing to output file - error was: \n " + e);
			}
		}
	} //GEN-LAST:event_btnExportActionPerformed

	private void currentFileActionPerformed(java.awt.event.ActionEvent evt) {
		//GEN-FIRST:event_currentFileActionPerformed
		Object newSelected = currentFile.getSelectedItem();
		if (newSelected != selected) {
			MoleculeFile mf =
				(MoleculeFile) theList.getLoaderMap().get(selected);
			if (mf != null) {
				//mf.removeFromScene( MolModel.getModel() );
			}
		}

		selected = currentFile.getSelectedItem();
		if (selected instanceof File) {
			MoleculeFile mf =
				(MoleculeFile) theList.getLoaderMap().get(selected);
			atoms = mf.getAtoms();
			//mf.addToScene( MolModel.getModel() );
			System.out.println("Changed file...");
			updateBins();
			DefaultComboBoxModel cm = new DefaultComboBoxModel();
			Iterator i = bins.keySet().iterator();
			while (i.hasNext()) {
				cm.addElement(i.next());
			}
			cm.addElement("All Types");
			cmbTypes.setModel(cm);
		}
	} //GEN-LAST:event_currentFileActionPerformed

	private void btnSettingsActionPerformed(java.awt.event.ActionEvent evt) {
		//GEN-FIRST:event_btnSettingsActionPerformed
		new ReflectDialog(this, settings, true).show();
		if (maxBond == settings.getMaximumBondSize()
			&& maxRank == settings.getMaximumRingSize()) {
			return;
		}

		this.maxBond = settings.getMaximumBondSize();
		this.maxRank = settings.getMaximumRingSize();
		finished.clear();
	} //GEN-LAST:event_btnSettingsActionPerformed

	private PlotCanvas pc;
	private JDialog pcDialog;
	private boolean firstCall = true;
	private void btnHistogramActionPerformed(java.awt.event.ActionEvent evt) {
		//GEN-FIRST:event_btnHistogramActionPerformed
		Runnable callback = new Runnable() {
			public void run() {
				btnHistogramActionPerformed(
					new ActionEvent(this, 0, "refresh"));
			}
		};

		if (working) {
			return;
		}
		int[] bins = new int[this.maxRank * 2];
		int maxVal = 0;
		if (selected != allFiles) {
			if (!isLoaded(selected)) {
				new CallbackThread(updateThread, callback).start();
				return;
			} else {
				Iterator i = atoms.iterator();
				maxVal = addBins(i, bins, maxVal);
			}
		} else {
			if (firstCall) {
				firstCall = false;
				new CallbackThread(updateThread, callback).start();
				return;
			} else {
				Iterator i = theList.getLoaders().iterator();
				while (i.hasNext()) {
					MoleculeFile mf = (MoleculeFile) i.next();
					maxVal = addBins(mf.getAtoms().iterator(), bins, maxVal);
				}
				firstCall = true;
			}
		}

		Log.writeLog("Finished calculation - rendering...");
		if (pc == null) {
			pcDialog = new JDialog(this, "Ring Histogram", false);
			pc = new PlotCanvas(bins, maxVal * 3 / 2, 1, maxRank, 1);
			pcDialog.setSize(500, 500);
			pcDialog.getContentPane().add(pc);
			pcDialog.show();
		} else {
			pc.setValues(bins, maxVal * 3 / 2, 1, maxRank, 1);
		}

		pc.setInterpolate(false);

		pcDialog.setTitle("Ring Histogram - " + selected);
		pcDialog.show();
		pcDialog.invalidate();
		pcDialog.repaint();
	} //GEN-LAST:event_btnHistogramActionPerformed
	private void btnRenderActionPerformed(java.awt.event.ActionEvent evt) {
		//GEN-FIRST:event_btnRenderActionPerformed
		Runnable callback = new Runnable() {
			public void run() {
				render();
			}
		};

		//if (!isLoaded(selected)) {
		new CallbackThread(updateThread, callback).start();
		return;
		//}

	} //GEN-LAST:event_btnRenderActionPerformed

	private void btnRemoveFilterActionPerformed(
		java.awt.event.ActionEvent evt) {
		//GEN-FIRST:event_btnRemoveFilterActionPerformed
		DefaultTableModel dm = (DefaultTableModel) filterTable.getModel();
		int row = filterTable.getSelectedRow();
		if (row != -1) {
			dm.removeRow(row);
		}
	} //GEN-LAST:event_btnRemoveFilterActionPerformed

	private void btnAddFilterActionPerformed(java.awt.event.ActionEvent evt) {
		//GEN-FIRST:event_btnAddFilterActionPerformed
		DefaultTableModel dm = (DefaultTableModel) filterTable.getModel();
		Vector v = dm.getDataVector();
		int r = filterTable.getSelectedRow();
		Vector rowVec = null;
		if (r == -1) {
			rowVec = new Vector();
			rowVec.add(new Integer(5));
			rowVec.add(new Integer(5));
		} else {
			rowVec = (Vector) v.get(r);
		}

		dm.addRow(
			new Object[] {
				new Integer(((Integer) rowVec.get(0)).intValue()),
				new Integer(((Integer) rowVec.get(1)).intValue())});

	} //GEN-LAST:event_btnAddFilterActionPerformed

	private void btnHelpActionPerformed(java.awt.event.ActionEvent evt) {
		//GEN-FIRST:event_btnHelpActionPerformed
		molmaster.util.GetText.showHelpDialog(this, "RingDialog.help");
	} //GEN-LAST:event_btnHelpActionPerformed

	/** Closes the dialog */
	private void closeDialog(
		java.awt.event.WindowEvent evt) { //GEN-FIRST:event_closeDialog
		setVisible(false);
		dispose();
	} //GEN-LAST:event_closeDialog

	protected void updateFile() {
		if (working == true) {
			getToolkit().beep();
			molmaster.Log.writeLog("Please be patient!");
			MainFrame.setStatusMessage("Please be patient!");
			return;
		}

		File tempDir = new File("./temp");
		tempDir.mkdir();

		final ProgressMonitor pg =
			new ProgressMonitor(
				this,
				"Performing ring analysis...",
				"Initializing data",
				0,
				100);
		pg.setMillisToDecideToPopup(0);
		pg.setMillisToPopup(0);

		try {
			//  loading all files...
			if (selected == allFiles) {
				ComboBoxModel cm = currentFile.getModel();
				for (int i = 0; i < cm.getSize(); i++) {
					Object o = cm.getElementAt(i);
					if (o != allFiles) {
						currentFile.setSelectedItem(o);
						selected = o;
						updateFile();
					}
				}
				currentFile.setSelectedItem(allFiles);
				return;
			} else {
				working = true;
				Boolean b = (Boolean) finished.get(selected);
				if (!updateBins()) {
					MainFrame.setStatusMessage("Couldn't load file...");
					working = false;
					return;
				}
				if (b != null && b.booleanValue()) {
					pg.setProgress(90);
					pg.setNote(
						"Using previously calculated values... extracting data");
				} else {
					writeInputFile();
					pg.setProgress(5);
					writeDataFile();
					pg.setProgress(10);
					pg.setNote("Running analysis...");
					runAnalysis();
					pg.setProgress(90);
					pg.setNote("Finished analysis.  Extracting data...");
					readValues();
					pg.setProgress(95);
					finished.put(selected, new Boolean(true));
				}
				pg.setProgress(100);
				MainFrame.setStatusMessage("Finished calculation.");
			}
		} catch (IOException e) {
			new molmaster.gui.tools.ErrorDialog(e).show();
		}
		working = false;
	}

	public void actionPerformed(java.awt.event.ActionEvent actionEvent) {
		loadCombo();
		currentFileActionPerformed(actionEvent);
		selected = currentFile.getSelectedItem();
	}

	protected void loadCombo() {
		DefaultComboBoxModel cm = new DefaultComboBoxModel();
		Iterator i = theList.getFiles().iterator();
		while (i.hasNext()) {
			File f = (File) i.next();
			cm.addElement(f);
		}
		cm.addElement(allFiles);

		currentFile.setModel(cm);
	}

	protected boolean updateBins() {
		Collection c = theList.getLoaders();
		//  put each atom into a bin...
		if (selected == null) {
			java.awt.Toolkit.getDefaultToolkit().beep();
			return false;
		}

		bins = new HashMap();
		MoleculeFile mf = (MoleculeFile) theList.getLoaderMap().get(selected);
		atoms = mf.getAtoms();
		Log.writeLog("Creating bins for: " + atoms.size() + " atoms.");
		Iterator i = atoms.iterator();
		while (i.hasNext()) {
			Atom a = (Atom) i.next();
			AtomAttributes aa = (AtomAttributes) a.getAttributes();
			List l = (List) bins.get(aa.atomicSymbol);
			if (l == null) {
				l = new LinkedList();
				bins.put(aa.atomicSymbol, l);
			}
			l.add(a);
		}
		return true;
	}
	protected void writeInputFile() throws java.io.IOException {
		File outFile = new File("ringstat.ipt");
		FileWriter fw = new FileWriter(outFile);
		BufferedWriter bw = new BufferedWriter(fw);

		System.out.println("Writing input for: " + selected);
		bw.write("MOLMASTER INPUT FOR RINGSTAT\n");
		bw.write("molmaster.in\n"); //  input file
		bw.write("temp/" + selected.toString() + "\n"); //  output file
		bw.write("1 1 1\n");
		bw.write("original set\n"); // use all atoms
		bw.write(maxRank + " \n"); // maximum rank
		bw.write(maxBond + " \n"); // maximum bond size
		bw.write("all      \n"); // examine all atoms
		bw.write("-10 10000\n"); // examine all space

		if (bins == null) {
			java.awt.Toolkit.getDefaultToolkit().beep();
			return;
		}

		Set s = bins.keySet();

		//bw.write(s.size() * s.size() + "\n");

		Iterator i = s.iterator();
		int bondCount = 0;
		int countI = 0;
		String binOut = "";
		while (i.hasNext()) {
			countI++;
			int countJ = 0;
			List la = (List) bins.get(i.next());
			Atom a = (Atom) la.get(0);
			AtomAttributes aa = (AtomAttributes) a.getAttributes();
			Iterator j = s.iterator();
			while (j.hasNext()) {
				countJ++;
				List lb = (List) bins.get(j.next());
				Atom b = (Atom) lb.get(0);
				AtomAttributes ba = (AtomAttributes) b.getAttributes();
				if (ba.atomName.equals(aa.atomName)) {
					continue;
				} else {
					Bond bond = new Bond(a, b, Bond.BOND_TYPE_NORMAL);
					double bondSize = 0;
					String type = bond.getBondTypeID();
					if (specialBonds.containsKey(type)) {
						BondSize bOut = (BondSize) specialBonds.get(type);
						bondSize = bOut.getMaxBondSize();
					} else {
						bondSize = Bond.bondLength(a, b);
					}
					bondCount++;
					binOut
						+= new PrintfFormat("%-8s%-8s%f").sprintf(
							new Object[] {
								String.valueOf((char) (countI + 'a')),
								String.valueOf((char) (countJ + 'a')),
								new Float(bondSize)});
					binOut += "\n";
				}
			}
		}

		bw.write(bondCount + "\n");
		bw.write(binOut);

		bw.close();
		fw.close();
	}

	/** Writes a POSCAR type input file to be fed into
	 *  the ring analysis routines.  This file contains
	 *  all the atoms currently found in the file list
	 *  being used for analysis.
	 */
	protected void writeDataFile() throws java.io.IOException {
		if (bins == null) {
			java.awt.Toolkit.getDefaultToolkit().beep();
			return;
		}
		FileWriter fw = new FileWriter("molmaster.in");
		BufferedWriter bw = new BufferedWriter(fw);
		double cellSize = 10;
		MoleculeFile mf = (MoleculeFile) theList.getLoaderMap().get(selected);
		if (mf instanceof VaspLoader) {
			cellSize = ((VaspLoader) mf).getBoxScale();
		}

		bw.write("molmaster generated POSCAR file for ring analysis\n");
		bw.write(new PrintfFormat("%2.14f").sprintf(cellSize) + "\n");
		bw.write(
			".9999999999999998     .0000000000000000     .0000000000000000\n");
		bw.write(
			".0000000000000000     .9999999999999998     .0000000000000000\n");
		bw.write(
			".0000000000000000     .0000000000000000     .9999999999999998\n");

		Iterator i = bins.values().iterator();
		while (i.hasNext()) {
			bw.write(((List) i.next()).size() + "\n");
		}

		for (int j = 0; j < 200 - bins.size(); j++) {
			bw.write("0\n");
		}

		bw.write("Selective dynamics\n");
		bw.write("Direct\n");

		numToAtom = new HashMap();
		i = bins.values().iterator();
		int count = 0;
		PrintfFormat outF = new PrintfFormat("%0.16f");
		while (i.hasNext()) {
			Iterator j = ((List) i.next()).iterator();
			while (j.hasNext()) {
				Atom a = (Atom) j.next();
				Vector3d v = a.getLocation();

				//String x = outF.sprintf( v.x / cellSize ).substring(1) + "\t";
				String x = "" + v.x / cellSize + "\t";
				bw.write(x);
				//String y = outF.sprintf( v.y / cellSize ).substring(1) + "\t";
				String y = "" + v.y / cellSize + "\t";
				bw.write(y);
				//String z = outF.sprintf( v.z / cellSize ).substring(1) + "\t";
				String z = "" + v.z / cellSize + "\t";
				bw.write(z);
				bw.write("T\tT\tT");

				bw.write("\n");
				numToAtom.put(new Integer(count++), a);
			}
		}
		bw.close();
		fw.close();
	}

	protected native void runAnalysis();

	protected void readValues() {
		Set ringList = new HashSet();
		try {
			FileReader fr = new FileReader("molmaster.out");
			BufferedReader buf = new BufferedReader(fr);
			String line = null;
			boolean valid = true;

			while (valid) {
				Set ring = new HashSet();
				line = buf.readLine();
				if (line == null) {
					valid = false;
					break;
				}
				StringTokenizer tok = new StringTokenizer(line, " ");
				while (tok.hasMoreTokens()) {
					Integer val = new Integer(tok.nextToken());
					if (val.intValue() == 0) {
						continue;
					}
					Integer atomVal = new Integer(val.intValue() - 1);
					ring.add(numToAtom.get(atomVal));
				}
				ringList.add(ring);
			}
		} catch (Exception e) {
			molmaster.Log.writeLog("Couldn't read output file...");
			e.printStackTrace(molmaster.Log.getLogStream());
		}

		Iterator i = atoms.iterator();
		while (i.hasNext()) {
			Atom a = (Atom) i.next();
			a.setUserObject(new ArrayList());
		}

		i = ringList.iterator();
		while (i.hasNext()) {
			Set s = (Set) i.next();
			Iterator j = s.iterator();
			while (j.hasNext()) {
				Atom a = (Atom) j.next();
				ArrayList al = (ArrayList) a.getUserObject();
				al.add(new Integer(s.size()));
			}
		}
	}

	/** Determines whether the given input atom fulfills the
	 *  criteria set forth in the given filter.  Returns true
	 *  if the filter is satisfied by the ring status of the
	 *  atom.
	 */
	protected boolean valid(Atom a, List filter) {
		List l = (List) a.getUserObject();
		//  TODO - deal with type information
		//String type = (String)l.get( l.size() - 1);
		List temp = new ArrayList(l);
		AtomAttributes aa = (AtomAttributes) a.getAttributes();
		Iterator i = filter.iterator();
		while (i.hasNext()) {
			if (!temp.remove(i.next()))
				return false;
		}
		Object showType = cmbTypes.getSelectedItem();
		if (showType.equals("All Types")) {
			return true;
		} else if (showType.equals(aa.atomicSymbol)) {
			return true;
		}
		return false;
	}

	/** Calculates the list of criteria required of each atom
	 *  given by the filter table.
	 */
	protected List calcFilter() {
		List calcOut = new ArrayList();
		Vector rows =
			((DefaultTableModel) filterTable.getModel()).getDataVector();
		Iterator i = rows.iterator();
		while (i.hasNext()) {
			Vector row = (Vector) i.next();
			int count = ((Integer) row.get(0)).intValue();
			int rank = ((Integer) row.get(1)).intValue();
			//String type = (String) row.get(2);

			for (int j = 0; j < count; j++) {
				calcOut.add(new Integer(rank));
			}
			//calcOut.add( type );
		}
		return calcOut;
	}

	/** Renders the result of the given filter table input.  Atoms
	 *  that fulfill the filter criteria are selected, those that
	 *  do not are unselected.
	 */
	protected void render() {
		//  don't try and render something if it isn't a file...
		if (!(selected instanceof File)) {
			return;
		}
		List filter = calcFilter();
		Iterator i = atoms.iterator();
		List renderOut = new LinkedList();
		while (i.hasNext()) {
			Atom a = (Atom) i.next();
			AtomView av = molmaster.gui.AtomView.atomToView(a);
			if (av != null) {
				if (valid(a, filter)) {
					av.select();
				} else {
					av.unselect();
				}
			}
		}
	}

	private boolean isLoaded(Object file) {
		Boolean b = (Boolean) finished.get(file);
		if (b == null) {
			return false;
		}
		return b.booleanValue();
	}

	private int addBins(Iterator i, int[] bins, int maxVal) {
		while (i.hasNext()) {
			Atom a = (Atom) i.next();
			if (a == null) {
				continue;
			}
			List l = (List) a.getUserObject();
			if (l == null) {
				continue;
			}
			Iterator j = l.iterator();
			while (j.hasNext()) {
				int m = ((Integer) j.next()).intValue();
				bins[m]++;
				if (bins[m] > maxVal) {
					maxVal = bins[m];
				}
			}
		}
		return maxVal;
	}

	public void componentHidden(java.awt.event.ComponentEvent componentEvent) {
	}
	public void componentMoved(java.awt.event.ComponentEvent componentEvent) {
	}
	public void componentResized(
		java.awt.event.ComponentEvent componentEvent) {
	}
	public void componentShown(java.awt.event.ComponentEvent componentEvent) {
		loadCombo();
		currentFileActionPerformed(new ActionEvent(this, 0, "update"));
	}

	// Variables declaration - do not modify//GEN-BEGIN:variables
	private javax.swing.JPanel topPanel;
	private javax.swing.JButton btnRemoveFilter;
	private javax.swing.JLabel jLabel1;
	private javax.swing.JButton btnAddFilter;
	private javax.swing.JTable filterTable;
	private javax.swing.JLabel lblFiles;
	private javax.swing.JPanel btnPanel;
	private javax.swing.JComboBox cmbTypes;
	private javax.swing.JScrollPane listScroll;
	private javax.swing.JButton btnExport;
	private javax.swing.JButton btnSettings;
	private javax.swing.JComboBox currentFile;
	private javax.swing.JButton btnRender;
	private javax.swing.JSplitPane valSplitter;
	private javax.swing.JPanel lstPanel;
	private javax.swing.JButton btnHistogram;
	private javax.swing.JButton btnHelp;
	private javax.swing.JSplitPane splitter;
	// End of variables declaration//GEN-END:variables

}
