package molmaster.gui.toolbar;

import java.io.File;
import javax.swing.*;
import java.awt.*;
import java.awt.image.*;
import java.awt.event.*;

import java.util.*;

/** Implements a button that is located on a ToolBar.  The
 *  EditButton class provides functionality for rearranging
 *  components and adding and removing components from the bar.
 *
 *  Also implemented are dynamic drop shadows - drop shadows are
 *  calculated for a controls icon when the control is created.
 *
 * @author  Russell Power
 */
public class EditButton extends JToggleButton
implements MouseMotionListener, MouseListener  {
    
    protected ImageIcon myIcon;
    protected ImageIcon shadowIcon;
    
    protected String myText;
    private boolean dragged = false;
    private boolean highlight = false;
    private boolean selected = false;
    
    static public int imageWidth = 20;
    static public int imageHeight = 20;
    
    protected static final BufferedImage blank = 
    new BufferedImage( imageWidth, imageHeight, BufferedImage.TYPE_INT_ARGB );
    protected static final ImageIcon blankIcon =  new ImageIcon( blank );
    
    static public javax.vecmath.Color3f dragColor = new
    javax.vecmath.Color3f( 0.55f, 0.78f, 0.91f );
    
    static public javax.vecmath.Color3f defaultColor = new
    javax.vecmath.Color3f( 1.0f, 1.0f, 1.0f );
    
    static public javax.vecmath.Color3f selectedColor= new
    javax.vecmath.Color3f( 0.7f, 0.5f, 0.5f );
    
    static {
        molmaster.gui.tools.OptionWindow options =
        molmaster.gui.tools.OptionWindow.getWindow();
        
        options.addValue(
        "Toolbar", "Drag color", "The color seen when a toolbar button is moved.",
        EditButton.class, "dragColor" );
        
        options.addValue(
        "Toolbar", "Back color", "The background color used for toolbar buttons.",
        EditButton.class, "defaultColor" );
        
        options.addValue(
        "Toolbar", "Selected color", "The background color used for a selected button.",
        EditButton.class, "selectedColor" );
        
        options.addValue(
        "Toolbar", "Icon Width", "The width to use when creating toolbar icons.",
        EditButton.class, "imageWidth" );
        
        options.addValue(
        "Toolbar", "Icon Height", "The height to use when creating toolbar icons.",
        EditButton.class, "imageWidth" );
    }
    
    public EditButton( String text ) {
        super( text, false );
        
        myText = text;
        myIcon = findImage( text );
        addMouseMotionListener( this );
        addMouseListener( this );
        
        setBackground( defaultColor.get() );
        setToolTipText( text );
        setName( text );
        setBorderPainted( false );
        
        if ( myIcon != null ) {
            Image img = myIcon.getImage();
            int w = imageWidth;
            int h = imageHeight;
            
            BufferedImage scale = new BufferedImage (
            imageWidth, imageHeight, BufferedImage.TYPE_INT_ARGB );
            scale.getGraphics().drawImage( img, 0, 0, imageWidth, imageHeight, this );
            myIcon = new ImageIcon( scale );
            setIcon( myIcon );
            
            BufferedImage icon = new BufferedImage( 
            imageWidth, imageHeight, BufferedImage.TYPE_INT_ARGB );
            
            BufferedImage back = new BufferedImage(
            imageWidth, imageHeight, BufferedImage.TYPE_INT_ARGB );

            back.getGraphics().drawImage( img, 0, 0, imageWidth, imageHeight, this );
            
            WritableRaster wr = back.getRaster();
            int[] pixel = new int[4];
            int[] trans = new int[] { 0, 0, 0, 0 };
            int[] black = new int[] { 128, 128, 128, Integer.MAX_VALUE };
            for ( int i = 0; i < w; i++ ){
                for ( int j = 0; j < h; j++ ) {
                    wr.getPixel( i, j, pixel );
                    if ( pixel[3] > 0 ) {
                        wr.setPixel( i, j, black );
                    }else {
                        wr.setPixel( i, j, trans );
                    }
                }
            }
            
            Graphics g = icon.getGraphics();
            g.drawImage( back, 1, 1, this );
            g.drawImage( back, 2, 2, this );
            g.drawImage( back, 3, 3, this );
            g.drawImage( scale, 0, 0, this );
            shadowIcon = new ImageIcon( icon );
        }else {
        	myIcon = new ImageIcon(blank);
        }
        
        setSmall( true );
        //setBorder( BorderFactory.createLineBorder( java.awt.Color.black ) );
    }
    
    /** Sets the layout of this button into small
     *  or expanded mode.
     */
    public void setSmall( boolean  mode) {
        if ( mode ) {
            setText( "" );
            if ( myIcon != null ) { setIcon( myIcon ); }
        }else {
            setText( myText );
        }
    }
    
    private ImageIcon findImage( String name ) {
        File f = new File( "graphics/toolbar/" + name + ".gif" );
        if ( f.exists() ) { return new ImageIcon( f.getAbsolutePath() ); }
        return null;
    }
    
    public Icon getIcon() { 
        if ( highlight ) { return shadowIcon; }
        return super.getIcon();
    }
    
    public void paint( Graphics g ) {
        if ( isSelected() ){
            g.setColor( selectedColor.get() );
        }else if ( highlight ) {
            g.setColor( dragColor.get() );
        }else {
            g.setColor( defaultColor.get() );
        }
        
        g.fillRect( 0, 0, getWidth(), getHeight() );
        g.setFont( getFont() );
        
        Icon i = getIcon();
        
		int ix, iy = 0;
        ix = 4;
        iy = (getHeight() - imageHeight) / 2;
        
        if ( i != null )
        	i.paintIcon( this, g, ix, iy);
        
        g.setColor( Color.black );
        g.drawString( getText(), imageWidth + ix + getIconTextGap(), getHeight() / 2 );
    }
    
    public void mouseDragged(java.awt.event.MouseEvent evt) {
        if ( !dragged ) {
            if ( evt.isAltDown() ) {
                setBackground(
                new java.awt.Color( dragColor.x, dragColor.y, dragColor.z, 0.5f ));
                dragged = true;
                Point p = getLocation();
                java.awt.Container c = getParent();
                c.remove(this);
                c.add( this, 0 );
                setLocation( p );
            }
        }
        if ( dragged ) {
            if( isSmall() ) {
                setLocation( (getX() - getWidth() /2 ) + evt.getX(), getY() );
            }else {
                setLocation( getX(), getY() - getHeight() / 2 + evt.getY() );
            }
        }
    }
    
    public boolean isSmall() {
        return ((JToolBar)getParent()).getOrientation() == JToolBar.HORIZONTAL;
    }
    
    protected void fireActionPerformed( ActionEvent e ) {
        if ( !dragged )
            super.fireActionPerformed( e );
    }
    
    public void mouseMoved(java.awt.event.MouseEvent mouseEvent) {}
    public void mouseClicked(java.awt.event.MouseEvent mouseEvent) {}
    
    public void mouseEntered(java.awt.event.MouseEvent mouseEvent) {
        highlight = true; repaint();}
    public void mouseExited(java.awt.event.MouseEvent mouseEvent) {
        highlight = false; repaint();
    }
    public void mousePressed(java.awt.event.MouseEvent mouseEvent) {}
    public void mouseReleased(java.awt.event.MouseEvent mouseEvent) {
        if( !dragged ) { return; }
        dragged = false;
        before.clear(); after.clear();
        java.awt.Component[] c = getParent().getComponents();
        boolean testH = isSmall();
        
        int myLoc = testH ? getX() : getY();
        
        for ( int i = 0; i < c.length; i++ ) {
            int testLoc = testH ? c[i].getX() : c[i].getY();
            if ( testLoc < myLoc ) { before.add( c[i] ); }
            else { after.add( c[i] ); }
        }
        
        java.awt.Container parent = getParent();
        parent.removeAll();
        
        Iterator i = before.iterator();
        while ( i.hasNext() ) { parent.add( ((Component)i.next()) ); }
        parent.add( this );
        i = after.iterator();
        while ( i.hasNext() ) { parent.add( ((Component)i.next()) ); }
    }
    
    static private ArrayList before = new ArrayList();
    static private ArrayList after = new ArrayList();
    
}
