package molmaster.gui.tools;

import java.net.*;
import java.io.*;
import java.util.*;
import javax.swing.*;


/**
 * MailSender is a simple class to send SMTP e-mail
 * based on the classes Letter and MailMergeFile.
 * This class is a Runnable, and the run() method
 * actually does the mailing.  It is also an
 * Observable, and sends progress reports to its
 * observers.
 * <p>
 * This class is not even close to a partial subset
 * implementation of SMTP.  For real production
 * applications, use of the JavaMail standard extension
 * would be a better choice.
 *
 * @author Neal Ziring, ziring@home.com
 */

public class MailSender extends Observable implements Runnable {
    
    private String	  server;
    private String	  sender;
    private String        dest;
    private Thread	  myThread;
    private Socket	  sock;
    private int		  smtpPort;
    private String       exMessage;
    private boolean	  abort;
    private JDialog       cancel;
    
    public boolean        finished = false;
    public boolean        good = false;
    
    /**
     * Create a MailSender object
     */
    public MailSender(String sv, String from, String to, int port, String message, JDialog cancel  ) {
        sock = null;
        server = sv;
        smtpPort = port;
        sender = from;
        dest = to;
        exMessage = message;
        abort = false;
        this.cancel = cancel;
        myThread = new Thread(this);
    }
    
    /**
     * Abort the run, if running.
     */
    public synchronized void abortRun() {
        abort = true;
    }
    
    /**
     * Start the run (can only be called once per object.)
     */
    public synchronized void start() {
        abort = false;
        myThread.start();
    }
    
    protected boolean doAbort() { return abort; }
    
    /**
     * Send all the letters to the mail server,
     * report activities to our observers.
     */
    public void run() {
        final String CRLF = "\r\n";
        BufferedReader br = null;
        PrintStream ps;
        String line;
        try {
            sock = new Socket(server, smtpPort);
            ps = new PrintStream(sock.getOutputStream());
            br = new BufferedReader(new InputStreamReader(sock.getInputStream()));
            line = br.readLine();
        }
        catch (Exception e) {
            setChanged();
            notifyObservers("Problem setting up connection: " + e);
            cancel.setVisible(false); cancel.dispose();
            return;
        }
        
        setChanged();
        notifyObservers("Connected to mail server " + server + ":" + smtpPort);
        if (doAbort()) {
            try { ps.close(); br.close(); sock.close(); }
            catch (Exception ex) { }
            setChanged();
            notifyObservers("Job aborted early, connection closed.");
            cancel.setVisible(false); cancel.dispose();
            return;
        }
        
        try {
            ps.print("MAIL FROM:"); ps.print(sender); ps.print(CRLF);
            line = br.readLine();
            if (line.startsWith("25")) {
                setChanged();
                notifyObservers("Sending to: " + dest);
            }
            else {
                setChanged();
                notifyObservers("Error from mail server (on MAIL): " + line);
            }
            ps.print("RCPT TO:"); ps.print(dest); ps.print(CRLF);
            line = br.readLine();
            if (!line.startsWith("25")) {
                setChanged();
                notifyObservers("Error from mail server (on RCPT): " + line);
            }
            ps.print("DATA"); ps.print(CRLF);
            line = br.readLine();
            if (!line.startsWith("35")) {
                setChanged();
                notifyObservers("Error from mail server (on DATA): " + line);
            }
            ps.print( "From: molmaster" + CRLF);
            ps.print( "To: \"Russell Power\" <rjpower@washington.edu>" + CRLF);
            ps.print( "Subject: Molmaster Error" + CRLF + CRLF + CRLF);
            
            ps.print( exMessage );
            ps.print(CRLF);
            ps.print(".");
            ps.print(CRLF);
            setChanged();
            notifyObservers("Mail sent to " + dest + " ok.");
            good = true;
            line = br.readLine();
            
            if (doAbort()) {
                try { ps.close(); br.close(); sock.close(); }
                catch (Exception ex) { }
                setChanged();
                notifyObservers("Job aborted early, connection closed.");
                cancel.setVisible(false); cancel.dispose();
                return;
            }
        }
        catch (Exception e2) {
            setChanged();
            notifyObservers("Problem reading information from server: " + e2);
        }
        
        try {
            br.close();
            ps.close();
            sock.close();
        }
        catch (Exception e3) {
            setChanged();
            notifyObservers("Problem closing down connection: " + e3);
        }
        
        good = true;
        cancel.setVisible(false); cancel.dispose();
        return;
    }
    
}

