package molmaster.gui.tools;

import molmaster.*;
import molmaster.model.*;
import molmaster.model.loaders.*;
import molmaster.gui.*;
import molmaster.gui.tools.reflect.*;

import java.io.*;
import java.util.*;
import java.awt.image.*;

import image.codec.*;

import javax.swing.*;
import javax.vecmath.*;

/** Represents a 'movie' - a collection of keyframes representing
 *  atoms and bonds, which are interpolated between via a cubic spline.
 *
 * @author  rjpower
 */
public class Movie implements CapturedImageObserver, WatchedObject {
    
    protected MovieFile movieBase;
    protected int interpSteps = 1;
    protected int currentPos = 0;
    protected int sleepTime = 5;
    protected int currentFrameWrite = 0;
    protected int startPos = 0;
    protected double boundary = 1;
    
    protected java.awt.Dimension canvasSize = MainFrame.canvas.getSize();
    
    protected boolean calcBonds = false;
    protected boolean showBonds = false;
    protected boolean antialiased = false;
    protected boolean recording = false ;
    protected boolean imageDone = false;
    protected boolean running = false;
    protected boolean paused = false;
    
    protected Runnable movieThread;
    
    protected File destDir = new File("f:\\temp\\");
    
    protected List atoms = new ArrayList();
    protected double getMaxDist( List a ) {
        Vector3d s = ((Atom)a.get(0)).getLocation();
        Vector3d m = ((Atom)a.get(a.size() - 1)).getLocation();
        Vector3d l = ((Atom)a.get(a.size() / 2)).getLocation();
        
        Vector3d temp = new Vector3d();
        temp.sub( s, l );
        double sl = temp.lengthSquared();
        temp.sub( s, m );
        double sm = temp.lengthSquared();
        temp.sub( m, l );
        double ml = temp.lengthSquared();
        
        double max = sl;
        if ( sm > max ) { max = sm; }
        if ( ml > max ) { max = ml; }
        
        return max;
    }
    
    public Movie(MovieFile m) throws IllegalArgumentException{
        movieBase = m;
        List a = m.getAtomFrame( 0 );
        
        boundary = getMaxDist( a ) / 2;
        
        if ( m.getFrameCount() < 2 ) {
            throw new IllegalArgumentException( "Movie file must contain more than one frame!" );
        }
        
        initMovie(a);
    }
    protected void initMovie(List a) {
        atoms.clear();
        Iterator i = a.iterator();
        while ( i.hasNext() ) {
            MovieAtom ma = new MovieAtom( (Atom)i.next(), a, boundary );
            atoms.add( ma );
        }
    }
    
    public MovieFile getFile() { return movieBase; }
    public double getBoundarySize() { return Math.sqrt(boundary); }
    /** Sets the distance to use for computing bonding.  All values result in correct
     * results - larger values indicate more atoms checked per individual frame, while
     * smaller values may force a recomputation if the movement of atoms exceeds the
     * boundary size.
     */
    public void setBoundarySize(double d) {
        boundary = Math.pow(d,2);
        initMovie( movieBase.getAtomFrame( 0 ) );
    }
    
    public File getDestinationDirectory() { return destDir; }
    /** Sets the destination for this movies output images.
     */
    public void setDestinationDirectory(File dest) {
        boolean created = false;
        if ( !dest.exists() ) {
            try {
                dest.mkdirs();
            }catch(Exception e){
                e.printStackTrace( Log.getLogStream() );
            }
        }
        if ( !dest.isDirectory() ) {
            throw new IllegalArgumentException( "Only directories may be specified." );
        }
        if ( created ){
            dest.delete();
        }
        destDir = dest;
    }
    
    public int getMoviePosition() { return currentPos; }
    public int getFrameCount() { return interpSteps * (movieBase.getFrameCount() - 2); }
    /** Sets the number of frames this movie should have in total.  This number is
     * rounded down to the nearest multiple of the number of frames the movie is
     * based on.
     */
    public void setFrameCount( int val ) { interpSteps = (val - 1 )/ (movieBase.getFrameCount()-1); }
    
    public int getSpeed() { return 101 - sleepTime; }
    public RangeValue SpeedRange() {
        RangeValue r = new RangeValue();
        r.min = 0; r.max = 100;
        r.labelTable.put( new Integer(0), new JLabel("Slow") );
        r.labelTable.put( new Integer(100), new JLabel("Fast") );
        return r;
    }
    
    /** Sets the playback speed of the movie( used while reviewing ).
     */
    public void setSpeed( int speed ) {
        if ( speed > 100 ) { speed = 100; }
        sleepTime = 101 - speed;
    }
    
    /** Sets the width of the movie canvas.
     */
    public void setWidth( int w ) { canvasSize.width = w; MainFrame.canvas.setSize( canvasSize ); }
    /** Sets the height of the movie canvas.
     */
    public void setHeight( int h ) { canvasSize.height = h; MainFrame.canvas.setSize( canvasSize ); }
    
    public int getWidth() { return canvasSize.width; }
    public int getHeight() { return canvasSize.height; }
    
    public void setRecalculateBonds( boolean val ) { calcBonds = val; }
    public boolean getRecalculateBonds() { return calcBonds; }
    
    public boolean getRecording() { return recording; }
    public void setRecording( boolean val ) { recording = val; }
    
    public boolean getAntialiasScene() { return antialiased; }
    /** Sets whether to activate scene antialiasing when movie recording begins.
     */
    public void setAntialiasScene(boolean val) { antialiased = val; }
    
    public boolean getShowBonds() { return showBonds; }
    public void setShowBonds( boolean val ) { showBonds = val; }
    
    /** Sets the number of steps to interpolate between base frames.
     */
    public void setInterpolationSteps( int interp ) {interpSteps = interp; }
    public int getInterpolationSteps() { return interpSteps; }
    
    /** Stops the movie currently playing.
     */
    public void stop() {
        if ( !running ) { return; }
        running = false;
        paused = false;
        synchronized( movieThread ) {
            setMoviePosition( startPos );
        }
    }
    
    /** Pauses the current movie.
     */
    public void pause() {
        if (!running) { return; }
        if ( paused ) { paused = false; return; }
        //synchronized ( movieThread ) {
        paused = true;
        //}
    }
    
    /** Starts the movie from the current position.
     */
    public void start() {
        
        if ( running && !paused ) { return; }
        if ( paused ) { paused = false; return; }
        
        running = true;
        
        MainFrame.uni.getViewer().getView().setSceneAntialiasingEnable( antialiased );
        
        startPos = currentPos = 0;
        currentFrameWrite = 0;
        if ( !destDir.exists() ){
            try{
                destDir.mkdirs();
            }
            catch(Exception e){
                javax.swing.JOptionPane.showMessageDialog( MainFrame.getActiveWindow(),
                "Couldn't create directory for storing movie files - error message follows:\n\n" + e,
                "Directory creation error",
                javax.swing.JOptionPane.ERROR_MESSAGE );
                return;
            }
        }
        
        movieThread = new Runnable() {
            public void run() {
                int startFrame = currentPos * movieBase.getFrameCount() / getFrameCount();
                startPos = startFrame * interpSteps;
                HashMap movements = new HashMap();
                
                for ( int i = startFrame; i < movieBase.getFrameCount() - 1; i++ ) {
                    synchronized( this ) {
                        if ( !running ) {
                            notifyAll();
                            return;
                        }
                        while ( paused ) {
                            notifyAll();
                            try {
                                Thread.sleep( 100 );
                            }catch(Exception e){}
                        }
                        
                        movements.clear();
                        getFrameMovements( movements, i, i+1 );
                        for ( int j = 0; j < currentPos - startPos; j++ ) {
                            stepFrame( atoms.iterator(), movements );
                            if (recording) {
                                imageDone = false;
                                Log.writeLog("Waiting for image to be written.");
                                MainFrame.canvas.addCaptureObserver( Movie.this );
                                while ( !imageDone ) {
                                    try {
                                        Thread.sleep( 10 );
                                    }catch(Exception e){}
                                }
                                Log.info( "continuing..." );
                            }
                            else {
                                try {
                                    Thread.sleep( sleepTime );
                                }catch(Exception e){}
                            }
                            
                        }
                        startPos = i*interpSteps;
                        currentPos = (i+1)*interpSteps;
                    }
                }
                running = false;
                MainFrame.uni.getViewer().getView().setSceneAntialiasingEnable( false );
                //setMoviePosition( startPos );
            }
        };
        
        new Thread(movieThread).start();

    }
    
    /** Sets the key/value pairs in the given map according the linear interpolation
     * of each atom from the given starting frame to the end frame.  Each atom is
     * associated with a given vector representing the movement.
     */
    protected void getFrameMovements( Map movements, int frameStart, int frameEnd ) {
        Iterator i = movieBase.getAtomFrame( frameStart ).iterator();
        Iterator j = movieBase.getAtomFrame( frameEnd ).iterator();
        
        int count = 0;
        while ( i.hasNext() && j.hasNext() ) {
            Atom st = (Atom)i.next();
            Atom end = (Atom)j.next();
            Vector3d dist = new Vector3d();
            dist.sub( end.getLocation(), st.getLocation() );
            dist.scale( 1.0/interpSteps );
            
            MovieAtom rep = (MovieAtom)atoms.get(count++);
            rep.setLocation( st.getLocation(), calcBonds );
            movements.put( rep, dist );
        }
    }
    
    protected void stepFrame(Iterator i, Map movements) {
        fire();
        Log.info( "Movie at: " + startPos );
        double md = 0; double md2 = 0;
        MovieAtom ma, ma2;
        
        while ( i.hasNext() ) {
            MovieAtom a = (MovieAtom)i.next();
            Vector3d move = (Vector3d)movements.get( a );
            if ( move == null ) { continue; }
            Vector3d cur = new Vector3d(a.getLocation());
            cur.add( move );
            Log.info(  "Moving from " + a.getLocation() + " to " + cur );
            a.setLocation( cur, calcBonds );
            if ( a.getMovementSquared() > md ) { md = a.getMovementSquared(); ma = a; }
            if ( a.getMovementSquared() > md2 ) { md2 = a.getMovementSquared(); ma2 = a; }
        }
        
        molmaster.Log.info( "Maximum distance: " +  (md+md2) );
        if ( md + md2 > boundary ) {
            molmaster.Log.info( "Boundary skin exceeded - recalculating centers..." );
            i = atoms.iterator();
            while ( i.hasNext() ){
                MovieAtom a = (MovieAtom)i.next();
                a.updateCenter();
                a.updateBoundaryList( atoms, boundary );
            }
        }
        
    }
    /** Resets the state of the scene.  This simply moves every atom back to the
     * initial location in the file and recalculates the bonding.
     */
    protected void resetScene() {
        Iterator i = atoms.iterator();
        while( i.hasNext() ) {
            ((MovieAtom)i.next()).resetScene();
        }
    }
    
    /** Moves the movie to the given position.
     */
    public void setMoviePosition( int pos ) {
        if ( pos > getFrameCount() - 1 || pos < 0 ) { return; }
        if ( pos == currentPos ) { return; }
        
        currentPos = pos;
        resetScene();
        
        Map movements = new Hashtable();
        int startFrame = currentPos * movieBase.getFrameCount() / getFrameCount();
        if ( startFrame >= movieBase.getFrameCount() - 1) { startFrame--; }
        
        getFrameMovements( movements, startFrame, startFrame + 1 );
        
        fire();
        int startPos = startFrame * interpSteps;
        for ( int k = 0; k < currentPos - startPos; k++ ) {
            Iterator i = atoms.iterator();
            stepFrame(i, movements );
        }
    }
    
    /**
     * Notification that an image has been captured from the canvas and is
     * ready for processing.
     *
     * @param img The image that was captured
     * @return Whether this observer should remain in the listening queue.
     */
    public boolean canvasImageCaptured(BufferedImage img) {
        String outName = "";
        if ( currentFrameWrite < 10 ) {
            outName += "000" + currentFrameWrite;
        }
        else if ( currentFrameWrite < 100 ) {
            outName += "00" + currentFrameWrite;
        }
        else if ( currentFrameWrite < 1000 ) {
            outName += "0" + currentFrameWrite;
        }
        else {
            outName += "" + currentFrameWrite;
        }
        
        outName += ".bmp";
        
        File out = new File( destDir, outName );
        Log.info( "Writing to file: " + outName );
        
        BufferedImage tmp = new BufferedImage( img.getWidth(), img.getHeight(), BufferedImage.TYPE_3BYTE_BGR );
        java.awt.Graphics g = tmp.getGraphics();
        g.drawImage( img, 0, 0, new java.awt.Component() {} );
        g.dispose();
        
        img = tmp;
        try {
            FileOutputStream fs = new FileOutputStream( out );
            BufferedOutputStream bs = new BufferedOutputStream( fs );
            BMPImageEncoder enc = new BMPImageEncoder( bs, new BMPEncodeParam()  );
            enc.encode( img );
            bs.close();
            fs.close();
        } catch(Exception e) {
            Log.info( "Error writing out image to file." );
            Log.info( e );
        }
        
        Log.writeLog( "Wrote frame " + currentFrameWrite + " to file." );
        currentFrameWrite++;
        imageDone = true ;
        return false;
    }
    
    protected List listeners = new ArrayList();
    
    protected void fire() {
        Iterator i = listeners.iterator();
        while ( i.hasNext() ){
            ((Listener)i.next()).wakeUp( this, 0,  null );
        }
    }
    /** Adds the given listener to the notification list.
     */
    public void addListener(Listener l) {
        if ( !listeners.contains( l ) ) {
            listeners.add( l );
        }
    }
    
    /** Removes the given listener from the notification list.
     */
    public void removeListener(Listener l) {
        listeners.remove( l );
    }
    
}

class MovieAtom extends Atom {
    private List nearby = new ArrayList();
    private Atom delegate;
    private Vector3d original = new Vector3d();
    private Vector3d rOld = new Vector3d();
    
    private static final Vector3d tempVec = new Vector3d();
    
    MovieAtom( Atom me, List atoms, double boundaryDist  ) {
        super( me );
        delegate = me ;
        original.set( me.getLocation() );
        rOld.set( original );
        updateBoundaryList( atoms, boundaryDist );
    }
    
    public void updateCenter( Vector3d newCenter ) {
        rOld = newCenter;
    }
    public void updateCenter() {
        updateCenter( delegate.getLocation() );
    }
    
    public void updateBoundaryList(List atoms, double boundaryDist ) {
        Iterator i = atoms.iterator();
        while ( i.hasNext() ) {
            Atom a = (Atom)i.next();
            if ( a == delegate ) { continue; }
            Vector3d tempV = new Vector3d();
            tempV.sub( a.getLocation(), rOld );
            if ( tempV.lengthSquared() < boundaryDist ) {
                nearby.add( a );
            }
        }
    }
    
    public void resetScene() {
        setLocation( original );
        rOld.set( original );
        delegate.setLocation( original );
    }
    
    public double getMovementSquared() {
        tempVec.sub( location, original );
        return tempVec.lengthSquared();
    }
    
    public void setLocation( Vector3d loc, boolean calc ) {
        //if ( location.equals( loc ) ) { return; }
        delegate.setLocation( loc );
        location.set( loc );
        
        fire( MOVED, loc );
        
        if ( calc ) {
            for ( int i = 0; i < nearby.size(); i++ ) {
                //TODO - test hydrogen...
                Atom a = (Atom)nearby.get(i);
                boolean connect = connected.contains( a );
                //Log.writeLog( connected );
                Bond b = new Bond( delegate, a, Bond.BOND_TYPE_NORMAL );
                if ( b.isValid() ) {
                    if ( connect ) {
                        continue;
                    }
                    else {
                        Log.info( "Bonding " + delegate.getID() + " to " + a.getID() );
                        MolModel.getModel().connect( this, a, b);
                    }
                }
            }
        }
        
        Iterator it = bondList.iterator();
        while ( it.hasNext() ) {
            Bond b = (Bond)it.next();
            if ( !b.isValid() ) {
                it.remove();
                MolModel.getModel().removeBond( b );
            }
        }
        
        it = bondList.iterator();
        while ( it.hasNext() ){
            BondView bv = BondView.bondToView( it.next() );
            if ( bv != null )
                bv.refresh();
        }
    }
}
