package molmaster.gui.tools;

import java.awt.event.*;
import java.util.*;
import javax.swing.*;

import molmaster.*;
import molmaster.util.*;
import molmaster.gui.tools.reflect.*;

/** Displays options available for the user to edit while
 *  running molmaster.
 *
 * @author  Russell Power
 */
public class OptionWindow extends JDialog implements 
    javax.swing.event.ListSelectionListener, MouseMotionListener {
    
    /** Data class to hold the option categories.   */
    class CatGroup {
        String name;
        JPanel display;
        CatGroup(String s, JPanel p) { name = s; display = p; }
        public String toString() { return name; }
    }
    
    static public OptionWindow getWindow() {
        if ( myWindow == null ) { myWindow = new OptionWindow(); }
        return myWindow;
    }
    
    class ValueInfo {
        protected String group;
        protected String name;
        protected String description;
        protected String field;
        
        protected Object source;
        
        protected MethodValue methodValue;
        
        ValueInfo( String g, String n, String d, Object source, String field ) {
            this( g, n, d, source );
            this.field = field;
            try {
                Class ref = null;
                if ( source instanceof Class ) { ref = (Class)source; }
                else { ref = source.getClass(); }
                
                boolean hasMethod = false;
                try {
                    ref.getMethod( "get" + field, new Class[]{} );
                    hasMethod = true;
                }catch(Exception e){}
                if ( hasMethod ) {
                    setMethod( new MethodValue( source, field ) );
                }else {
                    setMethod( new MethodValue( source, field, true ) );
                }
                methodValue.setName(n);
            }catch(Exception e){
            	e.printStackTrace();
            }
        }
        
        ValueInfo( String g, String n, String d, Object s ) {
            group = g; name = n; description = d; source = s;
        }
        
        /** Sets the setter/getter pair for this value to the
         *  given instance.  Also searches the on-disk cache
         *  of saved entries to restore old values, if found.
         */
        protected void setMethod( MethodValue mv ) {
            methodValue = mv;
            Object o = optionStore.get( source + "#" + mv.getName() );
            if ( o != null ) { mv.set( o ); }
        }
    }
    
    /** Adds the given value to the list of editable
     *  properties available in the options dialog.  Properties
     *  added here are saved as a resource when the program is
     *  shutdown, and restored afterward.
     */
    public void addValue(
    String group,
    String name,
    String description,
    Object source,
    String fieldName ) {
        
        List l = (List)valueMap.get( group );
        if ( l == null ){
            l = new ArrayList();
            valueMap.put( group, l );
        }
        ValueInfo vi = new ValueInfo( group, name, description, source, fieldName );
        valueList.add( vi );
        l.add( vi );
        createGroups();
        
        Object o = optionStore.get( source.toString() + "#" + name );
        if ( o == null ) { return; }
        vi.methodValue.set( o );
    }
    
    /** Adds all setter/getter pairs found in the given
     *  source class to the given group.    */
    public void addValues( String group, Class methodSource ) {
        ReflectModel rm = new ReflectModel( methodSource, ReflectModel.SHOW_STATIC );
        List l = (List)valueMap.get( group );
        if ( l == null ){
            l = new ArrayList();
            valueMap.put( group, l );
        }
        for ( int i = 0; i < rm.getRowCount(); i++ ) {
            MethodValue mv = rm.getMethod( i, 1 );
            ValueInfo vi = new ValueInfo(group, mv.getName(), mv.getName(), methodSource );
            vi.setMethod( mv );       
            l.add( vi );
            valueList.add( vi );
        }
        createGroups();
    }
    
    /** Recreates the list and table views found in the value map. */
    protected void createGroups() {
        DefaultListModel m = new DefaultListModel();
        Iterator i = valueMap.entrySet().iterator();
        while ( i.hasNext() ) {
            Map.Entry e = (Map.Entry)i.next();
            m.addElement( e.getKey() );
        }
        categoryList.setModel( m );
    }
    
    
    protected ValueInfo getValue( Object owner, String name ) {
        Iterator i = valueList.iterator();
        while ( i.hasNext() ) {
            ValueInfo vi = (ValueInfo)i.next();
            if ( vi.name.equals( name ) && vi.source == owner ) {
                return vi;
            }
        }
        return null;
    }
    
    /** Creates new form OptionWindow */
    public OptionWindow() {
        super( molmaster.gui.MainFrame.getActiveWindow(), "molmaster Options", false );
        
        optionStore = new DataStore( "optionSave" );
        molmaster.gui.ShutdownHook.addShutdownListener( new OptionSave() );
        
        initComponents();
        categoryList.addListSelectionListener(this);
        
        addValue( "Bonding", "Hydrogen fudge factor",
        "The amount of leeway to allow when creating hydrogen bonds", Bond.class,
        "HydrogenAdjustmentFactor" );
        
        addValue( "Bonding",  "Fudge factor",
        "The amount of leeway to allow when creating normal bonds", Bond.class,
        "BondingFudgeFactor" );
        
        addValues( "Rendering", molmaster.gui.AtomView.class );
        
        optionTable = new ReflectTable( new Object(){} );
        optionTable.addMouseMotionListener( this );
        tableScroll.setViewportView( optionTable );
        
        setSize( 500, 500 );
        
        optionTable.addMouseMotionListener( this );
        optionTable.addMouseListener( new MouseAdapter() {
            public void mouseExited( MouseEvent e ) {
                commentBar.setText(" ");
            }
        });

    }
    
    private void initComponents() {//GEN-BEGIN:initComponents
        categoryList = new javax.swing.JList();
        tableScroll = new javax.swing.JScrollPane();
        commentBar = new javax.swing.JLabel();

        setDefaultCloseOperation(javax.swing.WindowConstants.DISPOSE_ON_CLOSE);
        categoryList.setBorder(new javax.swing.border.TitledBorder(new javax.swing.border.EtchedBorder(), "Option Categories"));
        categoryList.setSelectionMode(javax.swing.ListSelectionModel.SINGLE_SELECTION);
        categoryList.setToolTipText("Select the option category to view");
        categoryList.setPreferredSize(new java.awt.Dimension(150, 0));
        getContentPane().add(categoryList, java.awt.BorderLayout.WEST);

        getContentPane().add(tableScroll, java.awt.BorderLayout.CENTER);

        commentBar.setBackground(new java.awt.Color(133, 200, 233));
        commentBar.setText(" ");
        commentBar.setOpaque(true);
        getContentPane().add(commentBar, java.awt.BorderLayout.SOUTH);

        pack();
        java.awt.Dimension screenSize = java.awt.Toolkit.getDefaultToolkit().getScreenSize();
        setSize(new java.awt.Dimension(600, 400));
        setLocation((screenSize.width-600)/2,(screenSize.height-400)/2);
    }//GEN-END:initComponents
    
    public void valueChanged(javax.swing.event.ListSelectionEvent evt) {
        ReflectModel model = new ReflectModel( new Object() {} );
        
        Object o = categoryList.getSelectedValue();
        if ( o != null ) {
            List l = (List)valueMap.get( o );
            Iterator i = l.iterator();
            while ( i.hasNext() ) {
                ValueInfo vi = (ValueInfo)i.next();
                model.addRow( vi.methodValue );
            }
        }
        
        optionTable.setModel( model );
    }
    
    public void mouseDragged(java.awt.event.MouseEvent mouseEvent) {}

    /** Identifies the currently selected item and shows status bar
     *  text for the given item.
     */
    public void mouseMoved(java.awt.event.MouseEvent e) {
        int row = optionTable.rowAtPoint( e.getPoint() );
        if ( row == -1 ) { return; }
        
        MethodValue mv = ((ReflectModel)optionTable.getModel()).getMethod( row,  1 );
        ValueInfo vi = getValue( mv.getOwner(), mv.getName() );
        commentBar.setText( vi.name + "\t\t: " + vi.description );
    }
    
    /** Saves options to disk when the program closes.  These
     *  are then restored when the program resumes. 
     */
    class OptionSave implements Runnable {
        public void run() {
            Iterator i = valueList.iterator();
            while ( i.hasNext() ) {
                ValueInfo vi = (ValueInfo)i.next();   
                try {
                    optionStore.put(  
                    vi.source.toString() + "#" + vi.methodValue.getName(), 
                    (java.io.Serializable)vi.methodValue.get() );
                }catch(Exception e){
                    System.out.println( "Couldn't write " + vi.methodValue.getName() + " to disk." );
                    System.out.println(e);
                }
            }
            optionStore.flush();
        }
    }
    
    static private OptionWindow myWindow;
    private Map valueMap = new HashMap();
    private List valueList = new ArrayList();
    private DataStore optionStore;
    private ReflectTable optionTable;
    
    // Variables declaration - do not modify//GEN-BEGIN:variables
    private javax.swing.JList categoryList;
    private javax.swing.JLabel commentBar;
    private javax.swing.JScrollPane tableScroll;
    // End of variables declaration//GEN-END:variables
    
}
