package molmaster.gui.tools.reflect;

import molmaster.*;

import java.util.*;
import java.lang.reflect.*;

/** Encapsulates a reflected setter/getter pair
 *  and provides convenience methods for accessing
 *  the given methods.
 *
 * @author  Russell Power
 */
public class MethodValue implements Comparator {
    protected String suffixName;
    protected Class ownerType;
    
    protected Method setMethod;
    protected Method getMethod;
    protected Field myField;
    
    protected Object owner;
    protected Object fieldOwner;
    
    protected Object array;
    protected boolean isArray;
    protected int arrayIndex;
    
    static final Object[] empty = new Object[]{};
    
    /** Creates a blank method value that will operate on the
     *  given owner object.  No operations can be performed
     *  until the set and get methods are defined.
     */
    public MethodValue( Object owner ) {
        this.owner = owner;
    }
    
    private Class getClass( Object owner ) {
        if ( owner instanceof Class ) { return (Class)owner; }
        else { return owner.getClass(); }
    }
    
    /** Creates a Method setter/getter pair that matches with
     *  the given suffix.
     */
    public MethodValue( Object owner, String suffix ) {
        suffixName = suffix;
        this.owner = owner;
        
        Class ref = getClass( owner );
        Method[] m = ref.getMethods();
        for ( int i = 0; i < m.length; i++ ) {
            if ( m[i].getName().equals( "get" + suffix ) ) {
                getMethod = m[i];
                ownerType = getMethod.getReturnType();
            }
            if ( m[i].getName().equals( "set" + suffix ) ) {
                setMethod = m[i];
                ownerType = setMethod.getParameterTypes()[0];
            }
        }
    }
    
    public MethodValue( Object owner, String fieldName, boolean forceField ) 
    throws NoSuchMethodException, NoSuchFieldException {
            Class ref = getClass(owner);
            myField = ref.getField( fieldName );
            
            this.fieldOwner = owner;
            this.owner = this;
            
            ownerType = myField.getType();
            setMethod = getClass().getMethod( "fieldSet", new Class[] {Object.class} );
            getMethod = getClass().getMethod( "fieldGet", new Class[] {} );
            suffixName = fieldName;
    }
    
    public String getName() {
        return suffixName;
    }
    
    public void setName( String name ) {
        suffixName = name; 
    }
    
    public Object getOwner() { 
        if( fieldOwner != null ) { return fieldOwner; }
        return owner; 
    }
    
    public Object get() {
        try {
            if (!isArray ){
                return getMethod.invoke( owner, empty );
            }else {
                array = getMethod.invoke( owner,  empty );
                return Array.get( array, arrayIndex );
            }
        } catch(Exception e){
            Log.warn( "Couldn't get value from object." );
            Log.warn( e );
        }
        return null;
    }
    
    public void set(Object val){
        try {
            molmaster.Log.debug(  
            "Setting value for " + setMethod.getParameterTypes()[0] + " to " + val );
            if( !isArray ){
                setMethod.invoke( owner, new Object[]{ val } );
            }else {
                Array.set( array, arrayIndex,  val );
            }
        }
        catch(InvocationTargetException e) {
            if ( e.getTargetException() instanceof IllegalArgumentException ) {
                //new ErrorDialog( e, "Error setting owner...", true ).show();
                java.awt.Toolkit.getDefaultToolkit().beep();
            }
        }
        catch(Exception e){
            Log.warn( "Couldn't set object value." );
            Log.warn( e );
            //e.printStackTrace( Log.getLogStream() );
        }
    }
    
    public String toString(){
        Object o = get();
        if ( o != null ) { return o.toString(); }
        return "(null)";
    }
    
    public int compare(Object o1, Object o2) {
        MethodValue m1 = (MethodValue)o1;
        MethodValue m2 = (MethodValue)o2;
        return m1.suffixName.compareTo( m2.suffixName );
    }
    
    /** Wrapper around a field value to allow for the function
     *  of rest of the reflected model to remain oblivious to
     *  the fact that we are working with a field and not a
     *  setter/getter method pair.
     */
    public void fieldSet( Object o ) {
        try {
            myField.set( fieldOwner, o );
        }catch(Exception e){
            java.awt.Toolkit.getDefaultToolkit().beep();
            Log.warn(  "Couldn't set value - error was:\n\n" + e );
        }
    }
    
    /** Wrapper around a field value - implements a get function
     *  by returning the value of the field itself.
     */
    public Object fieldGet() {
        try {
            return myField.get( fieldOwner );
        }catch( Exception e ) {
            java.awt.Toolkit.getDefaultToolkit().beep();
            Log.warn(  "Couldn't get field value - error was:\n\n" + e );
        }
        return null;
    }
}
