package molmaster.gui.tools.reflect;

import molmaster.*;
import molmaster.gui.tools.reflect.parser.*;

import java.util.*;
import java.lang.reflect.*;

import javax.swing.table.*;

/** Got tired of implementing table stuff over and over and
 *  *over* again.  This table captures all public set and get
 *  methods and uses them to access and display table
 *  entries.
 *
 * @author  Russell
 */
public class ReflectModel extends DefaultTableModel {
    
    protected Object value;
    protected Map methodMap = new HashMap();
    protected Method refreshMethod = null;
    protected int methodTypes = SHOW_INSTANCE;
    
    protected List ignoreList = new ArrayList();
    
    /** If the showMethod var has this field
     *  enabled, then static setter and getter
     *  functions will be queried for information.
     */
    public static final int SHOW_STATIC =   0x01;
    
    /** Instance setter and getter functions will
     *  be queried if this field is active.
     */
    public static final int SHOW_INSTANCE = 0x02;
    
    /** Both static and instance methods will be
     *  visible if the model is created with this 
     *  flag.
     */
    public static final int SHOW_ALL    =   0x03;
    
    /** This flag determines whether *public* field values
     *  should be exposed.  This is by default off.
     */
    public static final int SHOW_FIELDS =   0x10;
    
    public ReflectModel() {}
    
    public ReflectModel(Object val) {
        this( val, SHOW_INSTANCE );
    }
    
    /** Initializes the model with the given object and flags.  The flag parameter
     * indicates what type of methods should be displayed - static, instance, or
     * both.
     */    
    public ReflectModel(Object val, int methodTypes) {
        super( new Object[][] {}, new Object[] {"Name", "Value"} );
        ignoreList.add( "Class" ); 
        
        this.methodTypes = methodTypes;
        value = val;
        if ( val == null ) {return;}
        
        scanObject();
        initTable();
    }
    
    /** Returns the current object that this model is operating on. */    
    public Object getCurrentObject() { return value; }
    
    /** Returns the setter/getter pair associated with a given
     * table position.
     */    
    public MethodValue getMethod( int row, int col ) {
        return (MethodValue)super.getValueAt( row,  col );
    }
    
    protected void initTable() {
        Collection c = methodMap.values();
        Iterator i = c.iterator();
        List l = new ArrayList();
        while ( i.hasNext() ) {
            l.add( i.next() );
        }
        Collections.sort( l, new MethodValue( null ) );
        
        i = l.iterator();
        
        while ( getRowCount() > 0 ) { removeRow(0); }
        while ( i.hasNext() ){
            MethodValue m = (MethodValue)i.next();
            Log.debug( "Inserting " + m.suffixName + " into table." );
            addRow( new Object[] { m.suffixName,  m } );
        }
    }
    
    /** Adds the given MethodValue to this table.   */
    public void addRow( MethodValue mv ) {
        methodMap.put( mv.getName(),  mv );
        initTable();
    }
    
    /** Extracts all available(public) setter/getter method
     *  pairs, as well as public fields for editing.
     */
    protected void scanObject() {
        Method[] methods = null;
        if ( !(value instanceof Class) ) {
            methods = value.getClass().getMethods();
        }else{
            methods = ((Class)value).getMethods();
        }
        try {
            Method m = value.getClass().getMethod( "reflectIgnore", new Class[] {} );
            List res = (List)m.invoke( value, new Object[] {} );
            ignoreList.addAll( res );
        }catch(Exception e) {}
        
        for ( int i = 0; i < methods.length; i++ ) {
            String methodName = methods[i].getName();
            int att = methods[i].getModifiers();
            
            if ( !Modifier.isPublic( att )) { continue; }
            if ( Modifier.isStatic( att ) ){
                if ( (methodTypes & SHOW_STATIC) == 0 ) { continue; }
            }
            if ( !Modifier.isStatic(att) ) {
                if ((methodTypes & SHOW_INSTANCE) == 0) { continue; }
            }
            
            String prefix = methodName.substring(0,3);
            String suffix = methodName.substring(3);
            
            if ( ignoreList.contains( suffix ) ) { continue; }
            
            Log.debug( "Found method: " + suffix );
            Log.debug( "Prefix: " + prefix );
            
            if ( prefix.equals("set") || prefix.equals("get") ){
                MethodValue m = new MethodValue(value, suffix);
                m.suffixName = suffix;
                methodMap.put( suffix, m );
                Log.debug( "Adding method: " + suffix );
            }
            if ( methodName.equals( "refresh" ) ) {
                refreshMethod = methods[i];
            }
        }
        
        
    }
    
    protected void refresh() {
        if ( refreshMethod == null ) { return; }
        try {
            refreshMethod.invoke( value, new Object[]{} );
        } catch( Exception e ){}
    }
    
    public boolean isCellEditable(int row, int col) {
        return (col == 0)?false:true;
    }
    
    /** Uses the getter methods extracted for the target objects class definition
     * to determine what values to load for this table location.
     */    
    public Object getValueAt( int row, int col ) {
        if ( col == 0 ) { return super.getValueAt( row, col ); }
        MethodValue m = (MethodValue)super.getValueAt( row, col );
        return m.get();
    }
    
    /** Uses the setter methods found on the target object to set the value at
     * the given location.  The input Object is given the the set**** method
     * of the target object.  If the object is of the incorrect type, no
     * change is made.
     */    
    public void setValueAt( Object val, int row, int col ) {
        MethodValue m = (MethodValue)super.getValueAt( row, col );
        if ( m == null )
            return;
        
        Object o = null;
        if ( val instanceof String && !(m.ownerType == String.class) ) {
            ValueParser vp = (ValueParser)ValueParser.parseMap.get( m.ownerType );
            
            if ( vp == null ) { return; }
            try {
                Log.debug( "Setting value to: " + val );
                o = vp.parse( (String)val );
            }catch( Exception e ){
                new molmaster.gui.tools.ErrorDialog( e, "Parse Error", true ).show();
                return;
            }
            
        }else{
            o = val;
        }
        
        m.set( o );
        fire( o );
        refresh();
    }
    
    public void addListener( molmaster.model.Listener l ) {
        listeners.add( l );
    }
    
    public void removeListener( molmaster.model.Listener l ) {
        listeners.remove( l );
    }
    protected void fire( Object data ) {
        Iterator i = listeners.iterator();
        while ( i.hasNext() ) {
            ((molmaster.model.Listener)i.next()).wakeUp( this, 0, data );
        }
    }

    protected ArrayList listeners = new ArrayList();
}