package molmaster.gui.tools.reflect.parser;

import java.util.*;
import javax.vecmath.*;

/**
 *
 * @author  Russell
 */
public abstract class ValueParser {
    /** Represents all types that can be parsed by the ReflectTable.  The table examines
     * this map when attempting to derive an object reference from the input string.
     * If a parser is found for the given class here, it is called to extract a usable
     * value to set.
     */    
    public static final java.util.Map parseMap = new java.util.HashMap();
    static {
        parseMap.put( javax.vecmath.Vector3d.class, new Vector3dParser() );
        parseMap.put( javax.vecmath.Vector3f.class, new Vector3fParser() );
        parseMap.put( javax.vecmath.Quat4f.class, new Quat4fParser() );
        parseMap.put( javax.vecmath.Quat4d.class, new Quat4dParser() );
        parseMap.put( javax.vecmath.AxisAngle4d.class, new AxisAngle4dParser() );
        parseMap.put( javax.vecmath.AxisAngle4f.class, new AxisAngle4fParser() );
        parseMap.put( javax.vecmath.Color3f.class, new ColorParser() );
        parseMap.put( Float.class, new FloatParser() );
        parseMap.put( Double.class, new DoubleParser() );
        parseMap.put( Integer.class, new IntegerParser() );
        parseMap.put( Boolean.class, new BooleanParser() );
        parseMap.put( Float.TYPE, new FloatParser() );
        parseMap.put( Double.TYPE, new DoubleParser() );
        parseMap.put( Integer.TYPE, new IntegerParser() );
        parseMap.put( Boolean.TYPE, new BooleanParser() );
        
        
    }
    /** Attempt to parse the given input string and resolve an object of the
     * type that this parser has been registered for.  If the input string
     * is invalid, throw an InvalidArgumentException.
     */    
    public abstract Object parse( String s ) throws IllegalArgumentException;

}

class Vector3fParser extends TupleParser {
    public Object parse(String s){
        Vector3d d = (Vector3d)super.parse(s);
        return new Vector3f( d );
    }
}
class Vector3dParser extends TupleParser {}
class AxisAngle4fParser extends Tuple4Parser {
    public Object parse(String s){
        AxisAngle4f p = new AxisAngle4f();
        p.set( (Quat4d)super.parse( s ) );
        return p;
    }
}
class AxisAngle4dParser extends Tuple4Parser {
    public Object parse(String s){
        AxisAngle4d p = new AxisAngle4d();
        p.set( (Quat4d)super.parse( s ) );
        return p;
    }
}

class Quat4fParser extends Tuple4Parser {
    public Object parse(String s){
        return new Quat4f( (Tuple4d)super.parse( s ) );
    }
}
class Quat4dParser extends Tuple4Parser {}
class Tuple4Parser extends ValueParser {
    public Object parse(String s){
        try{
            StringTokenizer tok = new StringTokenizer( s, "(,)" );
            double x= Double.parseDouble( tok.nextToken() );
            double y= Double.parseDouble( tok.nextToken() );
            double z= Double.parseDouble( tok.nextToken() );
            double w= Double.parseDouble( tok.nextToken() );
            return new Quat4d( x, y, z, w );
        }catch(Exception e){
            throw new IllegalArgumentException(
            "Invalid vector format.  Specify vector as: \"x,y,z,w\" ." );
        }
    }
}
class ColorParser extends TupleParser {
    public Object parse(String s ) throws IllegalArgumentException { 
        return new Color3f( (Tuple3d)super.parse(s) ); 
    }
}

class TupleParser extends ValueParser {
    public Object parse(String s) {
        try {
            String t = s;
            if ( s.indexOf( '(' ) != -1 )  {
                t = s.substring( s.indexOf( '(' ) + 1);
            }
            StringTokenizer tok = new StringTokenizer( s, "(,)" );
            double x = Double.parseDouble( tok.nextToken() );
            double y = Double.parseDouble( tok.nextToken() );
            double z = Double.parseDouble( tok.nextToken() );
            return new javax.vecmath.Vector3d( x, y, z );
        } catch (Exception e){
            throw new IllegalArgumentException(
            "Invalid vector format.  Specify vector as: \"x,y,z\" ." );
        }
    }
}

class IntegerParser extends ValueParser {
    public Object parse(String s){
        try {
            return new Integer(s);
        }catch(Exception e){
            throw new IllegalArgumentException(
            "Invalid number format." ); }
    }
}

class FloatParser extends ValueParser {
    public Object parse(String s){
        try {
            return new Float(s);
        }catch(Exception e){
            throw new IllegalArgumentException(
            "Invalid number format." ); }
    }
}

class DoubleParser extends ValueParser {
    public Object parse(String s){
        try {
            return new Double(s);
        }catch(Exception e){
            throw new IllegalArgumentException(
            "Invalid number format." ); }
    }
}

class BooleanParser extends ValueParser {
    public Object parse(String s){
        try {
            return new Boolean(s);
        }catch(Exception e){
            throw new IllegalArgumentException(
            "Invalid format - boolean values can assume 'true' or 'false'." ); }
    }
}