package molmaster.model;

import java.io.*;
import java.awt.*;
import java.util.*;
import javax.vecmath.*;
import javax.media.j3d.*;

import molmaster.*;

/** Reads a list of atoms from the given resource file,
 *  and presents them for use to the rest of the program.
 *
 *  Each atom in the table is referenced by it's atomic
 *  number.
 *
 * @author  Russell
 */
public class AtomTable {

	/** The location of the atom table on disk.
	 */
	protected String dbFile;
	/** Maps atom names to AtomAttribute objects.
	 */
	protected Map atomMap = new HashMap();
	/** Carries the geometry used by AtomView.
	 */
	protected Map geometryMap = new HashMap();
	/** Maps an atoms name to an Appearance.
	 */
	protected Map appearanceMap = new HashMap();

	/** The amount to scale input radaii by.
	 */
	static protected float radiusScale = 1.0f;
	/** The AtomTable exhibits the singleton pattern.  This is the only AtomTable object constructed.
	 */
	static protected AtomTable activeTable;
	/** The appearance used by selected objects.
	 */
	static public Appearance selectedAppearance;
	/** The default appearance applied to objects.
	 */
	protected Appearance defaultAppearance = new Appearance();

	/** Creates an instance of the AtomTable using the default location for the
	 * table on disk.
	 */
	public AtomTable() {
		dbFile = "elements.dat";
		loadEntries();
		init();
	}

	protected void init() {
		activeTable = this;
		selectedAppearance = new Appearance();
		Material selectedMaterial = new Material();
		selectedMaterial.setAmbientColor(0.8f, 0.1f, 0.1f);
		selectedAppearance.setMaterial(selectedMaterial);

	}
	/** Creates an AtomTable using the specified disk file.
	 */
	public AtomTable(String fileName) {
		init();
		//ClassLoader.getSystemResourceAsStream( fileName );
		//dbFile = Molmaster.sourceDir.getAbsolutePath() + "/" + fileName;//new File(fileName);
		dbFile = fileName; //getClass().getResource(fileName);
		loadEntries();
		activeTable = this;
	}

	/** Loads (or reloads) all the entries for this table from
	 *  disk.  Each entry in the table is associated with size
	 *  and color information for a particular atom.    */
	public void loadEntries() {
		try {
			//Reader input = new BufferedReader( new FileReader( dbFile ) );
			Log.writeLog("Reading table from: " + dbFile);
			Reader input =
				new BufferedReader(
					new InputStreamReader(
						ClassLoader.getSystemResourceAsStream(dbFile)));

			StreamTokenizer tokenizer = new StreamTokenizer(input);
			readInput(tokenizer);
			input.close();
		} catch (Exception e) {
			e.printStackTrace();
		}

	}

	/** Reads the set of tokens given by the StreamTokenizer into
	 *  the AtomTable.  */
	protected void readInput(StreamTokenizer tokens) {
		try {
			int tokenType;
			tokens.eolIsSignificant(true);
			while (tokens.ttype != StreamTokenizer.TT_EOL) {
				tokens.nextToken();
			}
			tokens.eolIsSignificant(false);
			while (tokens.ttype != StreamTokenizer.TT_EOF) {
				AtomAttributes at = new AtomAttributes();

				tokens.nextToken();
				if (tokens.sval == null) {
					break;
				}
				at.atomName = tokens.sval.toLowerCase();

				tokens.nextToken();
				at.atomicSymbol = tokens.sval;

				tokens.nextToken();
				at.atomicNumber = (int) tokens.nval;

				tokens.nextToken();
				at.atomicMass = tokens.nval;

				tokens.nextToken();
				at.vanDerWaal = tokens.nval;

				tokens.nextToken();
				at.bondingRadius = tokens.nval;

				at.radius = at.bondingRadius * radiusScale;

				float c1, c2, c3;
				tokens.nextToken();
				c1 = (float) tokens.nval / 255;
				tokens.nextToken();
				c2 = (float) tokens.nval / 255;
				tokens.nextToken();
				c3 = (float) tokens.nval / 255;

				at.color = new javax.vecmath.Color3f(c1, c2, c3);

				atomMap.put(at.atomName, at);
				if (!atomMap.containsKey(at.atomicSymbol)) {
					atomMap.put(at.atomicSymbol, at);
				}
				appearanceMap.put(at.atomName, getAppearance(at.color));
			}
			Log.writeLog(
				"Read in " + appearanceMap.size() + " entries from disk.");
		} catch (Exception e) {
			System.out.println("An error occured while processing entries...");
			//e.printStackTrace();
		}
	}

	/** Returns an Appearance object using the given color consistent with
	 * application to an AtomView object.
	 */
	public Appearance getAppearance(Color3f c) {
		Appearance a = new Appearance();

		Material m = new Material();
		m.setAmbientColor(c);
		m.setDiffuseColor(c);
		m.setSpecularColor(1.0f, 1.0f, 1.0f);
		m.setShininess(40.0f);
		m.setLightingEnable(true);

		a.setMaterial(m);

		return a;
	}

	int unknownCount = 0;
	/** Returns a new instance of the atom given by atomName.   */
	public Atom getAtom(String atomName) {
		if (atomName == null || atomName.length() == 0) {
			atomName = "Unnamed[" + unknownCount + "]";
			unknownCount++;
		}

		AtomAttributes a = (AtomAttributes) atomMap.get(atomName.toLowerCase());

		if (a == null) {
			System.out.println(
				"Couldn't find attributes in table for type: " + atomName);
			Atom at = createNewAtom(atomName.toLowerCase());
			atomMap.put(atomName.toLowerCase(), at.getAttributes());
			appearanceMap.put(
				atomName.toLowerCase(),
				getAppearance(at.getAttributes().color));
			return at;
		}

		return new Atom(a);
	}

	/** Returns a list containing the name of each atom
	 * found in the file.
	 */
	public Collection getAtomNames() {
		return atomMap.keySet();
	}

	/** Returns the list of all attributes derived from the table.
	 */
	public Collection getAtomAttributes() {
		return atomMap.values();
	}

	/** Returns the default appearance for the given atom,
	 *  as found in the resource file.  If the given atom type
	 *  was not found in the file, this returns a default appearance.
	 */
	public Appearance getAppearance(Atom a) {
		Appearance ap =
			(Appearance) appearanceMap.get(
				((AtomAttributes) a.getAttributes()).atomName);
		if (ap != null) {
			return ap;
		}
		//System.out.println("Couldn't find entry for specified atom.");
		return defaultAppearance;
	}

	/** Listing of colors available for use by types not recognized in the current
	 * table.
	 */
	protected static Color[] newAtomColors =
		new Color[] {
			Color.blue,
			Color.green,
			Color.orange,
			Color.magenta,
			Color.cyan };
	/** The current color to be used when generating a new Appearance for an unknown
	 * type.
	 */
	protected int currentColor = 0;

	/** Creates a new atom of the given unknown type.
	 */
	protected Atom createNewAtom(String type) {
		Color c = newAtomColors[((currentColor++) % newAtomColors.length)];
		AtomAttributes aa = new AtomAttributes();
		aa.color = new Color3f(c);
		aa.atomName = type;
		aa.atomicSymbol = "" + type.charAt(0);
		aa.atomicNumber = 1000;
		return new Atom(aa);
	}

	/** Returns an instance of the AtomTable.
	 */
	public static AtomTable getActive() {
		return activeTable;
	}
}