package molmaster.model;

import molmaster.*;
import java.util.*;

/** Represents the current scene depicted by molmaster.  Controls
 *  addition and removal of files, atoms and bonds to the scene.
 *
 * @author  Russell
 */
public class MolModel {
    
    protected static MolModel theModel = new MolModel();
    /** Returns an instance of the current MolModel.
     */    
    static public MolModel getModel() { return theModel; }
    
    protected Map atomTypes = new HashMap();
    protected Set atomSet = new HashSet();
    protected Set bonds = new HashSet();
    protected List files = new ArrayList();
    protected Set listeners = new HashSet();
    protected Set endListeners = new HashSet();
    
    
    protected MolModel() {}
    
    
    protected int uniqueID = 0;
    /** Returns a unique identifier.  This value is valid for the duration of the
     * program.
     */    
    public int getUniqueID() {
        return uniqueID++;
    }
    
    /** Adds an atom to the scene.
     */    
    public void addAtom( Atom a ){
        atomSet.add( a );
        String name = ((AtomAttributes)a.getAttributes()).atomName;
        incrementType( name, 1 );
        update( a, ModelListener.ATOM_ADDED );
        
    }
    
    /** Removes the given Atom from the scene.  If the Atom is not present in the scene
     * there is no effect.
     */    
    public void removeAtom( Atom a ){
        atomSet.remove(a);
        a.getOwner().getAtoms().remove( a );
        String name = ((AtomAttributes)a.getAttributes()).atomName;
        incrementType( name, -1 );
        
        Iterator i = new ArrayList(a.getBonds()).iterator();
        while ( i.hasNext() ) {
            removeBond( (Bond)i.next() );
        }
        
        update( a, ModelListener.ATOM_REMOVED );
    }
    
    /** Adds the given Bond into the scene, connecting the two given atoms and notifying
     * listeners of the change.
     */    
    public void connect( Atom a, Atom b, Bond bond ){
        updateConnections( bond );
        update( bond, ModelListener.BOND_ADDED);
    }
    
    private void updateConnections(Bond bond){
        if ( !bonds.contains( bond ) ) {
            bonds.add( bond );
        }
        
        Atom a = bond.getStart();
        Atom b = bond.getEnd();
        a.addBond( bond );
        b.addBond( bond );
        
    }
    
    protected void disconnect( Atom a, Atom b, Bond bond ) {
        bonds.remove( bond );
        a.removeBond( bond );
        b.removeBond( bond );
    }
    
    /** Returns all Atoms currently connected via Bonds to the given Atom.
     */    
    public Collection getConnected(Atom a){
        return a.getConnected();
    }
    
    /** Returns all Atoms currently present in the scene.
     */    
    public Collection getAtoms() {
        return atomSet;
    }
    
    /** Returns all Bonds present in the scene.
     */    
    public Collection getBonds() {
        return bonds;
    }
    
    /** Returns true if the scene contains the given Atom.
     */    
    public boolean contains( Atom a ) {
        return atomSet.contains(a);
    }
    
    /** Called in response to a change in the model's state.
     */    
    protected void update() {
        update( null, ModelListener.MODEL_UPDATE );
    }
    
    /** Removes the given Bond from the scene.
     */    
    public void removeBond( Bond b ) {
        disconnect( b.getStart(), b.getEnd(), b );
        update( b, ModelListener.BOND_REMOVED );
    }
    private boolean updating = false;
    
    protected void update( Object value, int updateType ){
        if ( updating ) { return; }
        updating = true;
        Iterator i = listeners.iterator();
        while ( i.hasNext() ) {
            ((ModelListener)i.next()).modelUpdated(this, updateType, value);
        }
        
        i = endListeners.iterator();
        while ( i.hasNext() ) {
            ((ModelListener)i.next()).modelUpdated(this, updateType, value);
        }
        updating = false;
    }
    
    /** Adds the given listener to the notification list.  All listeners will be
     * called when a change occurs to the model's state.
     */    
    public void addListener( ModelListener l ) {
        addListener( l, false );
    }
    
    /** Listeners added with atEnd set to true will be
     *  called after all regular listeners. */
    public void addListener( ModelListener l, boolean atEnd ) {
        if ( atEnd == true ) {
            endListeners.add( l );
            return;
        }
        listeners.add(l);
    }
    
    /** Returns all files that have been added to the scene.
     */    
    public List getFiles() {
        return files;
    }
    
    /** Removes the given listener from the notification list.
     */    
    public void removeListener( ModelListener l ) {
        listeners.remove( l );
    }
    
    /** Adds the given group of bonds to the scene.
     */    
    public void addBondGroup( Collection l ) {
        Iterator i = l.iterator();
        while ( i.hasNext() ) {
            Bond bond = (Bond)i.next();
            updateConnections( bond );
        }
        update( l, ModelListener.BOND_GROUP_ADDED );
    }
    
    /** Adds the given file to the scene.  The file is queried for it's list of Atoms,
     * which are then added.
     */    
    public void addAtomGroup( MoleculeFile c ) {
        files.add( c );
        Iterator i = c.getAtoms().iterator();
        while ( i.hasNext() ) {
            Atom a = (Atom)i.next();
            String name = ((AtomAttributes)a.getAttributes()).atomName;
            atomSet.add(a);
            incrementType( name, 1 );
        }
        
        update( c, ModelListener.FILE_ADDED );
    }
    protected void incrementType( String name, int increment ) {
        Integer num = (Integer)atomTypes.get( name );
        if ( num == null ) {
            num = new Integer(0);
        }
        num = new Integer( num.intValue() + increment );
        if ( num.intValue() == 0 ) {
            atomTypes.remove( name );
            return;
        }
        atomTypes.put( name, num );
    }
    
    /** Removes the given file from the scene.
     */    
    public void removeAtomGroup( MoleculeFile c ) {
        files.remove( c );
        Iterator i = c.getAtoms().iterator();
        while ( i.hasNext() ) {
            Atom a = (Atom)i.next();
            atomSet.remove(a);
            //a.getOwner().getAtoms().remove( a );
            incrementType( ((AtomAttributes)a.getAttributes()).atomName, -1 );
        }
        
        update( c, ModelListener.FILE_REMOVED );
    }
    
    /** Removes the given list of bonds from the scene.
     */    
    public void removeBondGroup( Collection c ){
        Iterator i = c.iterator();
        while ( i.hasNext() ) {
            removeBond( (Bond)i.next() );
        }
    }
    
    /** Returns an array consisting of all Atoms that have atomic names or symbols
     * identical to the input.
     */    
    public Atom[] getAtomsOfType( String atomType ){
        Iterator i = atomSet.iterator();
        HashSet out = new HashSet();
        
        while ( i.hasNext() ) {
            Atom a = (Atom)i.next();
            AtomAttributes attrib = (AtomAttributes)a.getAttributes();
            if ( attrib.atomName.equals( atomType ) || attrib.atomicSymbol.equals( atomType ) ){
                out.add( a );
            }
        }
        
        Object[] o = out.toArray();
        Atom[] atomOut = new Atom[o.length];
        for ( int j = 0; j < o.length; j++ ){
            atomOut[j] = (Atom)o[j];
        }
        
        return atomOut;
    }
    
    /** Returns a list of Strings indicating each type of Atom currently present in
     * the scene.
     */    
    public Collection getAtomTypes() {
        return atomTypes.keySet();
    }
    
}
