package molmaster.model;

import molmaster.*;
import molmaster.model.loaders.*;

import javax.vecmath.*;
import javax.media.j3d.*;

import java.io.*;
import java.util.*;

/** Base class for all file formats describing
 *  molecules.
 *
 * @author  rjpower
 */
public abstract class MoleculeFile {

	protected Collection atoms = new HashSet();
	protected Collection bonds = new HashSet();
	protected String myName;
	protected File diskFile;

	static protected Map loaderTypes = new HashMap();

	static {
		loaderTypes.put("con", ConLoader.class);
		loaderTypes.put("xyz", XYZLoader.class);
		loaderTypes.put("mol", MolLoader.class);
		loaderTypes.put("vasp", VaspLoader.class);
		loaderTypes.put("chgcar", ChgCarLoader.class);
		loaderTypes.put("pdb", PDBLoader.class);
		loaderTypes.put("psi", PsiLoader.class);
		loaderTypes.put("ps", PostscriptWriter.class);
		loaderTypes.put("", MaxRhoLoader.class);
	}

	/** Returns a loader for the given filename.  This loader is based initially on the
	 * file suffix - if a file suffix matches that recognized by a loader, the
	 * approriate loader is returned.  Otherwise a dialog is shown prompting the user
	 * to select the correct loader type.
	 */
	static public MoleculeFile getLoader(String name) throws Exception {
		try {
			String ext = null;
			if (name.length() < 2) {
				ext = name;
				ext = ext.toLowerCase();
				Class c = (Class) loaderTypes.get(ext);
				MoleculeFile mf = (MoleculeFile) c.newInstance();
				return mf;
			}
			for (int extSize = 2; extSize < 6; extSize++) {
				ext = name.substring(name.length() - extSize);
				ext = ext.toLowerCase();
				Class c = (Class) loaderTypes.get(ext);
				if (c != null) {
					MoleculeFile mf = (MoleculeFile) c.newInstance();
					if (mf != null) {
						return mf;
					}
				}
			}
			throw new Exception("Couldn't find loader...");
		} catch (Exception e) {
			MultiLoader m =
				new MultiLoader(
					molmaster.gui.MainFrame.getActiveWindow(),
					true);
			m.setVisible(true);
			Class c = m.getLoader();
			if (c == null) {
				throw new Exception("File load canceled.");
			}
			try {
				MoleculeFile mf = (MoleculeFile) c.newInstance();
				return mf;
			} catch (Exception ex) {
				throw new Exception("Couldn't load file.");
			}
		}
	}

	/** Informs the display table to ignore the given 
	 *  reflected values.   
	 */
	public List reflectIgnore() {
		return Arrays.asList(
			new String[] {
				"Atoms",
				"Bonds",
				"Bounds",
				"BoxSize",
				"BoundingBox",
				"UserData",
				"AdditionalData" });
	}
	/** Returns a friendly name for the given file.  By default this is diskFile.getName().
	 */
	public String getName() {
		if (myName == null) {
			myName = diskFile.getName();
		}
		return myName;
	}

	/** Remove all atoms and bonds associated with this
	 *  file from the scene.
	 */
	public void removeFromScene(molmaster.model.MolModel scene) {
		scene.removeAtomGroup(this);
	}

	/** Add all atoms and bonds from this file to the current
	 *  scene.  */
	public void addToScene(molmaster.model.MolModel scene) {
		calcBox();
		scene.addAtomGroup(this);
		scene.addBondGroup(bonds);
	}

	/** Returns a Collection of atoms found in the file.    */
	public Collection getAtoms() {
		return atoms;
	}

	/** Returns a Collection of bonds found in the file.    */
	public Collection getBonds() {
		return bonds;
	}

	/** Save the current contents of this resource to
	 *  the filename specified. */
	public void saveToFile(String fileName) throws java.io.IOException {
		throw new java.io.IOException("This operation not supported.");
	}

	/** Load all information from the given file.  This should be
	 *  done in the current thread - if a user needs to perform
	 *  other processing while this takes place, they should spawn
	 *  a new thread before calling this function.
	 */
	public abstract void loadFromFile(String fileName)
		throws java.io.IOException;

	protected BranchGroup rootGroup = new javax.media.j3d.BranchGroup();
	protected TransformGroup boxTrans = new TransformGroup();

	/** Returns any additional structures that might be
	 *  present in the file.    */
	public javax.media.j3d.BranchGroup getAdditionalData() {
		return rootGroup;
	}

	/** Return a string representation of the current file.  By default this is
	 * this.getName().
	 */
	public String toString() {
		return getName();
	}

	protected float scaleFactor = 1.0f;
	/** Returns the current value Atoms from this file are scaled by.
	 */
	public float getScaleFactor() {
		return scaleFactor;
	}
	/** Set the amount to scale this files Atoms.  Values greater then 1 increase the
	 * radius of all Atoms contained within this file, value less than 1 decrease it.
	 */
	public void setScaleFactor(float f) {
		scaleFactor = f;
		Iterator i = atoms.iterator();
		while (i.hasNext()) {
			molmaster.gui.AtomView.atomToView((Atom) i.next()).setScaleFactor(
				f);
		}
		i = MolModel.getModel().getBonds().iterator();
		while (i.hasNext()) {
			Bond b = (Bond) i.next();
			if (b.getOwner() == this) {
				molmaster.gui.BondView.bondToView(b).setScaleFactor(f);
			}
		}
	}

	protected void skipLine(StreamTokenizer tok) throws IOException {
		tok.eolIsSignificant(true);
		int count;

		while (tok.ttype != StreamTokenizer.TT_EOF
			&& tok.ttype != StreamTokenizer.TT_EOL) {
			tok.nextToken();
		}
		tok.nextToken();
		tok.eolIsSignificant(false);
		tok.pushBack();
	}

	StringBuffer name = new StringBuffer(10);
	protected double getNumber(Reader r, double start)
		throws java.io.IOException {
		int next;
		name.setLength(0);

		while (true) {
			next = r.read();
			if ((char) next == '\n' || (char) next == ' ' || next == -1) {
				break;
			}
			name.append((char) next);
		}

		String s = name.toString();
		int mx = s.indexOf('e');
		if (s.indexOf('+') > mx) {
			mx = s.indexOf('+');
		}
		if (s.indexOf('E') > mx) {
			mx = s.indexOf('E');
		}
		if (mx < 0) {
			return start;
		}
		String sub = s.substring(mx + 1);
		return Math.pow(10, Integer.parseInt(sub)) * start;
	}

	protected String getString(BufferedReader br) throws java.io.IOException {
		StringBuffer buf = new StringBuffer();
		int next = 0;
		while ((next = br.read()) != -1) {
			if (Character.isWhitespace((char) next)) {
				if (buf.length() > 0) {
					break;
				}
			} else {
				if (Character.isLetterOrDigit((char) next)) {
					buf.append((char) next);
				}
			}
		}
		return buf.toString();
	}

	protected double getNumber(StreamTokenizer tok)
		throws java.io.IOException {
		return tok.nval;
	}

	protected double getFixed(Reader r) throws java.io.IOException {
		double base = 0;
		int exp = 0;

		name.setLength(0);
		int nl = 0;
		while ((nl = r.read()) != -1) {
			if (Character.isWhitespace((char) nl)) {
				continue;
			}
			break;
		}

		name.append((char) nl);

		while ((nl = r.read()) != -1) {
			if (Character.isDigit((char) nl) || (nl == '.')) {
				name.append((char) nl);
			} else {
				break;
			}
		}

		base = Double.parseDouble(name.toString());
		if (nl == 'e' || nl == 'E') {
			name.setLength(0);
			//r.read();    // E
			nl = r.read(); // +-
			if (nl != '+') {
				name.append((char) nl);
			}

			nl = r.read();
			while (Character.isDigit((char) nl)) {
				name.append((char) nl);
				nl = r.read();
			}
			exp = Integer.parseInt(name.toString());
			return base * Math.pow(10, exp);
		}

		//Log.writeLog("" + base);
		return base;
	}

	protected Object userData;

	/** Sets the user object associated with this file.
	 */
	public void setUserData(Object o) {
		userData = o;
	}
	/** Returns a reference to the user object held by this file.
	 */
	public Object getUserData() {
		return userData;
	}

	/** Returns the disk file this loader obtained it's data from.
	 */
	public File getDiskFile() {
		return diskFile;
	}

	public BoundingBox getBounds() {
		calcBox();
		return new BoundingBox(
			new Point3d(mx, my, mz),
			new Point3d(sx, sy, sz));
	}

	public double[] getBoxSize() {
		return new double[] { mx - sx, my - sy, mz - sz };
	}

	protected void calcBox() {

		mx = my = mz = Double.POSITIVE_INFINITY;
		sx = sy = sz = Double.NEGATIVE_INFINITY;
		Iterator i = atoms.iterator();
		while (i.hasNext()) {
			Vector3d loc = ((Atom) i.next()).getLocation();
			if (loc.x > sx) {
				sx = loc.x;
			}
			if (loc.y > sy) {
				sy = loc.y;
			}
			if (loc.z > sz) {
				sz = loc.z;
			}
			if (loc.x < mx) {
				mx = loc.x;
			}
			if (loc.y < my) {
				my = loc.y;
			}
			if (loc.z < mz) {
				mz = loc.z;
			}
		}

		if (!boxFinished) {
			makeBox(
				(float) mx,
				(float) my,
				(float) mz,
				(float) sx,
				(float) sy,
				(float) sz);
		}
	}

	private double sx, sy, sz;
	private double mx, my, mz;
	private boolean boxFinished = false;

	protected void makeBox(float sx, float sy, float sz) {
		makeBox(-sx, -sy, -sz, sx, sy, sz);
	}

	protected void makeBox(
		float mx,
		float my,
		float mz,
		float sx,
		float sy,
		float sz) {
		boxFinished = true;
		GeometryArray ga = new QuadArray(24, QuadArray.COORDINATES);
		//| QuadArray.NORMALS);
		float[] coords = new float[72];

		//  front face
		coords[0] = sx;
		coords[1] = my;
		coords[2] = sz;
		coords[3] = sx;
		coords[4] = sy;
		coords[5] = sz;
		coords[6] = mx;
		coords[7] = sy;
		coords[8] = sz;
		coords[9] = mx;
		coords[10] = my;
		coords[11] = sz;

		// left face
		coords[12] = mx;
		coords[13] = my;
		coords[14] = sz;
		coords[15] = mx;
		coords[16] = sy;
		coords[17] = sz;
		coords[18] = mx;
		coords[19] = sy;
		coords[20] = mz;
		coords[21] = mx;
		coords[22] = my;
		coords[23] = mz;

		// back face
		coords[24] = mx;
		coords[25] = my;
		coords[26] = mz;
		coords[27] = mx;
		coords[28] = sy;
		coords[29] = mz;
		coords[30] = sx;
		coords[31] = sy;
		coords[32] = mz;
		coords[33] = sx;
		coords[34] = my;
		coords[35] = mz;

		// right face
		coords[36] = sx;
		coords[37] = my;
		coords[38] = mz;
		coords[39] = sx;
		coords[40] = sy;
		coords[41] = mz;
		coords[42] = sx;
		coords[43] = sy;
		coords[44] = sz;
		coords[45] = sx;
		coords[46] = my;
		coords[47] = sz;

		// top face
		coords[48] = sx;
		coords[49] = sy;
		coords[50] = sz;
		coords[51] = sx;
		coords[52] = sy;
		coords[53] = mz;
		coords[54] = mx;
		coords[55] = sy;
		coords[56] = mz;
		coords[57] = mx;
		coords[58] = sy;
		coords[59] = sz;

		// bottom face
		coords[60] = mx;
		coords[61] = my;
		coords[62] = mz;
		coords[63] = sx;
		coords[64] = my;
		coords[65] = mz;
		coords[66] = sx;
		coords[67] = my;
		coords[68] = sz;
		coords[69] = mx;
		coords[70] = my;
		coords[71] = sz;

		float[] normals = new float[72];

		// front face
		normals[0] = 0;
		normals[1] = 0;
		normals[2] = 1;
		normals[3] = 0;
		normals[4] = 0;
		normals[5] = 1;
		normals[6] = 0;
		normals[7] = 0;
		normals[8] = 1;
		normals[9] = 0;
		normals[10] = 0;
		normals[11] = 1;

		// left face
		normals[12] = -1;
		normals[13] = 0;
		normals[14] = 0;
		normals[15] = -1;
		normals[16] = 0;
		normals[17] = 0;
		normals[18] = -1;
		normals[19] = 0;
		normals[20] = 0;
		normals[21] = -1;
		normals[22] = 0;
		normals[23] = 0;

		// back face
		normals[24] = 0;
		normals[25] = 0;
		normals[26] = -1;
		normals[27] = 0;
		normals[28] = 0;
		normals[29] = -1;
		normals[30] = 0;
		normals[31] = 0;
		normals[32] = -1;
		normals[33] = 0;
		normals[34] = 0;
		normals[35] = -1;

		// right face
		normals[36] = 1;
		normals[37] = 0;
		normals[38] = 0;
		normals[39] = 1;
		normals[40] = 0;
		normals[41] = 0;
		normals[42] = 1;
		normals[43] = 0;
		normals[44] = 0;
		normals[45] = 1;
		normals[46] = 0;
		normals[47] = 0;

		// top face
		normals[48] = 0;
		normals[49] = 1;
		normals[50] = 0;
		normals[51] = 0;
		normals[52] = 1;
		normals[53] = 0;
		normals[54] = 0;
		normals[55] = 1;
		normals[56] = 0;
		normals[57] = 0;
		normals[58] = 1;
		normals[59] = 0;

		// bottom face
		normals[60] = 0;
		normals[61] = -1;
		normals[62] = 0;
		normals[63] = 0;
		normals[64] = -1;
		normals[65] = 0;
		normals[66] = 0;
		normals[67] = -1;
		normals[68] = 0;
		normals[69] = 0;
		normals[70] = -1;
		normals[71] = 0;

		ga.setCoordinates(0, coords);
		//ga.setNormals( 0, normals );

		//  generate view box
		Appearance boxAppearance = new Appearance();
		PolygonAttributes pa = new PolygonAttributes();
		ColoringAttributes coloringAttr = new ColoringAttributes();
		coloringAttr.setColor(0.0f, 0.0f, 0.0f);
		coloringAttr.setShadeModel(ColoringAttributes.FASTEST);

		pa.setPolygonMode(PolygonAttributes.POLYGON_LINE);
		pa.setBackFaceNormalFlip(true);
		pa.setCullFace(PolygonAttributes.CULL_NONE);

		LineAttributes la = new LineAttributes();
		la.setLineAntialiasingEnable(true);

		boxAppearance.setLineAttributes(la);
		boxAppearance.setPolygonAttributes(pa);
		boxAppearance.setColoringAttributes(coloringAttr);

		Shape3D b = new Shape3D(ga);
		//new Box( (float)viewBox[0], (float)viewBox[1], (float)viewBox[2], boxAppearance );//Shape3D( ga );
		b.setPickable(false);
		b.setAppearance(boxAppearance);

		boxAppearance = new Appearance();
		pa = new PolygonAttributes();
		pa.setPolygonMode(PolygonAttributes.POLYGON_POINT);
		pa.setCullFace(PolygonAttributes.CULL_NONE);

		boxAppearance.setPolygonAttributes(pa);

		PointAttributes pointAttr = new PointAttributes(5.0f, true);
		boxAppearance.setPointAttributes(pointAttr);
		boxAppearance.setColoringAttributes(coloringAttr);

		Shape3D points = new Shape3D(ga);
		points.setAppearance(boxAppearance);
		points.setPickable(false);

		boxTrans.setCapability(TransformGroup.ALLOW_TRANSFORM_WRITE);
		boxTrans.setCapability(TransformGroup.ALLOW_TRANSFORM_READ);
		boxTrans.clearCapabilityIsFrequent(TransformGroup.ALLOW_TRANSFORM_READ);
		boxTrans.clearCapabilityIsFrequent(TransformGroup.ALLOW_TRANSFORM_WRITE);

		boxTrans.addChild(b);
		boxTrans.addChild(points);
		BranchGroup boxGroup = new BranchGroup();
		boxGroup.addChild(boxTrans);
		rootGroup.addChild(boxGroup);
		rootGroup.setCapability(BranchGroup.ALLOW_DETACH);
	}
}
