package molmaster.model.loaders;

import molmaster.*;
import molmaster.model.*;

import java.io.*;
import java.util.*;
import javax.vecmath.*;

/** Implements a simple loader for a subset of the PDB( protein data bank )
 *  file format.  This loader currently supports the reading of basic atom
 *  and bond information.  The loader supports the following PDB records:
 *  <code>
 *
 *  HEADER
 *  ATOM
 *  HETATM
 *  CONECT
 *  CRYST1
 *  SCALE
 *
 */
public class PDBLoader extends MoleculeFile implements CubeFile {
    
    protected double[] boxSize;
    protected Vector3d[] scaleFactors = new Vector3d[]{
        new Vector3d( 1, 0, 0 ),
        new Vector3d( 0, 1, 0 ),
        new Vector3d( 0, 0, 1 )
    };
    
    
    public PDBLoader() {}
    
    public void loadFromFile( String fileName ) throws java.io.IOException {
        diskFile = new File( fileName );
        FileReader reader = new FileReader( diskFile );
        BufferedReader input = new BufferedReader( reader );
        Map atomMap = new HashMap();
        
        String lineIn = null;
        while ( (lineIn = input.readLine()) != null  ){
            if ( lineIn.startsWith( "HEADER" ) ){
                myName = lineIn.substring( 6, 47 );
            }
            else if ( lineIn.startsWith( "ATOM" ) || lineIn.startsWith( "HETATM" ) ){
                String atomType = lineIn.substring( 12, 16).trim();
                //System.out.println(atomType);
                Atom a = AtomTable.getActive().getAtom( atomType );
                if ( ((AtomAttributes)a.getAttributes()).atomName == null ) {
                    System.out.println( a );
                }
                a.setID( Integer.parseInt( lineIn.substring( 6,12 ).trim() ) );
                Vector3d location = new Vector3d();
                location.x = Double.parseDouble( lineIn.substring( 30, 38 ).trim() );
                location.y = Double.parseDouble( lineIn.substring( 38, 46 ).trim() );
                location.z = Double.parseDouble( lineIn.substring( 46, 54 ).trim() );
                a.setLocation( location );
                a.setOwner( this );
                if ( !atomMap.containsKey( new Integer(a.getID()) ) ) {
                    atoms.add( a );
                    atomMap.put( new Integer( a.getID() ), a );
                }
            }
            else if ( lineIn.startsWith( "CONECT" ) ) {
                Atom startAtom = (Atom)atomMap.get( new Integer( lineIn.substring( 7,11 ).trim() ) );
                int bondType = Bond.BOND_TYPE_NORMAL;
                for ( int i = 0; i < 10; i++ ) {
                    if ( i == 4 ) { bondType = Bond.BOND_TYPE_HYDROGEN; }
                    try {
                        String next = lineIn.substring( 11+i*5, 16+i*5 );
                        next = next.trim();
                        Atom endAtom = (Atom)atomMap.get( new Integer( next ) );
                        Bond b = new Bond( startAtom, endAtom, bondType );
                        b.setOwner( this );
                        bonds.add( b );
                    }catch(Exception e){}
                }
            }
            else if ( lineIn.startsWith( "CRYST1" ) ){
                boxSize = new double[3];
                boxSize[0] = Double.parseDouble( lineIn.substring( 6,  15 ).trim() );
                boxSize[1] = Double.parseDouble( lineIn.substring( 15, 24 ).trim() );
                boxSize[2] = Double.parseDouble( lineIn.substring( 24, 33 ).trim() );
            }
            else if( lineIn.startsWith( "SCALE" ) ){
                int scaleLoc = lineIn.charAt(5) - '0';
                scaleFactors[scaleLoc-1] = new Vector3d(
                Double.parseDouble( lineIn.substring( 10, 20 ).trim() ), 
                Double.parseDouble( lineIn.substring( 20, 30 ).trim() ), 
                Double.parseDouble( lineIn.substring( 30, 40 ).trim() ) 
                );
            }
            else {
                
            }
        }
        reader.close();
        
        /** We now scale the input values by the scale factors that 
         *  were given to us.
         */
        if ( boxSize != null ) {
            scaleFactors[0].scale( boxSize[0] );
            scaleFactors[1].scale( boxSize[1] );
            scaleFactors[2].scale( boxSize[2] );
        }
        
        Iterator i = atoms.iterator();
        while ( i.hasNext() ) {
            Atom a = (Atom)i.next();
            a.setLocation( scaleLocation( a.getLocation() ) );
        }
        
    }
    
    /** Scales the coordinates by the scale factors found in
     *  the input file.
     */
    protected Vector3d scaleLocation( Vector3d loc ) {
        double x = loc.x; double y = loc.y; double z = loc.z;
        Vector3d out = new Vector3d();
        out.scaleAdd( x, scaleFactors[0], out );
        out.scaleAdd( y, scaleFactors[1], out );
        out.scaleAdd( z, scaleFactors[2], out );
        return out;
    }
    
    public void saveToFile( String fileName ) throws java.io.IOException {
        throw new IOException( "This feature is not implemented!" );
    }
    
    public double[] getBoundingBox() {
        if ( boxSize == null ) {
            return getBoxSize();
        }
        return boxSize;
    }
    
}
