package molmaster.model.loaders;

import java.io.*;

import javax.swing.JOptionPane;

import molmaster.*;
import molmaster.gui.*;

/** Implements a writer to ouput a crude PostScript
 *  approximation of the molmaster scene to a file. The
 *  output file represents a depiction of the scene as
 *  viewed from the current viewing position.
 *
 *  As of the current revision, no shading or lighting
 *  is supported in the output graph.  Only atoms and
 *  bonds are rendered at the current time.
 *
 * @author  Russell Power
 */
public class PostscriptWriter extends molmaster.model.MoleculeFile {

	private static final int FEEDBACK_SIZE_INCREMENT = 1024 * 1024;
	private static final double fogLevel = 0.005;

	public static float brightness = 2.0f;
	public static double scaleFactor = 400;
	public static int resolution = 10;
	public static double bondWidth = 0.5;
	public static double pageWidth = 8;
	public static double pageHeight = 11;

	static {
		molmaster.gui.tools.OptionWindow.getWindow().addValue(
			"PostScript",
			"Output Quality",
			"Determines the output quality when writing to postscript files.",
			PostscriptWriter.class,
			"resolution");
		molmaster.gui.tools.OptionWindow.getWindow().addValue(
			"PostScript",
			"Bond Width",
			"The width of bonds in the output postscript document.",
			PostscriptWriter.class,
			"bondWidth");
		molmaster.gui.tools.OptionWindow.getWindow().addValue(
			"PostScript",
			"Page Width",
			"The width of the output image.",
			PostscriptWriter.class,
			"pageWidth");
		molmaster.gui.tools.OptionWindow.getWindow().addValue(
			"PostScript",
			"Page Height",
			"The height of the output image.",
			PostscriptWriter.class,
			"pageHeight");
		molmaster.gui.tools.OptionWindow.getWindow().addValue(
			"PostScript",
			"Scale Factor",
			"The quantity object to be multiplied against objects "
				+ "sizes to determine rendered size",
			PostscriptWriter.class,
			"scaleFactor");
		molmaster.gui.tools.OptionWindow.getWindow().addValue(
			"PostScript",
			"Brightness",
			"Increases or decreases the relative brightness of atoms.",
			PostscriptWriter.class,
			"brightness");
	}

	/** Load all information from the given file.  This should be
	 *  done in the current thread - if a user needs to perform
	 *  other processing while this takes place, they should spawn
	 *  a new thread before calling this function.
	 *
	 */
	public void loadFromFile(String fileName) throws java.io.IOException {
		throw new IOException(
			"Sorry - I don't know how to read from PostScript files! \n"
				+ "This operation is not permitted.");
	}

	private native void startPage(int feedbackSize, String fileName);
	private native int endPage();

	private int bufferStatus;
	private int bufferSize = FEEDBACK_SIZE_INCREMENT;
	private boolean saveCanceled = false;
	private boolean pageStarted = false;
	static private boolean loadSuccess = false;

	static {
		try {
			String nativePath =
				System.getProperty("libdir")
					+ System.getProperty("file.separator")
					+ "gl2ps"
					+ System.getProperty("file.separator")
					+ "gl2ps.dll";
			molmaster.Log.writeLog("Loading native library: " + nativePath);
			Runtime.getRuntime().load(nativePath);
			loadSuccess = true;
		} catch (Throwable t) {
			molmaster.Log.writeLog(
				"Couldn't load native postscript analysis library.");
			molmaster.Log.writeLog(t);
		}

	}

	/**
	 * Saves the current scene (as viewed through the main canvas) to the
	 * specified file in postscript format.  This is available thanks to the
	 * incredibly useful gl2ps library.  Look in the lib/gl2ps directory for
	 * more information.
	 * 
	 * @see molmaster.model.MoleculeFile#saveToFile(String)
	 */
	public void saveToFile(String fileName) throws java.io.IOException {

		if (!loadSuccess) {
			JOptionPane.showMessageDialog(
				null,
				"The postscript library failed to load.  Aborting.\n"
					+ "Please ensure that you have installed the native library pack for your OS.",
				"Error loading library",
				JOptionPane.ERROR_MESSAGE);
			return;
		}

		final String theFile = fileName;
		final ImageCaptureCanvas3D cv = MainFrame.getCanvas();

		bufferStatus = 0;
		bufferSize = FEEDBACK_SIZE_INCREMENT;
		saveCanceled = false;
		pageStarted = false;

		final JOptionPane op =
			new JOptionPane(
				"Your scene is being saved.\n  This may take a while - please be patient.  "
					+ "How about getting yourself a nice cup of coffee?",
				JOptionPane.INFORMATION_MESSAGE);

		op.setOptions(new Object[] { "Cancel" });

		final javax.swing.JDialog jd =
			op.createDialog(
				MainFrame.getActiveWindow(),
				"Postscript Operation in Progress");

		/**
		 * We create 2 actions - one handles setting up the gl2ps library for
		 * the beginning of the scene, and the other handles finishing a page.
		 */
		cv.addPreRenderAction(new Runnable() {
			public void run() {
				if (bufferStatus != 0 || saveCanceled) {
					Log.info("removing pre-render");
					cv.removePreRenderAction(this);
					return;
				}
				Log.info(
					"Attempting to create postscript image with feedback buffer size: "
						+ bufferSize);
				pageStarted = true;
				startPage(bufferSize, theFile);
			}
		});

		cv.addPostRenderAction(new Runnable() {
			public void run() {

				if (saveCanceled) {
					cv.removePostRenderAction(this);
					return;
				}
				if (!pageStarted)
					return;

				Log.info("Ending postscript page...\n");

				bufferStatus = endPage();
				bufferSize += FEEDBACK_SIZE_INCREMENT;

				if (bufferStatus != 0) {
					Log.info("removing post-render");
					cv.removePostRenderAction(this);
					MainFrame.setProgressMeter(99);
					MainFrame.setStatusMessage(
						"Finished saving postscript image.");
					//	destroy the message dialog - we're done
					javax.swing.SwingUtilities.invokeLater(new Runnable() {
						public void run() {
							jd.setVisible(false);
						}
					});
					return;
				}
			}
		});

		jd.setVisible(true);
		saveCanceled = true;

	}
}
