package molmaster.model.loaders;

import java.io.*;
import java.util.*;
import javax.vecmath.*;

/** Inputs a field describing the the values of psi in 
 *  an area of the scene.
 *
 * @author  Russell Power
 */
public class PsiLoader extends molmaster.model.MoleculeFile implements FieldFile {
    
    private static final int MAX_FIELDS = 100;
    
    private ArrayList fields = new ArrayList();
    private ArrayList files = new ArrayList();
    
    private int[][] max = new int[MAX_FIELDS][3];
    private double[][][] myField;
    private Vector3d xVec;
    private Vector3d yVec;
    private Vector3d zVec;
    
    private double minField;
    private double maxField;
    
    protected PsiLoader( double[][][] field, String name, double min, double max ) {
        myField = field;
        myName = name;
        minField = min;
        maxField = max;
        xVec = new Vector3d(field.length, 0, 0);
        yVec = new Vector3d(0, field[0].length, 0);
        zVec = new Vector3d(0, 0, field[0][0].length );
    }
    
    public PsiLoader() {} 
    
    public void loadFromFile(String fileName) throws java.io.IOException {
        diskFile = new File( fileName );
        myName = diskFile.getName();
        FileReader fileIn = new FileReader( fileName );
        BufferedReader buf = new BufferedReader( fileIn );
        
        String line = "";
        long fLength = diskFile.length();
        int linesRead = 0;
        int bytesRead = 0;
        molmaster.gui.MainFrame.setStatusMessage( "Finding field sizes..." );
        while ( (line = buf.readLine()) != null ) {
            StringTokenizer tok = new StringTokenizer( line );
            int f = Integer.parseInt( tok.nextToken() );
            int x = Integer.parseInt( tok.nextToken() );
            int y = Integer.parseInt( tok.nextToken() );
            int z = Integer.parseInt( tok.nextToken() );
            max[f][0] = (x>max[f][0]) ? x : max[f][0];
            max[f][1] = (y>max[f][1]) ? y : max[f][1];
            max[f][2] = (z>max[f][2]) ? z : max[f][2];
            linesRead++;
            bytesRead += line.length() + 1;
            
            if ( (linesRead % 20) == 0 ) {
                long val = ((bytesRead* 100)/fLength);
                molmaster.gui.MainFrame.setProgressMeter( (int)val );
            }
        }
        
        molmaster.gui.MainFrame.setStatusMessage( "Loading field values..." );
        fileIn.close();
        fileIn = new FileReader( diskFile );
        buf = new BufferedReader( fileIn );

        for ( int i = 0; i < MAX_FIELDS; i++ ) {
            if ( max[i][0] == 0 ) { 
                fields.add( new float[1][1][1] );
                continue;
            }
            double[][][] field = new double[ max[i][0] ][ max[i][1] ][ max[i][2] ];
            fields.add( field );
        }
        
        int linesMax = linesRead;
        linesRead = 0;
        
        double[] minF = new double[ MAX_FIELDS ];
        double[] maxF = new double[ MAX_FIELDS ];
        
        for ( int i = 0; i < MAX_FIELDS; i++ ) {
            minF[i] = Double.POSITIVE_INFINITY;
            maxF[i] = Double.NEGATIVE_INFINITY;
        }
        
        while ( (line = buf.readLine()) != null ) {
            StringReader sr = new StringReader( line );
            int f = (int)getFixed( sr );
            int x = (int)getFixed( sr );
            int y = (int)getFixed( sr );
            int z = (int)getFixed( sr );
            double[][][] field = (double[][][])fields.get( f );
            double fieldVal = getFixed( sr );
            fieldVal *= 100000;
            field[x-1][y-1][z-1] = fieldVal;
            
            minF[f] = (fieldVal<minF[f])? fieldVal : minF[f];
            maxF[f] = (fieldVal>maxF[f])? fieldVal : maxF[f];
            
            linesRead++;
            if ( ( linesRead % 100 ) == 0 ) {
                molmaster.gui.MainFrame.setProgressMeter( linesRead * 100 / linesMax );
            }
        }

        for ( int i = 0; i < MAX_FIELDS; i++ ) {
            if ( max[i][0] == 0 ) { continue; }
            molmaster.gui.input.OrbitBehavior.getActive().setRotationCenter(
            new Point3d( max[i][0]/2, max[i][1]/2, max[i][2]/2 ) 
            );
            PsiLoader ps = new PsiLoader( (double[][][])fields.get(i), myName + " - " + i, minF[i], maxF[i] );
            ps.diskFile = diskFile;
            files.add( ps );
        }
        
        molmaster.gui.MainFrame.setStatusMessage( "Finished." );
        molmaster.gui.MainFrame.setProgressMeter( 0 );
    }
    
    public void addToScene( molmaster.model.MolModel scene ){
        Iterator i = files.iterator();
        while ( i.hasNext() ) {
            PsiLoader psiFile = (PsiLoader)i.next();
            scene.addAtomGroup( psiFile );
        }
    }
    
    /** Returns the basis vectors used to obtain field locations.    */
    public javax.vecmath.Vector3d[] getBasis() {
        return new Vector3d[] { xVec, yVec, zVec };
    }
    
    /** Returns a 3d array representing the field values
     *  at points in space.  Each index position is multiplied
     *  by the basis vector to obtain the field location in space.   */
    public double[][][] getField() {
        return myField;
    }
    
    /** Maximum value taken by this field.     */
    public double getFieldMax() {
        return maxField;
    }
    
    /** Returns the minimum field value taken on by this field.  */
    public double getFieldMin() {
        return minField;
    }
    
}
