package molmaster.model.loaders;

import molmaster.*;
import molmaster.model.*;

import java.io.*;
import java.util.*;

import javax.vecmath.*;

/** Loads the CHGCAR and POSCAR type files output 
 *  from VASP.  If you need to load actual field 
 *  data from a VASP file, use the CHGCAR loader.
 *
 * @author  rjpower
 */
public class VaspLoader extends MoleculeFile {
    
    protected static final int DIRECTED =   0x10;
    protected static final int CARTESIAN =  0x20;
    
    protected StreamTokenizer tok;
    protected boolean holdOpen = false;
    protected Vector3d basis[] = new Vector3d[3];
    protected BufferedReader r;
    protected double boxScale;
    
    /** Creates a new instance of VaspLoader */
    public VaspLoader() {
    }
    
    public Vector3d[] getBasis() {
        return basis;
    }
    
    public double getBoxScale() { 
        return boxScale;
    }
    
    public void loadFromFile(String fileName) throws java.io.IOException {
        diskFile = new File( fileName );
        myName = diskFile.getName();
        System.out.println(myName);
        r = new BufferedReader( new FileReader( fileName ) );
        tok = new StreamTokenizer( r );
        
        List atomNames = new ArrayList();
        tok.eolIsSignificant( true );
        while ( tok.nextToken() != StreamTokenizer.TT_EOL ) {
            atomNames.add( tok.sval );
        }
        tok.eolIsSignificant( false );
        tok.nextToken();
        boxScale = getNumber( tok );
        
        skipLine( tok );
        for ( int i = 0; i <= 2; i++ ) {
            basis[i] = new Vector3d();
            tok.nextToken(); basis[i].x = getNumber( tok );
            tok.nextToken(); basis[i].y = getNumber( tok );
            tok.nextToken(); basis[i].z = getNumber( tok );
            basis[i].scale( boxScale );
            Log.writeLog( basis[i] );
            skipLine( tok );
        }
        
        Map atomMap = new HashMap();
        tok.eolIsSignificant( true );
        boolean overflow = false;
        int cur = 0;
        while ( tok.nextToken() != StreamTokenizer.TT_EOL ) {
            if ( cur < atomNames.size() ) {
                atomMap.put( atomNames.get(cur), new Integer( (int)tok.nval ) );
            }else {
                atomNames.add( String.valueOf(cur) );
                atomMap.put( String.valueOf(cur) , new Integer( (int)tok.nval ) );
            }
            System.out.println("Type: " + tok.nval );
            cur++;
        }
        if ( cur < atomNames.size() ) {
            while ( atomNames.size() > cur ) {
                atomNames.remove( atomNames.size() - 1 );
            }
        }
        //skipLine( tok );
        
        tok.nextToken();
        String typeName = tok.sval;
        if ( typeName != null ) {
            typeName = typeName.toLowerCase();
        }else {
            typeName = "direct";
        }
        int type = 0;
        if ( typeName.equals( "cartesian" ) )
            type = CARTESIAN;
        if ( typeName.equals( "direct" ) )
            type = DIRECTED;
        
        skipLine( tok );

        if ( type == 0 ){
            tok.nextToken();
            typeName = tok.sval;
            typeName = typeName.toLowerCase();
            System.out.println("Typename:" + typeName );
            if ( typeName.equals( "cartesian" ) )
                type = CARTESIAN;
            if ( typeName.equals( "direct" ) )
                type = DIRECTED;
        }
        double x = 0;
        double y = 0;
        double z = 0;
        
        int atomCount = 0;
        for ( int i = 0; i < atomNames.size(); i++ ) {
            String curName = (String)atomNames.get(i);
            int count = ((Integer)atomMap.get(curName)).intValue();
            for ( int j = 0; j < count; j++ ) {
                tok.nextToken(); x = getNumber( tok );
                tok.nextToken(); y = getNumber( tok );
                tok.nextToken(); z = getNumber( tok );
                Atom a = AtomTable.getActive().getAtom( curName );
                
                Vector3d loc = new Vector3d(x, y, z);
                if ( type == DIRECTED ) {
                    getDirected( loc );
                }
                else {
                    loc.scale( boxScale );
                }
                a.setLocation( loc );
                a.setOwner( this );
                a.setID( atomCount++ );
                
                super.atoms.add( a );
                if( (i == (atomNames.size() - 1)) && j == (count - 1) ){
                }else{ skipLine( tok );}
            }
        }
        /** Keep the file open for a subclass.  */
        if( !holdOpen )
            r.close();
        
        System.out.println("Atoms loaded: " + atoms.size() );
    }
    
    protected void getDirected( Vector3d loc ) {
        double x = loc.x; double y = loc.y; double z = loc.z;
        Vector3d out = new Vector3d();
        out.scaleAdd( x, basis[0], out );
        out.scaleAdd( y, basis[1], out );
        out.scaleAdd( z, basis[2], out );
        loc.set( out );
    }
    
    /** Save the current contents of this resource to
     * the filename specified.  */
    public void saveToFile(String fileName) throws java.io.IOException {
    }
    
    protected void skipLine( StreamTokenizer tok ) throws java.io.IOException {
        tok.eolIsSignificant( true );
        while ( !(tok.nextToken() == StreamTokenizer.TT_EOL) );
        tok.eolIsSignificant( false );
    }
}
