package molmaster.util;

import molmaster.*;
import molmaster.gui.*;
import molmaster.model.*;
import molmaster.gui.editbar.analysis.*;

import java.util.*;
import javax.vecmath.*;

/** Convenience methods for generating bonds
 *  between various atom types.
 *
 * @author  Russell Power
 */
public class BondGenerator {
    
    private String startAtom, endAtom;
    private float minLength = 0;
    private float maxLength = -1;
    private boolean autoBond = false;
    private boolean visible = true;
    private Color3f color;
    private LengthFunction comp;
    private Collection fileList;
    private Collection bonds = new HashSet();
    private int[] coord;
    private Map specialBonds = new HashMap();
    
    private int count = 0;
    private int hCount = 0;
    private int tCount = 0;
    
    public BondGenerator() {
        fileList = MolModel.getModel().getFiles();
        autoBond = true;
    }
    
    public BondGenerator( float minLength, float maxLength,
    String startType, String endType, Color3f color ) {
        fileList = MolModel.getModel().getFiles();
        this.startAtom = startType;
        this.endAtom = endType;
        this.maxLength = maxLength;
        this.minLength = minLength;
        this.color = color;
    }
    
    /** Initializes a BondGenerator for generating bonds between the two given atom
     * types.
     */    
    public BondGenerator( String start, String end ) {
        this.startAtom = start;
        this.endAtom = end;
        this.color = new Color3f();
    }
    
    /** Sets the comparison function to use instead of the default length based comparison
     * to determine whether a bond is valid.
     */    
    public final void setComparisonFunction( LengthFunction f ) { comp = f; }
    /** Sets whether to render the set of bonds generated.
     */    
    public final void setVisible( boolean vis ) { visible = vis; }
    /** Sets a map of BondSizes' that should be used in place of the default maximum
     * bond length if the Bond to be generated has the id specified.
     */    
    public final void setSpecial( Map special ) { specialBonds = special; }
    
    /** Returns the files currently being used to generate bonds.
     */
    public final Collection getFiles() {return fileList;}
    
    /** Sets the current list of files to generate bonds with.  */
    public final void setFiles( Collection l ) { fileList = l; }
    
    /** Searches for all possible bonds in the given set of files.
     *  This bonding is performed using the default bond size, or
     *  via a user defined comparison function.
     */
    public void autoBond() {
        maxLength = -1;
        
        MainFrame.setStatusMessage("Generating bonds for all atom types...");
        if ( fileList.size() <= 0 ) { return; }
        Iterator i = fileList.iterator();
        while ( i.hasNext() ){
            MoleculeFile cf = (MoleculeFile)i.next();
            if ( cf == null ) { continue; }
            Collection atoms = cf.getAtoms();
            int max = atoms.size();
            Log.writeLog( "Reading file of " + max + " atoms. " );
            Iterator j = atoms.iterator();
            while ( j.hasNext() ){
                Atom a = (Atom)j.next();
                Iterator k = atoms.iterator();
                while ( k.hasNext() ){
                    if ( tCount++ % 50 == 0 ) {
                        MainFrame.setProgressMeter( tCount * 100 / max );
                    }
                    Atom b = (Atom)k.next();
                    connect( a, b, bonds );
                }
            }
        }
        
        MainFrame.setProgressMeter( 50 );
        Log.writeLog( count + " regular bonds generated.");
        Log.writeLog( hCount + " hydrogen bonds generated." );
        Log.writeLog( "adding generated bonds to scene." );
        if ( visible ) {
            MolModel.getModel().addBondGroup( bonds );
        }else {
            i = bonds.iterator();
            while ( i.hasNext() ) {
                Bond b = (Bond)i.next();
                b.getStart().addBond( b );
                b.getEnd().addBond( b );
            }
        }
        
        MainFrame.setProgressMeter( 100 );
        Log.writeLog( "Bond generation complete." );
    }
    
    /** Attempts to connect the two given atoms.    */
    private void connect( Atom a, Atom b, Collection bonds ) {
        if ( a == b ) { return; }
        Bond bond = new Bond( a, b, Bond.BOND_TYPE_NORMAL, maxLength );
        if ( bonds.contains( bond ) ) {return;}
        
        boolean valid = false;
        if ( comp != null ) {
            valid = comp.isValid( a,  b, 0, Bond.bondLength( a, b ) );
        }
        else {
            if ( specialBonds.containsKey( bond.getBondTypeID() ) ) {
                double max = ((BondSize)specialBonds.get(bond.getBondTypeID())).getMaxBondSize();
                if ( bond.getLength() < max ) {
                    valid = true;
                }
            }
            else {
                valid = bond.isValid() && (bond.getLength() > minLength);
                if (
                ((AtomAttributes)a.getAttributes()).atomicSymbol.equals("H") ||
                ((AtomAttributes)b.getAttributes()).atomicSymbol.equals("H") ){
                    bond = new Bond( a, b, Bond.BOND_TYPE_HYDROGEN, maxLength );
                    if ( bonds.contains( bond ) ){ return; }
                    if ( bond.isValid()){
                        if ( bond.getLength() > minLength ) {
                            valid = false;
                            bonds.add(bond);
                            hCount++;
                        }
                    }
                }
            }
        }
        
        if( valid ) {
            bonds.add( bond );
            count++;
            return;
        }
    }
    
    /** Generates a set of bonds between two specific atom
     *  types.
     */
    public void genBondSet(){
        Atom[] si = MolModel.getModel().getAtomsOfType( startAtom );
        Atom[] ei = MolModel.getModel().getAtomsOfType( endAtom );
        
        MainFrame.setStatusMessage("Generating bonds...");
        
        if ( maxLength < 0 ) {
            maxLength = (float)Bond.bondLength( si[0], ei[0] );
        }
        
        int totalBonds = si.length * ei.length;
        int tCount = 0;
        Log.writeLog( "Charting bonds - bond length = " + maxLength );
        Iterator f = fileList.iterator();
        while ( f.hasNext() ){
            MoleculeFile currentFile = (MoleculeFile)f.next();
            for ( int i = 0; i < si.length; i++ ){
                Atom a = si[i];
                if ( !currentFile.getAtoms().contains( a ) ) { continue; }
                for ( int j = 0; j < ei.length; j++ ){
                    tCount++;
                    if ( tCount % 100 == 0 ) {
                        MainFrame.setProgressMeter( tCount * 50 / totalBonds );
                    }
                    Atom b = ei[j];
                    if ( !currentFile.getAtoms().contains(b) ) {continue;}
                    
                    connect( a, b, bonds );
                }
            }
        }
        
        MainFrame.setProgressMeter( 50 );
        Log.info( count + " regular bonds generated.");
        Log.info( hCount + " hydrogen bonds generated." );
        Log.info( "adding generated bonds to scene." );
        if ( visible ) {
            MolModel.getModel().addBondGroup( bonds );
            Log.writeLog( "Setting bond color to: " + color );
            Iterator i = bonds.iterator();
            while ( i.hasNext() ) {
                BondView.bondToView( i.next() ).setColor( color );
            }
        }else {
            Iterator i = bonds.iterator();
            while ( i.hasNext() ) {
                Bond b = (Bond)i.next();
                Atom s = b.getStart();
                if ( !s.getConnected().contains( b.getEnd() ) ) {
                    b.getStart().addBond( b );
                    b.getEnd().addBond( b );
                }
            }
        }
        
        MainFrame.setProgressMeter( 100 );
        Log.info( "Bond generation complete." );
    }
    
    /** Clears all bonds generated during the last step.    */
    public void clear() {
        if ( visible ) {
            MolModel.getModel().removeBondGroup( bonds );
        }else{
            //  we have to remove manually since this aren't
            //  actually being used by the model
            Iterator i = bonds.iterator();
            while( i.hasNext() ){
                Bond b = (Bond)i.next();
                b.getStart().removeBond( b );
                b.getEnd().removeBond( b );
            }
        }
    }
    
    /**
     */    
    public int getMaxCoordination() {
        getCoordination();
        return maxCoord;
    }
    
    public int getMinCoordination() {
        getCoordination();
        return minCoord;
    }
    
    private int minCoord, maxCoord;
    
    /** Returns a map containing the coordination numbers
     *  found in the current bond set.
     */
    public int[] getCoordination() {
        if ( coord != null ) {
            return coord;
        }
        
        coord = new int[100];
        HashSet seen = new HashSet();
        Iterator i = bonds.iterator();
        while ( i.hasNext() ) {
            Bond b = (Bond)i.next();
            Atom start = b.getStart();
            Atom end = b.getEnd();
            if ( !seen.contains( start ) ) {
                int size = start.getConnected().size();
                coord[size]++;
                if ( size > maxCoord ) { maxCoord = size; }
                if ( size < minCoord ) { minCoord = size; }
            }
            if ( !seen.contains( end ) ) {
                int size = end.getConnected().size();
                coord[size]++;
                if ( size > maxCoord ) { maxCoord = size; }
                if ( size < minCoord ) { minCoord = size; }
            }
            seen.add( start ); seen.add( end );
        }
        return coord;
    }
}
