package molmaster.util;

import java.util.*;
import javax.vecmath.*;

import molmaster.*;
/** Performs generic common neighbor (CN) analysis
 *  on the input group of atoms and bonds.  Each atom pair is
 *  studied and a triplet of values is found and stored for
 *  them.  The triplet represents the number of common
 *  neighbors belonging to each pair, the total number of
 *  common neighbor bonds, and the longest path existing
 *  between the common neighbors.
 *
 * @author  Russell Power
 */
public class CNAnalysis {
    
    private Bond[] bonds;
    private Atom[] atoms;
    
    /** Initialize the calculation with the given atom
     *  and bond list.  The references are transferred into
     *  arrays to allow for faster computation.
     *
     */
    public CNAnalysis(Collection inAtoms) {
        Iterator i = inAtoms.iterator();
        atoms = new Atom[ inAtoms.size() ];
        int count = 0;
        while ( i.hasNext() ){
            atoms[ count++ ] = (Atom)i.next();
        }
        
    }
    
    class StatInfo implements Comparable {
        int count;
        Vector3d type;
        
        public int compareTo(Object obj) {
            StatInfo s = (StatInfo)obj;
            if ( s.count > count ) { return 1; }
            if ( count > s.count ) { return -1; }
            return 0;
        }
        
    }
    
    /** Performs the CN calculations.  This procedure
     *  may be rather time consuming, so it is wise to
     *  run it in a separate thread to allow for UI updates
     *  if needed.
     */
    public void doCalculation() {
        HashMap cache = new HashMap();
        HashMap stats = new HashMap();
        
        //  for each atom
        for ( int i = 0; i < atoms.length; i++ ) {
            //  for each bond possible
            Atom start = atoms[i];
            List sConnect = start.getConnected();
            List cna = new ArrayList();
            for ( int j = 0; j < sConnect.size(); j++ ) {
                Atom partner = (Atom)sConnect.get(j);
                
                if ( cache.containsKey( partner.getID() + "" + start.getID() ) ) {
                    Vector4d v = (Vector4d)cache.get( partner.getID() + "" + start.getID() );
                    cna.add( v );
                    Vector3d type = new Vector3d( v.x, v.y, v.z );
                    StatInfo s = (StatInfo)stats.get(type);
                    if ( s == null ) {
                        s = new StatInfo();
                        s.type = new Vector3d( type.x, type.y, type.z );
                        stats.put( type, s );
                    }
                    s.count++;
                    continue;
                }
                
                Vector4d out = new Vector4d();
                out.w = partner.getID();
                List eConnect = partner.getConnected();
                
                merge = new HashSet();
                Iterator x = sConnect.iterator();
                while ( x.hasNext() ) {
                    Atom nb = (Atom)x.next();
                    if ( eConnect.contains( nb ) ) {
                        merge.add( nb );
                    }
                }
                
                out.x = merge.size();	// first member of tuple is number of
                                        // common neigbors
                
                Iterator it = merge.iterator();
                Set cnBonds = new HashSet();
                while ( it.hasNext() ) {
                    Atom cn = (Atom)it.next();
                    Iterator jt = cn.getConnected().iterator();
                    while ( jt.hasNext() ) {
                        Atom n = (Atom)jt.next();
                        if( merge.contains( n ) ) {
                            cnBonds.add( cn.getID() + "-" + n.getID() );
                            cnBonds.add( n.getID() + "-" + cn.getID() );
                        }
                    }
                }
                
                out.y = cnBonds.size() / 2;
                // second member of tuple is the number
                // of bonds between common neighbors
                
                //System.out.println("Finding longest path...");
                //merge.add( start );
                if ( out.x == 4 && out.y == 1 ) { 
                    log = true;
                }else {
                    log = false;
                }
                out.z = findLongestPath( merge );
                // third member is the longest path found inside common neighbors
                
                
                cna.add( out );
                cache.put( start.getID() + ""  + partner.getID(), out );
                Vector3d type = new Vector3d( out.x, out.y, out.z );
                StatInfo s = (StatInfo)stats.get(type);
                if ( s == null ) {
                    s = new StatInfo();
                    s.type = type;
                    stats.put( type, s );
                }
                s.count++;
            }
            start.setUserObject( cna );
            //System.out.println(i);
            System.out.flush();
            molmaster.gui.MainFrame.setProgressMeter( i / atoms.length );
        }
        
        ArrayList out = new ArrayList( stats.values() );
        Collections.sort( out );
        Iterator q = out.iterator();
        while ( q.hasNext() ) {
            StatInfo s = (StatInfo)q.next();
            System.out.println("Type:" + s.type + ", count: " + s.count / 2);
        }
    }
    
    private boolean log = false;
    private Set merge;
    
    /**	Find the longest path on the given group of atoms.
     * The length of a cycle is given as the number of bonds
     * along that cycle.
     */
    private int findLongestPath( Collection group ) {
        Iterator i = group.iterator();
        int max = 0;
 
        i = group.iterator();
        while ( i.hasNext() ){
            Atom a = (Atom)i.next();
            int l = findPath( a, new ArrayList() );
            if ( l > max ) { max = l; }
        }
        return max;
    }
    
    private int findPath( Atom start, List visited ) {
        //  cycle
        if ( visited.contains( start ) ) {
            if ( visited.size() > 2 ) {
                return visited.size();
            }
            return 0;
        }

        int max = visited.size();
        int init = max;
        
        visited.add( start );
        
        Iterator i = start.getConnected().iterator();
        while ( i.hasNext() ){
            Atom a = (Atom)i.next();
            if ( !merge.contains( a ) ) { continue; }
            
            List vis = new ArrayList( visited );
            int n = findPath( a, vis );
            
            if ( n > max ) {
                max = n;
            }
        }
        
        return max;
    }
    
}
