package molmaster.util;

import molmaster.*;
import java.io.*;
import java.util.*;

/** Provides methods to store and retrieve
 *  arbitrary data elements from a stable resource -
 *  in this case, a disk file.  Data is synced to
 *  disk at regular intervals.  Multiple instances
 *  of this class sharing the same resource will
 *  also share the same inner structure, allowing
 *  for consistency.
 *
 * @author  Russell Power
 */
public class DataStore {
    
    protected Map objectMap = null;
    protected File resourceFile = null;
    
    static protected Map resourceMap = new HashMap();
    
    /** Creates a new DataStore which will use the
     *  given string as the name of the resource
     *  set to use.
     */
    public DataStore(String resourceName) {
        File base = new File( Molmaster.dataStorage + "/data/");
        base.mkdirs();
        
        String fName = base.getAbsolutePath() + "/" + resourceName + ".res";
        resourceFile =  new File( fName );
        
        if ( resourceMap.containsKey( resourceName ) ) {
            objectMap = (Map)resourceMap.get( resourceName );
        }else {
            objectMap = new HashMap();
            resourceMap.put( resourceName, objectMap );
            
            if ( resourceFile.exists() ) {
                try {
                    FileInputStream fis = new FileInputStream(fName);
                    ObjectInputStream ois = new ObjectInputStream( fis );
                    readObjects(ois);
                    ois.close();
                    fis.close();
                }catch(Exception e){
                    molmaster.Log.writeLog( "Couldn't read input file." );
                }
            }
            /** Add a hook to get us flushed to disk when the app
             *  shuts down.
             */
            Runtime.getRuntime().addShutdownHook(
                new Thread() {
                    public void run() {
                        flush();
                    }
            });
        }
    }
    
    protected void readObjects( ObjectInputStream ois ){
        while ( true ) {
            try {
                Pair p = (Pair)ois.readObject();
                if ( p == null ) { return; }
                objectMap.put( p.key, p.value );
            }catch(EOFException eof){
                return;
            }catch(Exception e){
                molmaster.Log.writeLog("Error loading key/value pair");
                molmaster.Log.writeLog( e );
            }
        }
    }
    
    protected void writeObjects( ObjectOutputStream oos ) {
        Iterator i = objectMap.keySet().iterator();
        while ( i.hasNext() ){
            try {
                Serializable key = (Serializable)i.next();
                Serializable value = (Serializable)objectMap.get( key );
                oos.writeObject( new Pair( key, value ) );
            }catch( Exception e ){
                molmaster.Log.writeLog( "Couldn't write key/value pair" );
                molmaster.Log.writeLog( e );
            }
        }
        
    }
    
    /** Flushes the contents of this DataStore object to disk.
     */    
    synchronized public void flush() {
        molmaster.Log.writeLog( "Flushing store: " + resourceFile );
        try {
            FileOutputStream fos = new FileOutputStream( resourceFile );
            ObjectOutputStream oos = new ObjectOutputStream( fos );
            writeObjects( oos );
            oos.close();
            fos.close();
        }catch(Exception e){
            molmaster.Log.writeLog("Couldn't write resource file.");
            e.printStackTrace( molmaster.Log.getLogStream() );
            new molmaster.gui.tools.ErrorDialog( e ).show();
        }
    }
    
    /** Insert the given values into the data store.  These values will be written to
     * disk the next time the DataStore is flushed.
     */    
    public void put( Serializable key, Serializable value ) {
        objectMap.put( key, value );
    }
    
    /** Returns the value associated with the given input key.  If no value is found
     * null is returned.
     */    
    public Object get( Serializable key ) {
        return objectMap.get( key );
    }
    
    /** Removes the given key (and associated value) from the DataStore.
     */    
    public void remove( Serializable key ) {
        objectMap.remove( key );
    }
    
    /** Clears all key/value entries from the data store.
     */    
    public void clear() {
        objectMap.clear();
    }
    
    /** Returns the current number of key/value pairs in the datastore.
     */    
    public int getCount() {
        return objectMap.size();
    }
    
    /** Equivalent to ( getCount() == 0 );
     */    
    public boolean isEmpty() {
        return (getCount() == 0 );
    }
    
    protected void finalize() {
        flush();
    }
    
}

class Pair implements Serializable  {
    Serializable key;
    Serializable value;
    
    Pair( Serializable k, Serializable v ) {
        key = k; value = v;
    }
}
