package molmaster.util;

import molmaster.*;
import molmaster.model.*;

import java.io.*;
import java.util.*;
import javax.swing.*;

import java.awt.event.*;

/** Provides methods related to keeping track
 *  of a group of files.
 *
 * @author  Russell Power
 */
public class FileList implements molmaster.model.ModelListener {
    
    static private FileList defaultList = new FileList();
    /** Returns an instance of the default file list.
     */    
    static public FileList getDefaultList(){ return defaultList; }
    
    protected List fileSet = new ArrayList();
    protected FileListPanel theList;
    protected JDialog theEditor;
    protected Map fileToLoader = new HashMap();
    
    public FileList() {
        MolModel.getModel().addListener( this );
        fireActionEvent();

    }
    
    /** Returns the current list of open files. */
    public Collection getFiles() {
        updateFiles();
        return fileSet;
    }
    
    /** In the case that this file list is being used
     *  to keep track of MoleculeFiles, this returns
     *  all the MoleculeFile's currently loaded.
     */
    public Collection getLoaders() {
        updateFiles();
        return fileToLoader.values();
    }
    
    /** Returns the map associating files with a
     *  loader.
     */
    public Map getLoaderMap() {
        updateFiles();
        return fileToLoader;
    }
    
    /** Displays a dialog for adding files.  Any files
     *  selected are added into the current list of files.
     */
    public void showAddDialog(java.awt.Dialog parent) {
        if ( theEditor == null ) {
            theEditor = new JDialog( molmaster.gui.MainFrame.getActiveWindow(), "File selection", true );
            FileEditPanel panel = new FileEditPanel( this, theEditor );
            
            theEditor.getContentPane().add( panel, java.awt.BorderLayout.CENTER );
            java.awt.Dimension d = panel.getPreferredSize();
            d.width += 10;
            d.height += 40;
            
            theEditor.setSize( d );
        }
        theEditor.show();
    }
    
    /** Displays a dialog box for removing files from the
     *  current file listing.
     */
    public void showRemoveDialog(java.awt.Dialog parent) {
        showAddDialog( parent );
    }
    
    /** Displays a list of the currently selected files
     *  from this list, with buttons to add and remove
     *  files.
     */
    public void showListDialog(java.awt.Dialog parent){
        if ( theList == null ) {
            theList = new FileListPanel( this, molmaster.gui.MainFrame.getActiveWindow() );
        }
        theList.show();
    }
    
    /** Adds the given file to this list.   */
    public void addFile( File f ) {
        MolFile mf = new MolFile( f );
        if ( !fileSet.contains( mf ) ){
            fileSet.add( mf );
        }
        Log.writeLog( "Adding group to scene: " + mf );
        fireActionEvent();
    }
    
    /** Removes the given file from this list.  */
    public void removeFile( File f ) {
        fileSet.remove( new MolFile(f) );
        System.out.println("Removing group from scene: " + f.getName() );
        MoleculeFile mf = (MoleculeFile)fileToLoader.get( new MolFile(f) );
        if ( mf != null ) {    
            MolModel.getModel().removeAtomGroup( mf );
        }
        //System.out.println(fileSet.size());
        fireActionEvent();
    }
    
    /** Removes all items from the current file list.   */
    public void clear() {
        for ( int i = 0; i < fileSet.size(); i++ ){
            removeFile( (File)fileSet.get(i) );
        }
    }
    
    /** Returns a JComponent which contains a list of
     *  all currently selected files.
     */
    public JComponent getFileListPanel() {
        return null;
    }
    
    /** Adds a listener to this filelist.  Listeners are notified regarding changes
     * to the list state.
     */    
    public void addListener( java.awt.event.ActionListener e ) {
        listeners.add( e );
    }
    /** Removes the given listener from the file list.
     */    
    public void removeListener( java.awt.event.ActionListener e ){
        listeners.remove( e );
    }
    
    private boolean updating = false;
    protected void fireActionEvent() {
        //if (updating) { return; }
        updateFiles();
        /** Preload any changed files.  */
        fileToLoader.keySet().retainAll( getFiles() );
        Iterator i = getFiles().iterator();
        while ( i.hasNext() ) {
            File nextFile  = (File)i.next();
            MolFile f = new MolFile( nextFile );
            if ( !fileToLoader.containsKey( f ) ){
                try {
                    molmaster.Log.writeLog( "Loading from: " + f.getName() );
                    MoleculeFile mf = MoleculeFile.getLoader( f.getName() );
                    mf.loadFromFile( f.getAbsolutePath() );
                    mf.addToScene( MolModel.getModel() );
                    fileToLoader.put( f, mf );
                }catch(Exception e){
                    molmaster.Log.writeLog("An error occured while trying to read:" + f.getName() );
                    e.printStackTrace( molmaster.Log.getLogStream() );
                }
            }
        }
        
        //updating = true;
        i = listeners.iterator();
        while ( i.hasNext() ) {
            ((ActionListener)i.next()).actionPerformed(
            new ActionEvent( this, 0, "file list changed" )
            );
        }
        //updating = false;
    }
    
    protected void updateFiles() {
        Iterator i = MolModel.getModel().getFiles().iterator();
        while ( i.hasNext() ) {
            MoleculeFile m = (MoleculeFile)i.next();
            //System.out.println(m);
            File diskFile = m.getDiskFile();
            MolFile f = new MolFile( diskFile );
            if ( !fileSet.contains(f) ){
                fileSet.add( f );
            }
            fileToLoader.put( f, m );
        }
        Collections.sort( fileSet );
    }
    
    /** Used to obtain information on events happening to the model, and to keep the
     * FileList up to date with the model's state.
     */    
    public void modelUpdated(MolModel scene, int opType, Object opValue) {
        if ( opType == FILE_ADDED ){
            addFile( ((MoleculeFile)opValue).getDiskFile() );
        }else if ( opType == FILE_REMOVED ){
            removeFile ( ((MoleculeFile)opValue).getDiskFile() );
        }
        fireActionEvent();
    }
    
    protected List listeners = new ArrayList();
}

class MolFile extends File {
    /** Overrides the default String operator of File to provide a friendlier display.
     */    
    public MolFile( File f ) { super( f.getAbsolutePath() ); }
    public String toString() { return getName(); }

}
