package molmaster.util;

import java.awt.Color;
import java.awt.image.*;
import java.util.*;

import javax.vecmath.*;
import javax.media.j3d.*;

import molmaster.model.*;
import molmaster.model.loaders.*;

/** Utility class for generating a 2d slice representing isosurface
 *  values across a plane.
 *
 * @author  Russell Power
 */
public class TextureSlice {
    
    /** 
     *  Calculates the texture slice across the input image.  By varying the
     *  input parameters the speed of calculation can be varied.
     *
     */
    protected double[][] calcSlice( 
    FieldFile fFile, 
    Vector3d normal,
    int xRes, 
    int yRes ) {
        
        double fieldMax = fFile.getFieldMax();
        double fieldMin = fFile.getFieldMin();
        double[][][] field = fFile.getField();
        
        if ( xRes <= 0 || yRes <= 0 ) { 
            throw new IllegalArgumentException( "Image size cannot be <= 0!" );
        }
        
        Vector3d c = new Vector3d( field.length/2, field[0].length/2, field[0][0].length/2 );
        
        Vector4d plane = new Vector4d( normal.x, normal.y, normal.z, 0 );
        plane.w = normal.x * c.x + normal.y * c.y + normal.z * c.z;
        
        double[][] imageOut = new double[ xRes ][ yRes ];
        for ( int i = 0; i < xRes; i++ ) {
            for ( int j = 0; j < yRes; j++ ) {
                /** We perform a projection of the target point at the back of the 
                 *  cube onto the target planar surface.  We treat the x and y location
                 *  in space as constant, and find the z location on the plane that 
                 *  corresponds to their values.
                 */
                int xB = (field.length / xRes) * i;
                int yB = (field.length / yRes) * j;
                int zB = (int)((plane.w - (normal.x * xB + normal.y * yB)) / normal.z);
                try {
                    imageOut[i][j] = field[xB][yB][zB];
                }catch( ArrayIndexOutOfBoundsException e ){}
            }
        }
        return imageOut;
    }
    
    /** Returns an image corresponding the input value
     *  information.  genType determines how the output
     *  is classified - as a logarithmic or linear scale.
     */
    protected BufferedImage genImage( double[][] input, FieldFile f, int genType ) {
        BufferedImage bi = 
        new BufferedImage( input.length, input[0].length, BufferedImage.TYPE_INT_RGB );
        
        double min = f.getFieldMin();
        double max = f.getFieldMax();
        
        int[] out = new int[ input.length * input[0].length ];
        
        ColorInterp ci = new ColorInterp();
        ci.addColor( Color.blue );
        ci.addColor( Color.green );
        ci.addColor( Color.red );
        
        for ( int i = 0; i < input.length; i++ ) {
            for ( int j = 0; j < input[0].length; j++ ) {
                double rel = (input[i][j] - min) / max;
                out[ i * input.length + j ] = ci.getColor( rel );
            }
        }
        bi.getRaster().setSamples( 0, 0, bi.getWidth(), bi.getHeight(), 0, out );
        return bi;
    }
    
    public Shape3D getSlice( FieldFile f, int xRes, int yRes, Vector3d xMove, Vector3d yMove ) {
        Vector3d normal = new Vector3d();
        normal.cross( xMove, yMove );
        
        double[][] vals = calcSlice( f, normal, xRes, yRes );
        BufferedImage bi = genImage( vals, f, 0 );
        
        Texture2D tOut = molmaster.gui.TextureCache.getTexture( bi, false );
        QuadArray gOut = 
        new QuadArray( 4, GeometryArray.COORDINATES | GeometryArray.TEXTURE_COORDINATE_2 );
        
        MoleculeFile mf = (MoleculeFile)f;
        BoundingBox box = mf.getBounds();
        Point3d lower = new Point3d(); box.getLower( lower );
        Point3d size = new Point3d(); box.getUpper( size);
        
        Vector3d center = new Vector3d( lower.x + size.x/2, lower.y + size.y / 2, lower.z + size.z / 2 );
        center.scaleAdd( -0.5, xMove );
        center.scaleAdd( -0.5, yMove );
        
        return new Shape3D();
    }
    class ColorInterp {
        private List colors = new ArrayList();
        
        void addColor( java.awt.Color c ) {
            colors.add( c );
        }
        
        /** Returns a packed RGB integer representing the 
         *  input value, which is assumed to be between 0
         *  and 1.
         */
        int getColor( double in ) {
           Color low = (Color)colors.get( (int)Math.floor( colors.size() * in ) );
           int max = (int)Math.ceil( (int)colors.size() * in );
           if ( max > colors.size() - 1 ) { max = colors.size() - 1; }
           Color high = (Color)colors.get( max );
           int outR = low.getRed() + high.getRed() / 2;
           int outG = low.getGreen() + high.getGreen() / 2;
           int outB = low.getBlue() + high.getBlue() / 2;
           return (outR << 16) + (outG << 8) + outB;
        }
    }
}
