package molmaster.util;

import java.io.*;
import java.net.*;
import java.util.*;
import java.util.zip.*;

import javax.swing.*;

/** Checks for molmaster updates and prompts the user
 *  if a newer version is available.
 *
 * @author  Russell Power
 */
public class UpdateCheck implements Runnable {

	private File downloadFile;
	public static final int incrementAmount = 7;
	public static final DataStore updateStore = new DataStore("update");

	private static final String REMOTE_HOST =
		"http://bestla.chem.washington.edu/~molmaster";
		
	private static final String VERSION_FILE = "version.txt";
	private static final String REMOTE_FILE = "files/update.zip";

	/** Creates a new UpdateCheck object.
	 */
	public UpdateCheck() {
	}

	/** This instance of UpdateCheck will attempt to
	 *  download the latest version number of molmaster
	 *  from molmaster.sourceforge.net/version.txt.  If
	 *  successful, it compares the version found with
	 *  the version of molmaster currently running, and
	 *  provides a prompt if they are not the same.
	 */
	public void run() {
		molmaster.Log.writeLog("Checking for file updates...");
		Object o = updateStore.get("lastUpdateTime");
		if (o != null) {
			java.util.Date d = (java.util.Date) o;
			Calendar c = new GregorianCalendar();
			Date currentTime = c.getTime();
			if (currentTime.before(d)) {
				molmaster.Log.writeLog(
					"Waiting until hold period is complete.");
				return;
			}
		}

		try {
			downloadFile = File.createTempFile("mol", "");

			File ver = new File("./data/" + VERSION_FILE);
			FileReader fr = new FileReader(ver);
			BufferedReader buf = new BufferedReader(fr);
			final String buildNum = buf.readLine();
			buf.close();
			fr.close();

			URL location = new URL(REMOTE_HOST + "/version.txt");
			InputStream is = location.openStream();
			Reader ir = new InputStreamReader(is);
			buf = new BufferedReader(ir);
			final String newBuild = buf.readLine();
			String changelog = "";
			String line = "";
			while ((line = buf.readLine()) != null) {
				changelog += line + "\n";
			}
			is.close();

			final UpdateDialog update =
				new UpdateDialog(buildNum, newBuild, changelog);

			if (!(newBuild.equals(buildNum))) {
				SwingUtilities.invokeAndWait(new Runnable() {
					public void run() {
						update.show();
					}
				});
			} else {
				molmaster.Log.writeLog("Version is up-to-date.");
				return;
			}

			int updateCode = update.getReturnStatus();
			if (updateCode == UpdateDialog.RET_CANCEL) {
				Calendar c = new GregorianCalendar();
				c.add(Calendar.DATE, incrementAmount);
				updateStore.put("lastUpdateTime", c.getTime());
				updateStore.flush();
				molmaster.Log.writeLog("Canceling update.");
				return;
			} else {
				downloadUpdate();
			}

			/** If we've made it this far, we've successfully downloaded
			 *  the molmaster jar file, and are ready to replace the existing
			 *  one.
			 */
			FileOutputStream fOut = new FileOutputStream(ver);
			fOut.write(newBuild.getBytes());
			fOut.close();

			unzip(downloadFile);

			downloadFile.delete();

			JOptionPane.showMessageDialog(
				molmaster.gui.MainFrame.getActiveWindow(),
				"Update completed.  Molmaster will now shutdown - please restart to\n"
					+ "continue your session.");

			Runtime.getRuntime().halt(0);

		} catch (Exception httpException) {
			molmaster.Log.writeLog(httpException);
		}
	}

	private int fileCount = 0;
	private int currentP = 1;
	private ProgressMonitor progress = null;

	/** Extracts entries from the given zip file
	 *  and writes them into the current directory.
	 */
	protected void unzip(File f) throws Exception {

		File curDir = new File(".");

		ZipFile zf = new ZipFile(f);
		fileCount = zf.size();
		zf.close();

		progress =
			new ProgressMonitor(
				molmaster.gui.MainFrame.getActiveWindow(),
				"Extracting update contents.",
				"Current file: ",
				0,
				fileCount);
		progress.setMillisToDecideToPopup(0);
		progress.setMillisToPopup(0);
		progress.setProgress(1);

		FileInputStream fs = new FileInputStream(f);
		BufferedInputStream buf = new BufferedInputStream(fs);
		ZipInputStream zipIn = new ZipInputStream(buf);

		ZipEntry e = null;
		for (int i = 0; i < fileCount; i++) {
			e = zipIn.getNextEntry();
			writeFile(e, zipIn);
			zipIn.closeEntry();
		}

		System.out.println("Finished extraction.");
		progress.setProgress(fileCount);
		progress.close();
		System.out.println("Windowing system shutdown.");

		zipIn.close();
		System.out.println("Closed zip file.");
		buf.close();
		System.out.println("Clearing buffer.");
		fs.close();
		System.out.println("Done.");
	}

	static final byte[] fileBuffer = new byte[16384];

	protected void writeFile(ZipEntry e, ZipInputStream in) throws Exception {

		progress.setNote(e.getName());
		progress.setProgress(currentP++);
		if (e.isDirectory()) {
			new File(e.getName()).mkdirs();
			return;
		}

		System.out.print(e.getName() + ": Size = " + e.getSize());
		int remaining = (int) e.getSize();
		FileOutputStream fs = new FileOutputStream(e.getName());
		BufferedOutputStream bufOut = new BufferedOutputStream(fs);

		while (remaining > 0) {
			int toRead = remaining;
			if (toRead > 16384) {
				toRead = 16384;
			}

			int read = in.read(fileBuffer, 0, toRead);
			bufOut.write(fileBuffer, 0, read);

			remaining -= read;
		}

		System.out.println(" ...done");
		bufOut.close();
		fs.close();

		Thread.yield();
	}

	protected void downloadUpdate() throws Exception {
		updateStore.remove("lastUpdateTime");
		updateStore.flush();

		FileOutputStream fOut = new FileOutputStream(downloadFile);
		BufferedOutputStream bOut = new BufferedOutputStream(fOut);

		URL data =
			new URL(REMOTE_HOST + "/" + REMOTE_FILE);
		URLConnection connect = data.openConnection();
		InputStream zipIn = connect.getInputStream();
		int length = connect.getContentLength();

		ProgressMonitorInputStream progressIn =
			new ProgressMonitorInputStream(
				molmaster.gui.MainFrame.getActiveWindow(),
				"Downloading new version...",
				zipIn);
		ProgressMonitor pg = progressIn.getProgressMonitor();
		pg.setMaximum(length);
		pg.setMinimum(0);
		pg.setMillisToPopup(0);
		pg.setMillisToDecideToPopup(0);
		pg.setNote("Starting download...");

		byte[] buffer = new byte[4096];
		int len = 0;
		int cur = 0;
		while ((len = progressIn.read(buffer)) > 0) {
			if (pg.isCanceled()) {
				zipIn.close();
				bOut.close();
				fOut.close();
				downloadFile.delete();
				return;
			}
			cur += len;
			pg.setNote("Downloaded " + cur + " bytes of: " + length);
			bOut.write(buffer, 0, len);
		}
		zipIn.close();
		bOut.close();
		fOut.close();
	}

}
